//   Copyright (c)  2003-2011  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/BigInt.H"
#include "CoCoA/OpenMath.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"

#include<algorithm>
using std::swap;

#include<cmath>
using std::log;
using std::pow;
using std::abs;

#include<cstdlib>
using std::abs;

#include <iostream>
using std::ostream;
using std::istream;
using std::ios;

#include <limits>
using std::numeric_limits;

#include<string>
using std::string;



namespace CoCoA
{

  BigInt::BigInt()
  {
    mpz_init(myRepr);
  }


  // CAREFUL!  BigInt(0) goes to another ctor below (from mpz_t)
  BigInt::BigInt(const MachineInt& n)
  {
    if (IsNegative(n))
      mpz_init_set_si(myRepr, AsSignedLong(n));
    else
      mpz_init_set_ui(myRepr, AsUnsignedLong(n));
  }


  BigInt::BigInt(const std::string& str)
  {
    mpz_init(myRepr);
    convert(*this, str);
  }

//   BigInt::BigInt(const char* str)
//   {
//     mpz_init(myRepr);
//     // Must test for zero ptr here, because the call BigInt(0) comes here
//     // and not to the ctor taking a MachineInt!!
//     if (str != 0)
//       convert(*this, str);
//   }


  // CAREFUL!!  BigInt(0) comes here!!
  BigInt::BigInt(const mpz_t N)
  {
    if (N == 0/*nullptr*/)
    {
      mpz_init(myRepr);
      return;
    }
    mpz_init_set(myRepr, N);
  }

//   BigInt::BigInt(MP_INT to_be_owned)
//   {
//     myRepr[0] = to_be_owned;
//   }


//   BigInt::BigInt(CopyFromMPZMarker, const mpz_t to_be_copied)
//   {
//     mpz_init_set(myRepr, to_be_copied);
//   }

// COPY CONSTRUCTOR

  BigInt::BigInt(const BigInt& from)
  {
    mpz_init_set(myRepr, from.myRepr);
  }


  BigInt& BigInt::operator=(const BigInt& rhs)
  {
    if (this == &rhs) return *this;
    mpz_set(myRepr, rhs.myRepr);
    return *this;
  }



  void swap(BigInt& a, BigInt& b)
  {
    if (&a == &b) return;
    mpz_swap(mpzref(a), mpzref(b));
  }


  ostream& operator<<(ostream& out, const BigInt& N)
  {
//    Next two lines should do it all...
//    mpz_class Ncopy(N.myRepr);
//    return out << Ncopy;

    // Dispose of the exceptional case n=0 immediately.
    if (IsZero(N)) return out << '0';
    const int base = (out.flags() & ios::oct) ? 8 :
      (out.flags() & ios::hex) ? 16 : 10;

    // prefix is sign and/or base specification
    string prefix;

    // Firstly put the sign in the prefix...
    if (N < 0) prefix += '-';
    else if (out.flags() & ios::showpos) prefix += '+';

    // Next put base indication in the prefix if required...
    if (out.flags() & ios::showbase)
    {
      if (base == 8 || base == 16) prefix += '0';
      if (base == 16) prefix += 'x';
    }

    // Finally compute the digits... (without leading sign)
    string digits;
    bool OK;
    if (N > 0) OK = IsConvertible(digits, N, base);
    else OK = IsConvertible(digits, -N, base);
    if (!OK)
      return out << "(NUMBER TOO BIG: "
                 << NumDigits(N, base) << " digits base " << base << ")";
    out << prefix << digits;
    return out;
  }



  namespace  // anonymous namespace for file local functions
  {

    // Returns true if c is a valid digit for the given base
    // (both upper and lower cases are allowed for hexadecimal).
    // Note: base must be 8, 10 or 16 (otherwise always gives false)
    bool IsDigitBase(char c, int base) // base = 8, 10 or 16
    {
      switch(base)
      {
          case 10:
            return isdigit(c);
          case 16:
            return isxdigit(c);
          case 8:
            return (isdigit(c) &&  c != '8' && c != '9');
          default:
            return false;
      }
    }
  } // end of anonymous namespace

  istream& operator>>(istream& in, BigInt& N)
  {
    if (!in.good()) return in;
//     mpz_class buf;
//     in >> buf;
//     mpz_set(N.myRepr, buf.get_mpz_t());

    const int base = (in.flags() & ios::oct) ? 8 : (in.flags() & ios::hex) ? 16 : 10;
    ws(in);

    // Look for sign of number.
    const char FirstChar = in.peek(); // this may set eofbit
    if (!in.good()) return in;
    if (FirstChar == '-' || FirstChar == '+')
      in.ignore();

    // Read in as many digits as possible.
    string digits;
    while (true)
    {
      const char ch = in.peek(); // this may set eofbit
      if (!in.good() || !IsDigitBase(ch, base)) break;
      in.ignore();
      digits += ch;
    }
    if (digits.empty())
    {
      if (FirstChar == '-' || FirstChar == '+')
        in.putback(FirstChar);
      in.setstate(ios::failbit); // set failbit (perhaps in addition to eofbit)
      return in;
    }
    // We found some digits, so convert them into a number.
    convert(N, digits); // could throw if number is huge.
    if (FirstChar == '-') negate(N);
    return in;
  }


  OpenMathOutput& operator<<(OpenMathOutput& OMOut, const BigInt& N)
  {
    OMOut->mySend(N);
    return OMOut;
  }


  OpenMathInput& operator>>(OpenMathInput& OMIn, BigInt& /*N*/)
  {
    CoCoA_ERROR(ERR::NYI, "OpenMathInput fns");
    return OMIn;
  }


/////////////////////////////////////////////////////////////////////////////
// STANDARD ARITHMETIC

  const BigInt operator-(const BigInt& N)
  {
    BigInt ans(N);
    negate(ans);
    return ans;
  }


  const BigInt operator+(const BigInt& N1, const BigInt& N2)
  {
    BigInt ans;
    mpz_add(mpzref(ans), mpzref(N1), mpzref(N2));
    return ans;
  }


  const BigInt operator-(const BigInt& N1, const BigInt& N2)
  {
    BigInt ans;
    mpz_sub(mpzref(ans), mpzref(N1), mpzref(N2));
    return ans;
  }


  const BigInt operator*(const BigInt& N1, const BigInt& N2)
  {
    BigInt ans;
    mpz_mul(mpzref(ans), mpzref(N1), mpzref(N2));
    return ans;
  }


  const BigInt operator/(const BigInt& N1, const BigInt& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "BigInt / BigInt");
//???    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "BigInt / BigInt");
    BigInt ans;
    mpz_tdiv_q(mpzref(ans), mpzref(N1), mpzref(N2));
    return ans;
  }


  const BigInt operator%(const BigInt& N1, const BigInt& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::ZeroModulus, "BigInt % BigInt");
//???    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "BigInt % BigInt");
    BigInt ans;
    mpz_tdiv_r(mpzref(ans), mpzref(N1), mpzref(N2));
    return ans;
  }


  const BigInt power(const BigInt& base, const BigInt& exponent)
  {
    if (exponent < 0)
      CoCoA_ERROR(ERR::NegExp, "power(BigInt, BigInt)");
    if (IsZero(base) && IsZero(exponent))
      CoCoA_ERROR(ERR::BadPwrZero, "power(BigInt, BigInt)");
    // Could allow high powers of 0 and +/-1.  Is it worth it?
    unsigned long exp;
    if (!IsConvertible(exp, exponent))
      CoCoA_ERROR(ERR::ExpTooBig, "power(BigInt, BigInt)");
    BigInt ans;
    mpz_pow_ui(mpzref(ans), mpzref(base), exp);
    return ans;
  }


  const BigInt power(const BigInt& base, const MachineInt& exponent)
  {
    if (IsNegative(exponent))
      CoCoA_ERROR(ERR::NegExp, "power(BigInt, MachineInt)");
    if (IsZero(base) && IsZero(exponent))
      CoCoA_ERROR(ERR::BadPwrZero, "power(BigInt, MachineInt)");
    BigInt ans;
    mpz_pow_ui(mpzref(ans), mpzref(base), AsUnsignedLong(exponent));
    return ans;
  }


  const BigInt power(const MachineInt& base, const BigInt& exponent)
  {
    const char* const FnName = "power(MachineInt, BigInt)";
    if (exponent < 0)
      CoCoA_ERROR(ERR::NegExp, FnName);
    if (IsZero(base) && IsZero(exponent))
      CoCoA_ERROR(ERR::BadPwrZero, FnName);
    // BUG???  Could allow high powers of 0 and +/-1.  Is it worth it?
    unsigned long exp;
    if (!IsConvertible(exp, exponent))
      CoCoA_ERROR(ERR::ExpTooBig, FnName);
    const bool NegateResult = IsNegative(base) & IsOdd(exp);
    const unsigned long B = abs(base);
    BigInt ans;
    mpz_ui_pow_ui(mpzref(ans), B, exp);
    if (NegateResult)
      negate(ans);
    return ans;
  }


  const BigInt power(const MachineInt& base, const MachineInt& exponent)
  {
    const char* const FnName = "power(MachineInt, MachineInt)";
    if (IsNegative(exponent))
      CoCoA_ERROR(ERR::NegExp, FnName);
    if (IsZero(base) && IsZero(exponent))
      CoCoA_ERROR(ERR::BadPwrZero, FnName);

    const bool NegateResult = IsNegative(base) & IsOdd(AsUnsignedLong(exponent));
    const unsigned long B = abs(base);
    BigInt ans;
    mpz_ui_pow_ui(mpzref(ans), B, AsUnsignedLong(exponent));
    if (NegateResult)
      negate(ans);
    return ans;
  }


// !!!!! SEE RANDOMZZSTREAM !!!!!
//   const BigInt random(const BigInt& lo, const BigInt& hi)
//   {
//     if (lo > hi) CoCoA_ERROR(ERR::BadArg, "random(lo,hi)");
//     if (lo == hi) return lo;
//     mpz_t rnd, range;
//     mpz_init(range);
//     mpz_sub(range, mpzref(hi), mpzref(lo));
// //    if (mpz_sgn(range) < 0) mpz_neg(range, range);
//     mpz_init(rnd);
//     mpz_random(rnd, 1+mpz_size(range));
//     mpz_add_ui(range, range, 1);
//     mpz_mod(rnd, rnd, range);
//     mpz_clear(range);
//     if (lo > hi) mpz_add(rnd, rnd, mpzref(hi));
//     else mpz_add(rnd, rnd, mpzref(lo));
//     return rnd[0];
//   }

//   const BigInt random(const MachineInt& lo, const MachineInt& hi)
//   {
//     return random(BigInt(lo), BigInt(hi));
//   }

//   const BigInt random(const MachineInt& lo, const BigInt& hi)
//   {
//     return random(BigInt(lo), hi);
//   }

//   const BigInt random(const BigInt& lo, const MachineInt& hi)
//   {
//     return random(lo, BigInt(hi));
//   }


  const BigInt RoundDiv(const BigInt& num, const MachineInt& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "RoundDiv(BigInt,MachInt)");
    return RoundDiv(num, BigInt(den));
  }

  const BigInt RoundDiv(const MachineInt& num, const BigInt& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "RoundDiv(MachInt,BigInt)");
    return RoundDiv(BigInt(num), den);
  }

  // NB halves are rounded towards zero.
  const BigInt RoundDiv(const BigInt& num, const BigInt& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "RoundDiv(BigInt,BigInt)");
    BigInt q,r;
    quorem(q, r, num, den);
    if (CmpAbs(2*r,den) > 0)
      q += sign(r)*sign(den);
    return q;
  }


  const BigInt iroot(const MachineInt& n, const MachineInt& r)
  {
    return iroot(BigInt(n), r);
  }

  const BigInt iroot(const MachineInt& n, const BigInt& R)
  {
    if (R < 1) CoCoA_ERROR(ERR::BadArg, "iroot: 2nd arg must be at least 1");
    unsigned long r;
    if (!IsConvertible(r, R)) CoCoA_ERROR(ERR::ArgTooBig, "iroot: 2nd arg is too big");
    return iroot(BigInt(n), r);
  }

  const BigInt iroot(const BigInt& N, const MachineInt& r)
  {
    if (IsNegative(r) || IsZero(r)) CoCoA_ERROR(ERR::BadArg, "iroot: 2nd arg must be at least 1");
    if (N < 0 && IsEven(r)) CoCoA_ERROR(ERR::BadArg, "iroot: even root of negative number");
    BigInt ans;
    mpz_root(mpzref(ans), mpzref(N), AsUnsignedLong(r));
    return ans;
  }

  const BigInt iroot(const BigInt& N, const BigInt& R)
  {
    if (R < 1) CoCoA_ERROR(ERR::BadArg, "iroot: 2nd arg must be at least 1");
    unsigned long r;
    if (!IsConvertible(r, R)) CoCoA_ERROR(ERR::ArgTooBig, "iroot: 2nd arg is too big");
    return iroot(N, r);
  }


  bool IsExactIroot(long& ans, const MachineInt& n, const MachineInt& r)
  {
    BigInt root;
    const bool IsExact = IsExactIroot(root, BigInt(n), r);
    if (!IsConvertible(ans, root))
      CoCoA_ERROR(ERR::ArgTooBig, "IsExactIroot");
    return IsExact;
  }

  bool IsExactIroot(BigInt& ans, const MachineInt& n, const MachineInt& r)
  {
    return IsExactIroot(ans, BigInt(n), r);
  }

  bool IsExactIroot(long& ans, const MachineInt& n, const BigInt& R)
  {
    unsigned long r;
    if (!IsConvertible(r, R))
      CoCoA_ERROR(ERR::ArgTooBig, "IsExactIroot: root index must be >=1 & fit into a ulong");
    BigInt root;
    const bool IsExact = IsExactIroot(root, BigInt(n), r);
    if (!IsConvertible(ans, root))
      CoCoA_ERROR(ERR::ArgTooBig, "IsExactIroot");
    return IsExact;
  }

  bool IsExactIroot(BigInt& ans, const MachineInt& n, const BigInt& R)
  {
    unsigned long r;
    if (!IsConvertible(r, R))
      CoCoA_ERROR(ERR::ArgTooBig, "IsExactIroot: root index must be >=1 & fit into a ulong");
    return IsExactIroot(ans, BigInt(n), r);
  }

  bool IsExactIroot(BigInt& ans, const BigInt& N, const MachineInt& r)
  {
    if (IsNegative(r) || IsZero(r))
      CoCoA_ERROR(ERR::BadArg, "IsExactIroot: root index must be at least 1");
    if (N < 0 && IsEven(AsUnsignedLong(r)))
      CoCoA_ERROR(ERR::BadArg, "IsExactIroot: even root of negative number");
    const bool IsExact = mpz_root(mpzref(ans), mpzref(N), AsUnsignedLong(r));
    return IsExact;
  }

  bool IsExactIroot(BigInt& ans, const BigInt& N, const BigInt& R)
  {
    unsigned long r;
    if (!IsConvertible(r, R))
      CoCoA_ERROR(ERR::ArgTooBig, "IsExactIroot: root index must be >=1 & fit into a ulong");
    return IsExactIroot(ans, N, r);
  }


  long isqrt(const MachineInt& n)
  {
    if (IsNegative(n)) CoCoA_ERROR(ERR::NotNonNegative, "isqrt(n)");
    long ans;
    IsConvertible(ans, isqrt(BigInt(n))); // conversion will succeed.
    return ans;
  }

  const BigInt isqrt(const BigInt& N)
  {
    if (N < 0) CoCoA_ERROR(ERR::NotNonNegative, "isqrt(N)");
    BigInt ans;
    mpz_sqrt(mpzref(ans), mpzref(N));
    return ans;
  }


  bool IsSquare(const MachineInt& n)
  {
    return IsSquare(BigInt(n));
  }

  bool IsSquare(const BigInt& N)
  {
    return mpz_perfect_square_p(mpzref(N));
  }


  bool IsPower(const MachineInt& n)
  {
    return IsPower(BigInt(n));
  }

  bool IsPower(const BigInt& N)
  {
    return mpz_perfect_power_p(mpzref(N));
  }


  bool IsZero(const BigInt& N)
  {
    return mpz_sgn(mpzref(N))==0;
  }


  bool IsOne(const BigInt& N)
  {
    return mpz_cmp_ui(mpzref(N), 1)==0;
  }


  bool IsMinusOne(const BigInt& N)
  {
    return mpz_cmp_si(mpzref(N),-1)==0;
  }


  int sign(const BigInt& N)
  {
    return mpz_sgn(mpzref(N));
  }


  bool IsOdd(const BigInt& N)
  {
    return mpz_odd_p(mpzref(N));
  }


  bool IsEven(const BigInt& N)
  {
    return mpz_even_p(mpzref(N));
  }


  long NumDigits(const BigInt& N, long base)
  {
    if (base < 2 || base > 36)
      CoCoA_ERROR(ERR::BadNumBase, "NumDigits(BigInt,base)");
    return NumericCast<long>(mpz_sizeinbase(mpzref(N), base));
  }


  double mantissa(const BigInt& N)
  {
    long DiscardExponent;
    return mpz_get_d_2exp(&DiscardExponent, mpzref(N));
  }


  long exponent(const BigInt& N)
  {
    return NumericCast<long>(mpz_sizeinbase(mpzref(N), 2));
  }


  double log(const BigInt& N)
  {
    if (IsZero(N))
      CoCoA_ERROR(ERR::LogZero, "log(BigInt)");
    long exp;
    const double mantissa = mpz_get_d_2exp(&exp, mpzref(N));
    const static double log2 = std::log(2.0);
    return std::log(std::abs(mantissa)) + exp*log2;
  }


  long ILogBase(const MachineInt& n, const MachineInt& base)
  {
    return ILogBase(BigInt(n), BigInt(base));
  }

  long ILogBase(const MachineInt& n, const BigInt& base)
  {
    return ILogBase(BigInt(n), base);
  }

  long ILogBase(const BigInt& N, const MachineInt& base)
  {
    return ILogBase(N, BigInt(base));
  }

  long ILogBase(const BigInt& N, const BigInt& base)
  {
    if (abs(base) < 2) CoCoA_ERROR(ERR::BadArg, "ILogBase: base must be at least 2");
    if (IsZero(N)) CoCoA_ERROR(ERR::LogZero, "ILogBase");
    const double ApproxLog = log(N)/log(base);
    // Check whether ApproxLog is far from integer.
    const double delta = 5* ApproxLog * numeric_limits<double>::epsilon();
    const long candidate = static_cast<long>(floor(ApproxLog+delta)); // probably right but could be too big by 1
    if (std::abs(ApproxLog - candidate) > delta)
      return candidate;
    // ApproxLog was close to integer, so we make a slow but sure check.
    const BigInt pwr = power(base, candidate);
    const int test = cmp(abs(N), pwr);
    if (test == 0) return candidate;
    if (test < 0) return candidate-1;
    if (abs(N) >= base*pwr) return candidate+1;
    return candidate;
  }


  // This fn comes out a long horrible mess because I cannot return a
  // machine int as a BigInt.  There must be a design error somewhere.
  BigInt RangeFactorial(const MachineInt& lo, const MachineInt& hi)
  {
    BigInt ans;
    // If zero is in the range, result is zero
    if ((IsZero(lo) || IsNegative(lo)) && !IsNegative(hi)) return ans;
    ans = 1;
    // First some checks for empty products.
    if (IsNegative(hi) && !IsNegative(lo)) return ans;
    // We now know that hi and lo have the same sign.
    // Two more checks for empty products.
    if (IsNegative(lo) && AsSignedLong(lo) > AsSignedLong(hi)) return ans;
    if (!IsNegative(lo) && AsUnsignedLong(lo) > AsUnsignedLong(hi)) return ans;
    // We now know that the product is not empty & does not span 0.
    unsigned long l = abs(lo);
    unsigned long h = abs(hi);
    if (IsNegative(hi))
    {
      std::swap(l, h);
      if (((h^l)&1) == 0) negate(ans);
    }

    if (h-l > 15)
    {
      const unsigned long mid = l + (h-l)/2; // equiv to (l+h)/2 but avoids overflow
      return RangeFactorial(l,mid) * RangeFactorial(1+mid,h);
    }
    // From here on 0 < l < h.
    for (; l <= h; ++l)
    {
      ans *= l;
//      mpz_mul_ui(mpzref(ans), mpzref(ans), l);
    }
    return ans;
  }



  // The basic comparsion functions: they return <0, =0, or >0.

  int cmp(const BigInt& N1, const BigInt& N2)
  {
    return sign(mpz_cmp(mpzref(N1), mpzref(N2)));
  }


  int cmp(const BigInt& N1, const MachineInt& n2)
  {
    if (IsNegative(n2))
      return sign(mpz_cmp_si(mpzref(N1), AsSignedLong(n2)));
    else
      return sign(mpz_cmp_ui(mpzref(N1), AsUnsignedLong(n2)));
  }


  int cmp(const MachineInt& n1, const BigInt& N2)
  {
    return -cmp(N2, n1);
  }


  // Comparison functions between two BigInts

  bool operator==(const BigInt& N1, const BigInt& N2)
  {
    return cmp(N1, N2) == 0;
  }


  bool operator!=(const BigInt& N1, const BigInt& N2)
  {
    return cmp(N1, N2) != 0;
  }


  bool operator> (const BigInt& N1, const BigInt& N2)
  {
    return cmp(N1, N2) > 0;
  }


  bool operator>=(const BigInt& N1, const BigInt& N2)
  {
    return cmp(N1, N2) >= 0;
  }


  bool operator< (const BigInt& N1, const BigInt& N2)
  {
    return cmp(N1, N2) < 0;
  }


  bool operator<=(const BigInt& N1, const BigInt& N2)
  {
    return cmp(N1, N2) <= 0;
  }


  // Comparison functions between BigInt and MachineInt

  bool operator==(const BigInt& N1, const MachineInt& n2)
  {
    return cmp(N1, n2) == 0;
  }


  bool operator!=(const BigInt& N1, const MachineInt& n2)
  {
    return cmp(N1, n2) != 0;
  }


  bool operator> (const BigInt& N1, const MachineInt& n2)
  {
    return cmp(N1, n2) > 0;
  }


  bool operator>=(const BigInt& N1, const MachineInt& n2)
  {
    return cmp(N1, n2) >= 0;
  }


  bool operator< (const BigInt& N1, const MachineInt& n2)
  {
    return cmp(N1, n2) < 0;
  }


  bool operator<=(const BigInt& N1, const MachineInt& n2)
  {
    return cmp(N1, n2) <= 0;
  }
			

  // Comparison functions between a MachineInt and a BigInt

  bool operator==(const MachineInt& n1, const BigInt& N2)
  {
    return cmp(n1, N2) == 0;
  }


  bool operator!=(const MachineInt& n1, const BigInt& N2)
  {
    return cmp(n1, N2) != 0;
  }


  bool operator> (const MachineInt& n1, const BigInt& N2)
  {
    return cmp(n1, N2) > 0;
  }


  bool operator>=(const MachineInt& n1, const BigInt& N2)
  {
    return cmp(n1, N2) >= 0;
  }


  bool operator< (const MachineInt& n1, const BigInt& N2)
  {
    return cmp(n1, N2) < 0;
  }


  bool operator<=(const MachineInt& n1, const BigInt& N2)
  {
    return cmp(n1, N2) <= 0;
  }
			


  //---------------------------------------------------------------------------
  // Assignment and assignment arithmetic functions

  BigInt& BigInt::operator+=(const BigInt& rhs)
  {
    mpz_add(myRepr, myRepr, rhs.myRepr);
    return *this;
  }


  BigInt& BigInt::operator-=(const BigInt& rhs)
  {
    mpz_sub(myRepr, myRepr, rhs.myRepr);
    return *this;
  }


  BigInt& BigInt::operator*=(const BigInt& rhs)
  {
    mpz_mul(myRepr, myRepr, rhs.myRepr);
    return *this;
  }


  BigInt& BigInt::operator/=(const BigInt& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::DivByZero, "BigInt /= BigInt");
//???    if (rhs < 0 && !mpz_divisible_p(myRepr, mpzref(rhs)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "BigInt /= BigInt");
    mpz_tdiv_q(myRepr, myRepr, rhs.myRepr);
    return *this;
  }


  BigInt& BigInt::operator%=(const BigInt& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::ZeroModulus, "BigInt %= BigInt");
//???    if (rhs < 0 && !mpz_divisible_p(myRepr, mpzref(rhs)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "BigInt %= BigInt");
    mpz_tdiv_r(myRepr, myRepr, rhs.myRepr);
    return *this;
  }


  //---------------------------------------------------------------------------
  // Assignment and assignment arithmetic with rhs a MachineInt

  BigInt& BigInt::operator= (const MachineInt& rhs)
  {
    if (IsNegative(rhs))
      mpz_set_si(myRepr, AsSignedLong(rhs));
    else
      mpz_set_ui(myRepr, AsUnsignedLong(rhs));

    return *this;
  }


  BigInt& BigInt::operator+=(const MachineInt& rhs)
  {
    if (IsNegative(rhs))
      mpz_sub_ui(myRepr, myRepr, abs(rhs));
    else
      mpz_add_ui(myRepr, myRepr, AsUnsignedLong(rhs));
    return *this;
  }


  BigInt& BigInt::operator-=(const MachineInt& rhs)
  {
    if (IsNegative(rhs))
      mpz_add_ui(myRepr, myRepr, abs(rhs));
    else
      mpz_sub_ui(myRepr, myRepr, AsUnsignedLong(rhs));
    return *this;
  }


  BigInt& BigInt::operator*=(const MachineInt& rhs)
  {
    if (IsNegative(rhs))
      mpz_mul_si(myRepr, myRepr, AsSignedLong(rhs));
    else
      mpz_mul_ui(myRepr, myRepr, AsUnsignedLong(rhs));
    return *this;
  }


  BigInt& BigInt::operator/=(const MachineInt& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::DivByZero, "BigInt /= MachineInt");
    if (IsNegative(rhs))
    {
//???      if (!mpz_divisible_ui_p(myRepr, abs(rhs)))
//???        CoCoA_ERROR(ERR::IntDivByNeg, "BigInt /= MachineInt");
      mpz_neg(myRepr, myRepr);
      mpz_tdiv_q_ui(myRepr, myRepr, abs(rhs));
    }
    else
      mpz_tdiv_q_ui(myRepr, myRepr, AsUnsignedLong(rhs));
    return *this;
  }


  BigInt& BigInt::operator%=(const MachineInt& rhs)
  {
    if (IsZero(rhs))
      CoCoA_ERROR(ERR::ZeroModulus, "BigInt %= MachineInt");
    if (IsNegative(rhs))
    {
//???      if (!mpz_divisible_ui_p(myRepr, abs(rhs)))
//???        CoCoA_ERROR(ERR::IntDivByNeg, "BigInt %= MachineInt");
      mpz_tdiv_r_ui(myRepr, myRepr, abs(rhs));
    }
    else
      mpz_tdiv_r_ui(myRepr, myRepr, AsUnsignedLong(rhs));
    return *this;
  }



  //---------------------------------------------------------------------------
  // increment and decrement

  BigInt& BigInt::operator++()
  {
    mpz_add_ui(myRepr, myRepr, 1);
    return *this;
  }


  BigInt& BigInt::operator--()
  {
    mpz_sub_ui(myRepr, myRepr, 1);
    return *this;
  }


  const BigInt BigInt::operator++(int)  // INEFFICIENT
  {
    BigInt ans(*this);
    ++*this;
    return ans;
  }


  const BigInt BigInt::operator--(int)  // INEFFICIENT
  {
    BigInt ans(*this);
    --*this;
    return ans;
  }


  //---------------------------------------------------------------------------

  const BigInt abs(const BigInt& N)
  {
    BigInt ans;
    mpz_abs(mpzref(ans), mpzref(N));
    return ans;
  }


//   void neg(BigInt& lhs, const BigInt& rhs)
//   {
//     mpz_neg(mpzref(lhs), mpzref(rhs));
//   }


//   void abs(BigInt& lhs, const BigInt& rhs)
//   {
//     mpz_abs(mpzref(lhs), mpzref(rhs));
//   }


  //---------------------------------------------------------------------------
  // Efficient 3 argument arithmetic functions -- should be inline?

  void add(BigInt& lhs, const BigInt& N1, const BigInt& N2)
  {
    mpz_add(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void sub(BigInt& lhs, const BigInt& N1, const BigInt& N2)
  {
    mpz_sub(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void mul(BigInt& lhs, const BigInt& N1, const BigInt& N2)
  {
    mpz_mul(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void div(BigInt& lhs, const BigInt& N1, const BigInt& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "div(BigInt, BigInt, BigInt)");
//???    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "div(BigInt, BigInt, BigInt)");
    mpz_tdiv_q(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void mod(BigInt& lhs, const BigInt& N1, const BigInt& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::ZeroModulus, "mod(BigInt, BigInt, BigInt)");
//???    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "mod(BigInt, BigInt, BigInt)");
    mpz_tdiv_r(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  void quorem(BigInt& quo, BigInt& rem, const BigInt& num, const BigInt& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "quorem(BigInt, BigInt, BigInt, BigInt)");
//???    if (den < 0 && !mpz_divisible_p(mpzref(num), mpzref(den)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "quorem(BigInt, BigInt, BigInt, BigInt)");
    mpz_tdiv_qr(mpzref(quo), mpzref(rem), mpzref(num), mpzref(den));
  }


  void divexact(BigInt& lhs, const BigInt& N1, const BigInt& N2)
  {
    CoCoA_ASSERT(N1%N2 == 0);
    mpz_divexact(mpzref(lhs), mpzref(N1), mpzref(N2));
  }


  // Efficient 3 arg arithmetic where last arg is a MachineInt

  void add(BigInt& lhs, const BigInt& N1, const MachineInt& n2)
  {
    if (IsNegative(n2))
      mpz_sub_ui(mpzref(lhs), mpzref(N1), abs(n2));
    else
      mpz_add_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void sub(BigInt& lhs, const BigInt& N1, const MachineInt& n2)
  {
    if (IsNegative(n2))
      mpz_add_ui(mpzref(lhs), mpzref(N1), abs(n2));
    else
      mpz_sub_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void mul(BigInt& lhs, const BigInt& N1, const MachineInt& n2)
  {
    if (IsNegative(n2))
      mpz_mul_si(mpzref(lhs), mpzref(N1), AsSignedLong(n2));
    else
      mpz_mul_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void div(BigInt& lhs, const BigInt& N1, const MachineInt& n2)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "div(BigInt, BigInt, MachineInt)");
    if (IsNegative(n2))
    {
//???      if (!mpz_divisible_ui_p(mpzref(N1), abs(n2)))
//???        CoCoA_ERROR(ERR::IntDivByNeg, "div(BigInt, BigInt, MachineInt)");
      mpz_tdiv_q_ui(mpzref(lhs), mpzref(N1), abs(n2));
      negate(lhs);
    }
    else
      mpz_tdiv_q_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void mod(BigInt& lhs, const BigInt& N1, const MachineInt& n2)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::ZeroModulus, "mod(BigInt, BigInt, MachineInt)");
    if (IsNegative(n2))
    {
//???      if (!mpz_divisible_ui_p(mpzref(N1), abs(n2)))
//???        CoCoA_ERROR(ERR::IntDivByNeg, "mod(BigInt, BigInt, MachineInt)");
      mpz_tdiv_r_ui(mpzref(lhs), mpzref(N1), abs(n2));
    }
    else
      mpz_tdiv_r_ui(mpzref(lhs), mpzref(N1), AsUnsignedLong(n2));
  }


  void quorem(BigInt& quo, BigInt& rem, const BigInt& num, const MachineInt& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "quorem(BigInt, BigInt, BigInt, MachineInt)");
    if (IsNegative(den))
    {
//???      if (!mpz_divisible_ui_p(mpzref(num), abs(den)))
//???        CoCoA_ERROR(ERR::IntDivByNeg, "quorem(BigInt, BigInt, BigInt, MachineInt)");
      mpz_tdiv_qr_ui(mpzref(quo), mpzref(rem), mpzref(num), abs(den));
      negate(quo);
    }
    else
      mpz_tdiv_qr_ui(mpzref(quo), mpzref(rem), mpzref(num), AsUnsignedLong(den));
  }



  // Efficient 3 arg arithmetic where second arg is a MachineInt

  void add(BigInt& lhs, const MachineInt& n1, const BigInt& N2)
  {
    add(lhs, N2, n1);
  }


  void sub(BigInt& lhs, const MachineInt& n1, const BigInt& N2)
  {
    sub(lhs, N2, n1);
    negate(lhs);
  }


  void mul(BigInt& lhs, const MachineInt& n1, const BigInt& N2)
  {
    mul(lhs, N2, n1);
  }


  // Impl is simple rather than fast.
  void div(BigInt& lhs, const MachineInt& n1, const BigInt& N2)
  {
    if (N2 == 0)
      CoCoA_ERROR(ERR::DivByZero, "div(BigInt, MachineInt, BigInt)");
    BigInt N1(n1);
//???    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
//???        CoCoA_ERROR(ERR::IntDivByNeg, "div(BigInt, MachineInt, BigInt)");
    div(lhs, N1, N2);
  }


  // Impl is simple rather than fast.
  void mod(BigInt& lhs, const MachineInt& n1, const BigInt& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::ZeroModulus, "mod(BigInt, MachineInt, BigInt)");
    BigInt N1(n1);
//???    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
//???        CoCoA_ERROR(ERR::IntDivByNeg, "mod(BigInt, MachineInt, BigInt)");
    mod(lhs, N1, N2);
  }


  // Impl is simple rather than fast.
  void quorem(BigInt& quo, BigInt& rem, const MachineInt& num, const BigInt& den)
  {
    if (IsZero(den))
      CoCoA_ERROR(ERR::DivByZero, "quorem(BigInt, BigInt, MachineInt, BigInt)");
    BigInt Num(num);
//???    if (den < 0 && !mpz_divisible_p(mpzref(Num), mpzref(den)))
//???        CoCoA_ERROR(ERR::IntDivByNeg, "quorem(BigInt, BigInt, MachineInt, BigInt)");
    quorem(quo, rem, Num, den);
  }



  //---------------------------------------------------------------------------

  const BigInt factorial(const BigInt& N)
  {
    if (N < 0)
      CoCoA_ERROR(ERR::NotNonNegative, "factorial(BigInt)");
    unsigned long n;
    if (!IsConvertible(n, N))
      CoCoA_ERROR(ERR::ArgTooBig, "factorial(BigInt)");
//    return factorial(n);
    BigInt ans;
    mpz_fac_ui(mpzref(ans), n);
    return ans;
  }


  const BigInt factorial(const MachineInt& n)
  {
    if (IsNegative(n))
      CoCoA_ERROR(ERR::NotNonNegative, "factorial(MachineInt)");
    BigInt ans;
    mpz_fac_ui(mpzref(ans), AsUnsignedLong(n));
    return ans;
  }


  const BigInt binomial(const BigInt& N, const BigInt& R)
  {
//    if (R < 0) return BigInt(0);
    if (R < 0)
      CoCoA_ERROR(ERR::BadArg, "binomial(BigInt,BigInt): 2nd arg must be non-neg");
    if (N < 0)
    {
      if (IsEven(R)) return binomial(R-N-1,R);
      else return -binomial(R-N-1,R);
    }
    if (R > N)
      return BigInt(0);

    // General case:  N >= R >= 0
    unsigned long r;
    if ((2*R < N && !IsConvertible(r, R)) || !IsConvertible(r, N-R))
      CoCoA_ERROR(ERR::ArgTooBig, "binomial(BigInt, BigInt)");

    BigInt ans;
    mpz_bin_ui(mpzref(ans), mpzref(N), r); // we know that r >= 0.
    return ans;
  }


  const BigInt binomial(const MachineInt& n, const BigInt& R)
  {
    if (R < 0)
      CoCoA_ERROR(ERR::BadArg, "binomial(MachineInt,BigInt): 2nd arg must be non-neg");
    return binomial(BigInt(n), R);
  }


  const BigInt binomial(const BigInt& N, const MachineInt& r)
  {
//    if (IsNegative(r)) return BigInt(0);
    if (IsNegative(r))
      CoCoA_ERROR(ERR::BadArg, "binomial(BigInt,MachineInt): 2nd arg must be non-neg");
    BigInt ans;
    mpz_bin_ui(mpzref(ans), mpzref(N), AsUnsignedLong(r));
    return ans;
  }


  const BigInt binomial(const MachineInt& n, const MachineInt& r)
  {
//    if (IsNegative(r)) return BigInt(0);
    if (IsNegative(r))
      CoCoA_ERROR(ERR::BadArg, "binomial(MachineInt,MachineInt): 2nd arg must be non-neg");

    BigInt ans;
    if (IsNegative(n))
      mpz_bin_ui(mpzref(ans), mpzref(BigInt(n)), AsUnsignedLong(r));
    else
      mpz_bin_uiui(mpzref(ans), AsUnsignedLong(n), AsUnsignedLong(r));
    return ans;
  }


  const BigInt fibonacci(const BigInt& N)
  {
    unsigned long n;
    if (!IsConvertible(n, N))
      CoCoA_ERROR(ERR::ArgTooBig, "fibonacci(BigInt)");
    return fibonacci(n);
  }


  const BigInt fibonacci(const MachineInt& n)
  {
    const bool EvenNegative = IsNegative(n) && IsEven(n);
    BigInt ans;
    unsigned long arg = abs(n);
    mpz_fib_ui(mpzref(ans), arg);
    if (EvenNegative)
      negate(ans);
    return ans;
  }

  //---------------------------------------------------------------------------

  const BigInt operator+(const BigInt& N1, const MachineInt& n2)
  {
    BigInt ans;
    if (IsNegative(n2))
      mpz_sub_ui(mpzref(ans), mpzref(N1), abs(n2));
    else
      mpz_add_ui(mpzref(ans), mpzref(N1), AsUnsignedLong(n2));
    return ans;
  }


  const BigInt operator-(const BigInt& N1, const MachineInt& n2)
  {
    BigInt ans;
    if (IsNegative(n2))
      mpz_add_ui(mpzref(ans), mpzref(N1), abs(n2));
    else
      mpz_sub_ui(mpzref(ans), mpzref(N1), AsUnsignedLong(n2));
    return ans;
  }


  const BigInt operator*(const BigInt& N1, const MachineInt& n2)
  {
    BigInt ans;
    if (IsNegative(n2))
      mpz_mul_si(mpzref(ans), mpzref(N1), AsSignedLong(n2));
    else
      mpz_mul_ui(mpzref(ans), mpzref(N1), AsUnsignedLong(n2));
    return ans;
  }


  const BigInt operator/(const BigInt& N1, const MachineInt& n2)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "BigInt / MachineInt");
//???    if (IsNegative(n2) && !mpz_divisible_ui_p(mpzref(N1), abs(n2)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "BigInt / MachineInt");
    BigInt ans;
    if (IsNegative(n2))
    {
      mpz_tdiv_q_ui(mpzref(ans), mpzref(N1), abs(n2));
      negate(ans);
    }
    else
      mpz_tdiv_q_ui(mpzref(ans), mpzref(N1), AsUnsignedLong(n2));
    return ans;
  }


  // This impl is valid for any second arg, but would then
  // have to return an unsigned long.  I have decided to forbid
  // 2nd args which are too big to fit into signed long; this
  // guarantees that the result will fit into signed long.
  // An alternative would be simply to NumericCast the result to
  // long (which will throw if the result doesn't fit).
  long operator%(const BigInt& N1, const MachineInt& n2)
  {
    if (IsZero(n2))
      CoCoA_ERROR(ERR::DivByZero, "BigInt % MachineInt");
    if (!IsSignedLong(n2))
      CoCoA_ERROR(ERR::ArgTooBig, "BigInt % MachineInt");
//     if (IsNegative(n2))
//     {
// //???      if (!mpz_divisible_ui_p(mpzref(N1), abs(n2)))
// //???        CoCoA_ERROR(ERR::IntDivByNeg, "BigInt % MachineInt");
//       return 0;
//     }
    const long ans = mpz_tdiv_ui(mpzref(N1), abs(n2)); // abs value of remainder
    if (sign(N1) < 0) return -ans;
    return ans;
  }


  const BigInt operator+(const MachineInt& n1, const BigInt& N2)
  {
    BigInt ans;
    if (IsNegative(n1))
      mpz_sub_ui(mpzref(ans), mpzref(N2), abs(n1));
    else
      mpz_add_ui(mpzref(ans), mpzref(N2), AsUnsignedLong(n1));
    return ans;
  }


  const BigInt operator-(const MachineInt& n1, const BigInt& N2)
  {
    BigInt ans;
    if (IsNegative(n1))
      mpz_add_ui(mpzref(ans), mpzref(N2), abs(n1));
    else
      mpz_sub_ui(mpzref(ans), mpzref(N2), AsUnsignedLong(n1));
    negate(ans);
    return ans;
  }


  const BigInt operator*(const MachineInt& n1, const BigInt& N2)
  {
    BigInt ans;
    if (IsNegative(n1))
      mpz_mul_si(mpzref(ans), mpzref(N2), AsSignedLong(n1));
    else
      mpz_mul_ui(mpzref(ans), mpzref(N2), AsUnsignedLong(n1));
    return ans;
  }


  // Impl is simple rather than fast.
  const BigInt operator/(const MachineInt& n1, const BigInt& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "MachineInt / BigInt");
//???    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "MachineInt / BigInt");
    BigInt ans(n1);
    ans /= N2;
    return ans;
  }


  // Impl is simple rather than fast.
  const BigInt operator%(const MachineInt& n1, const BigInt& N2)
  {
    if (IsZero(N2))
      CoCoA_ERROR(ERR::DivByZero, "MachineInt % BigInt");
//???    BigInt N1(n1);
//???    if (N2 < 0 && !mpz_divisible_p(mpzref(N1), mpzref(N2)))
//???      CoCoA_ERROR(ERR::IntDivByNeg, "MachineInt % BigInt");
    BigInt ans(n1);
    ans %= N2;
    return ans;
  }


  int CmpAbs(const BigInt& N1, const BigInt& N2)
  {
    return sign(mpz_cmpabs(mpzref(N1), mpzref(N2)));
  }

  int CmpAbs(const BigInt& N1, const MachineInt& n2)
  {
    return sign(mpz_cmpabs_ui(mpzref(N1), abs(n2)));
  }

  int CmpAbs(const MachineInt& n1, const BigInt& N2)
  {
    return -sign(mpz_cmpabs_ui(mpzref(N2), abs(n1)));
  }

  int CmpAbs(const MachineInt& n1, const MachineInt& n2)
  {
    const unsigned long a1 = abs(n1);
    const unsigned long a2 = abs(n2);
    if (a1 > a2) return 1;
    if (a1 < a2) return -1;
    return 0;
  }

} // end of namespace CoCoA



// The next few lines contain RCS header/log information.
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/BigInt.C,v 1.10 2012/02/07 10:47:12 bigatti Exp $
// $Log: BigInt.C,v $
// Revision 1.10  2012/02/07 10:47:12  bigatti
// -- updated comment to "BigInt"
//
// Revision 1.9  2011/11/09 14:03:40  bigatti
// -- renamed MachineInteger --> MachineInt
//
// Revision 1.8  2011/09/13 15:49:20  abbott
// Added "static" to log2 local var in log(BigInt).
//
// Revision 1.7  2011/09/06 15:21:53  abbott
// Changed "cmp" functions so that the return value is in {-1,0,+1}.
//
// Revision 1.6  2011/08/25 06:30:08  bigatti
// -- added log from ZZ.C
//
// Revision 1.5  2011/08/23 16:16:29  abbott
// Corrected defn of RoundDiv; corrected comment too.
//
// Revision 1.4  2011/08/17 11:56:57  abbott
// Added static_cast to keep compiler quiet.
//
// Revision 1.3  2011/08/14 15:52:17  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.2  2011/08/12 16:31:03  abbott
// COMMENTED OUT SOME FNS SO THAT BigInt CAN EXIST ALONGSIDE ZZ
// FOR THE TIME BEING.
//
// Revision 1.1  2011/08/12 15:21:26  abbott
// Added BigInt impl (derived from ZZ); not used by anyone yet.
//
//------ log with old name ZZ.C -------------------------------
// Revision 1.28  2011/03/21 11:13:25  abbott
// Changed return type for operator%(ZZ,MachineInteger) from
// unsigned long to long to fit in with the new coding conventions.
//
// Revision 1.27  2011/03/14 10:32:22  abbott
// Changed size_t into long (return types of NumDigits & exponent).
//
// Revision 1.26  2011/03/01 15:27:23  abbott
// Improved impl of ILogBase -- faster in most cases.
// Changed the error objects thrown in certain cases (e.g. ERR::BadArg --> ERR::LogZero).
//
// Revision 1.25  2011/01/14 17:21:15  abbott
// Added isqrt, iroot, IsExactIroot, IsSquare, IsPower.
//
// Revision 1.24  2010/12/26 13:03:15  abbott
// Added ILogBase function (to BigInt & BigRat).
//
// Revision 1.23  2010/04/23 13:54:44  abbott
// binomial function now triggers error if 2nd arg is negative.
//
// Revision 1.22  2010/03/23 11:59:35  abbott
// Cleaned mantissa -- GMP already has a fn for this!
// Removed use of shared static "temporary" ZZ value -- caused GMPAllocator to report an error!
//
// Revision 1.21  2010/03/22 21:01:31  abbott
// Cleaned impl of binomial (& fixed a couple of silly bugs).
//
// Revision 1.20  2010/03/22 11:50:31  abbott
// Added ctor from a string.
// Fixed stupid bug in operator-.
//
// Revision 1.19  2010/03/18 16:41:08  abbott
// Commented out unused arg (in NYI fn).
//
// Revision 1.18  2010/03/18 13:54:19  abbott
// Added openmath output fns (moved from OpenMath files).
//
// Revision 1.17  2009/12/29 22:44:32  abbott
// Removed buggy proxy class ZZ::rtn.
// Consequent changes for function prototypes also in NumTheory.
// Corrected some minor buglets in NumTheory.
//
// Revision 1.16  2009/12/11 11:42:16  abbott
// Changed convert into IsConvertible.
// Cleaned impl of operator>>(istream&,ZZ&).
//
// Revision 1.15  2009/11/26 16:18:00  bigatti
// -- including string ZZ.C instead of ZZ.H
//
// Revision 1.14  2009/10/08 13:39:47  abbott
// Renamed "round" into "RoundDiv".
// Added some new versions of "RoundDiv".
// Added a test for "RoundDiv".
//
// Revision 1.13  2009/06/11 14:05:29  abbott
// CLEANING: Removed several functions which are now gathered in NumTheory.H/C
//           (for example: gcd, lcm, PowerMod, InvMod).
//
// Revision 1.12  2009/06/05 12:08:27  abbott
// Changed return type of operator%(ZZ,MachineInteger); it is now unsigned long
// instead of ZZ.
//
// Revision 1.11  2009/01/12 14:34:10  abbott
// Corrected IsDigitBase (following a bug report from anonymous)
// for octal and hexadecimal digits.  Added 3 lines of comment too.
//
// Revision 1.10  2008/12/16 21:14:47  abbott
// In functions taking a machine integer changed arg type from MachineInteger
// to const-ref-MachineInteger.
//
// Revision 1.9  2008/11/18 10:25:48  abbott
// Added function round.
//
// Revision 1.8  2008/04/22 13:09:16  abbott
// Removed IsPositive and IsNegative functions for ZZs.
// Comparison between RingElem and 0 is now handled specially (specially fast?).
//
// Revision 1.7  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.6  2007/06/01 13:56:08  abbott
// Now gives error whenever you try to compute 0^0 -- previously some cases were not checked.
//
// Revision 1.5  2007/05/22 22:51:39  abbott
// Changed name of fn ndigits to NumDigits.
// Changed return type of exponent and NumDigits.
// Changed some exceptions from nonstandard to the appropriate standard one.
//
// Revision 1.4  2007/05/21 12:57:28  abbott
// New class for passing machine integers as args; includes some simple
// operations on machine integers (cmp, gcd, IsNegative,...).  Operations
// between ZZ and machine integers modified to use the new class.  Inexact
// integer division (of a ZZ) by a negative value now triggers an error;
// new error for reporting inexact integer division by a negative value.
//
// Revision 1.3  2007/03/23 18:38:42  abbott
// Separated the "convert" functions (and CheckedCast) into their own files.
// Many consequential changes.  Also corrected conversion to doubles.
//
// Revision 1.2  2007/03/16 17:43:05  abbott
// Added new convert function (from ZZ to double).
//
// Revision 1.1.1.1  2007/03/09 15:16:11  abbott
// Imported files
//
// Revision 1.9  2007/03/08 18:22:28  cocoa
// Just whitespace cleaning.
//
// Revision 1.8  2007/01/13 14:14:34  cocoa
// Overhaul of RingHom code: it nows uses SmartPtrIRC, and printing is more logical.
// Have not yet updated the documentation.
//
// Revision 1.7  2007/01/09 15:52:08  cocoa
// Changed QBGenerator to use std::vector instead of std::list for the result.
// Minor mod to configure script.
//
// Revision 1.6  2006/11/29 11:59:35  cocoa
// -- fixed: convert(double& z, const ZZ& num, const ZZ& den) now returns
//    bool (was void) and does not throw errors
//
// Revision 1.5  2006/11/27 13:06:22  cocoa
// Anna and Michael made me check without writing a proper message.
//
// Revision 1.4  2006/10/16 23:18:59  cocoa
// Corrected use of std::swap and various special swap functions.
// Improved myApply memfn for homs of RingDistrMPolyInlPP.
//
// Revision 1.3  2006/10/06 09:36:13  cocoa
// Fixed two minor bugs introduced at last check in.
//
// Revision 1.2  2006/09/29 11:46:54  cocoa
// Corrected bug in convert(ZZ, ZZ, double) -- now correct and simpler.
// Previously went into infinite loop on negative doubles.
//
// Revision 1.1.1.1  2006/05/30 11:39:37  cocoa
// Imported files
//
// Revision 1.5  2006/05/12 16:10:58  cocoa
// Added OpenMathFwd.H, and tidied OpenMath.H.
// Many consequential but trivial changes.
//
// Revision 1.4  2006/03/27 12:21:25  cocoa
// Minor silly changes to reduce number of complaints from some compiler or other.
//
// Revision 1.3  2006/03/14 15:01:49  cocoa
// Improved the implementation of ring member fns for computing powers.
// Should keep Intel C++ compiler quieter too.
//
// Revision 1.2  2006/03/12 21:28:33  cocoa
// Major check in after many changes
//
// Revision 1.1.1.1  2005/10/17 10:46:54  cocoa
// Imported files
//
// Revision 1.2  2005/08/08 16:36:32  cocoa
// Just checking in before going on holiday.
// Don't really recall what changes have been made.
// Added IsIndet function for RingElem, PPMonoidElem,
// and a member function of OrdvArith.
// Improved the way failed assertions are handled.
//
// Revision 1.1.1.1  2005/05/03 15:47:31  cocoa
// Imported files
//
// Revision 1.3  2005/04/20 15:40:47  cocoa
// Major change: modified the standard way errors are to be signalled
// (now via a macro which records filename and line number).  Updated
// documentation in error.txt accordingly.
//
// Improved the documentation in matrix.txt (still more work to be done).
//
// Revision 1.2  2005/04/19 14:06:03  cocoa
// Added GPL and GFDL licence stuff.
//
// Revision 1.1.1.1  2005/01/27 15:12:13  cocoa
// Imported files
//
// Revision 1.8  2004/11/12 15:49:29  cocoa
// Tidying prior to 0.90 release.
// (a) documentation improved (or marked as poor)
// (b) sundry minor improvements to the code
//
// Revision 1.7  2004/11/05 15:36:57  cocoa
// Removed call to new(nothrow) inside a convert function.  Replaced
// it by use of a std::vector -- code is certainly cleaner and clearer
// now, but possibly with an inconsequential reduction in run-time
// performance.
//
// Revision 1.6  2004/11/04 18:47:42  cocoa
// (1) Ring member functions which previously expected mpz_t args
//     now expect ZZ args.  Numerous minor consequential changes.
// (2) Renamed function which gives access to the mpz_t value inside
//     a ZZ object: previously was raw(...), now is mpzref(...).
//     Plenty of calls had to be altered.
//
// Revision 1.5  2004/05/27 16:14:02  cocoa
// Minor revision for new coding conventions.
//
// Revision 1.4  2004/05/24 15:52:13  cocoa
// Major update:
//   new error mechanism
//   many fixes
//   RingHoms almost work now
//   RingFloat much improved
//
// Revision 1.3  2004/02/03 16:16:20  cocoa
// Removed pointless IamGCDDomain functions from several concrete rings.
// Added IamOrderedDomain functions where appropriate.
// Tidied ctors for the small finite fields.
//
// Revision 1.2  2003/10/17 10:51:06  cocoa
// Major cleaning, and new naming convention.
//
// Revision 1.1.1.1  2003/09/24 12:55:43  cocoa
// Imported files
//
// Revision 1.2  2003/06/23 16:13:01  abbott
// Minor cleaning prior to public release.
//
// Revision 1.1  2003/05/14 17:12:40  abbott
// Initial revision
//
