#ifdef CoCoA_WITH_NORMALIZ
//   Copyright (c)  2010-2011 Anna M. Bigatti, Christof Soeger

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Source code for Normaliz integration

#include "CoCoA/ExternalLibs-Normaliz.H"
#include "CoCoA/BigInt.H"
#include "CoCoA/TmpPPVector.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"

#include "CoCoA/io.H"  // just for debugging
#include <iostream>  // just for debugging

#include "gmpxx.h"
//#include "libnormaliz.cpp"
using libnormaliz::Cone;
using libnormaliz::ConeProperties;



// #include <list>
// using std::list;
#include <vector>
using std::vector;


//template class libnormaliz::Cone<long>;
//template class libnormaliz::Cone<mpz_class>;


namespace CoCoA
{

namespace Normaliz
{

namespace ConeImpl
{

  cone ConeLong(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType);
  cone ConeMPZ(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType);

  class LongImpl: public ConeBase
  {
  public:
    // waiting for Normaliz Team decisions on input
    LongImpl(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType);
    LongImpl(const PPVector& v, libnormaliz::Type::InputType InputType);
    friend cone ConeLong(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType);
  private:
    void myComputation(const libnormaliz::ConeProperties& CPs) const;
    void myComputation(libnormaliz::ConeProperty::Enum CP) const;    
    void myHilbertBasis(std::vector<std::vector<BigInt> >& v) const;
    void myHt1Elements(std::vector<std::vector<BigInt> >& v) const;
    void mySupportHyperplanes(std::vector<std::vector<BigInt> >& v) const;
    void myHVector(std::vector<BigInt>& v) const;
    //    void myTriangulation(std::vector<std::vector<BigInt> >& v) const;
  private:
    mutable libnormaliz::Cone<long> myConeLong; // be careful!
  };


  class MPZImpl: public ConeBase
  {
  public:
    // waiting for Normaliz Team decisions on input
    MPZImpl(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType);
    MPZImpl(const PPVector& v, libnormaliz::Type::InputType InputType);
    friend cone ConeMPZ(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType);
  private:
    void myComputation(const libnormaliz::ConeProperties& CPs) const;
    void myComputation(libnormaliz::ConeProperty::Enum CP) const;    
    void myHilbertBasis(std::vector<std::vector<BigInt> >& v) const;
    void myHt1Elements(std::vector<std::vector<BigInt> >& v) const;
    void mySupportHyperplanes(std::vector<std::vector<BigInt> >& v) const;
    void myHVector(std::vector<BigInt>& v) const;
    //    void myTriangulation(std::vector<std::vector<BigInt> >& v) const;
  private:
    mutable libnormaliz::Cone<mpz_class> myConeMPZ; // be careful
  };
} // end of namespace ConeImpl


  std::vector<std::vector<BigInt> >  HilbertBasis(const cone& c)
  {
    std::vector<std::vector<BigInt> > v;
  	c->myHilbertBasis(v);
    return v;
  }

  void Ht1Elements(std::vector<std::vector<BigInt> >& v, const cone& c)
  { c->myHt1Elements(v); }
  
  void SupportHyperplanes(std::vector<std::vector<BigInt> >& v, const cone& c)
  { c->mySupportHyperplanes(v); }
  
  void HVector(std::vector<BigInt>& v, const cone& c)
  { c->myHVector(v); }

//   void Triangulation(std::vector<std::vector<BigInt> >& v, const cone& c)
//   { c->myTriangulation(v); }

 
  namespace  // conversion functions
  {
    std::vector<BigInt> LongVToBigIntV(const std::vector<long>& VIn)
    {
      std::vector<BigInt> v;
      v.reserve(VIn.size());
      for (vector<long>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
        v.push_back(BigInt(*it));
      return v;
    }


    std::vector<BigInt> MPZ_classVToBigIntV(const std::vector<mpz_class>& VIn)
    {
      std::vector<BigInt> v;
      v.reserve(VIn.size());
      for (vector<mpz_class>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
        v.push_back(BigInt(it->get_mpz_t()));
      return v;
    }


    std::vector<long> BigIntVToLongV(const std::vector<BigInt>& VIn)
    {
      std::vector<long> v;
      long n;
      v.reserve(VIn.size());
      for (vector<BigInt>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
        if (IsConvertible(n, *it))  v.push_back(n);
        else CoCoA_ERROR(ERR::BadConvert, "normaliz: BigIntVToLongV");
      return v;
    }


    std::vector<mpz_class> BigIntVToMPZ_classV(const std::vector<BigInt>& VIn)
    {
      std::vector<mpz_class> v;
      v.reserve(VIn.size());
      for (vector<BigInt>::const_iterator it=VIn.begin(); it!=VIn.end(); ++it)
        v.push_back(mpz_class(mpzref(*it)));
      return v;
    }


    void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<long> >& l)
    {
      v.clear();      // not exception safe
      v.reserve(l.size());
      //      transform(l.begin(), l.end(), v.begin(), LongVToBigIntV);
      for (vector<vector<long> >::const_iterator it=l.begin(); it!=l.end(); ++it)
        v.push_back(LongVToBigIntV(*it));
    }    


    void ConvertFromNormaliz(std::vector<std::vector<BigInt> >& v, const std::vector<std::vector<mpz_class> >& l)
    {
      v.clear();      // not exception safe
      v.reserve(l.size());
      //      transform(l.begin(), l.end(), v.begin(), LongVToBigIntV);
      for (vector<vector<mpz_class> >::const_iterator it=l.begin(); it!=l.end(); ++it)
        v.push_back(MPZ_classVToBigIntV(*it));
    }    


    std::vector<std::vector<long> >  ReturnVecVecLong(const std::vector<std::vector<BigInt> >& v)
    {
      std::vector<std::vector<long> > l;
      //      transform(v.begin(), v.end(), l.begin(), BigIntVToLongV);
      for (vector<vector<BigInt> >::const_iterator it=v.begin(); it!=v.end(); ++it)
        l.push_back(BigIntVToLongV(*it));
      //      std::clog << l << std::endl;
      return l;
    }


    std::vector<std::vector<mpz_class> >  ReturnVecVecMPZ_class(const std::vector<std::vector<BigInt> >& v)
    {
      std::vector<std::vector<mpz_class> > l;
      //      transform(v.begin(), v.end(), l.begin(), BigIntVToLongV);
      for (vector<vector<BigInt> >::const_iterator it=v.begin(); it!=v.end(); ++it)
        l.push_back(BigIntVToMPZ_classV(*it));
      //      std::clog << l << std::endl;
      return l;
    }


  std::vector<std::vector<BigInt> > MatrixToVecVecBigInt(ConstMatrixView M)
  {
    BigInt tmp;
    vector<vector<BigInt> > v;
    for (long i=0; i<NumRows(M); ++i)
    {
      v.push_back(vector<BigInt>());
      for (long j=0; j<NumCols(M); ++j)
      {
        if (!IsInteger(tmp, M(i,j)))
          CoCoA_ERROR("Not an integer", "MatrixToVecVecBigInt");
        v[i].push_back(tmp);
      }
    }
    return v;
  }

  std::vector<std::vector<BigInt> > PPVectorToVecVecBigInt(const PPVector& ppv)
  {
    vector<BigInt> tmp;
	long len =  ppv.myLen();
    vector<vector<BigInt> > v(len);
    for (long i=0; i<len; ++i)
    {
      BigExponents(tmp,PP(ppv[i]));
      v[i]=tmp;
    }
    return v;
  }

  void VecVecBigIntToPPVector(PPVector& ppv, const std::vector<std::vector<BigInt> >& M)
  {
  	PPMonoid ppm = PPM(ppv);
	long len =  M.size();
    for (long i=0; i<len; ++i)
    {
	  ppv.myPushBack(PPMonoidElem(ppm, BigIntVToLongV(M[i])));
	  //TODO Better use PPMonoidElem(const PPMonoid& Gamma, const std::vector<BigInt>& v); which is not implemented yet
	}
  }

  } // end of anonymous namespace


namespace ConeImpl
{
  LongImpl::LongImpl(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType):
    myConeLong(ReturnVecVecLong(v), InputType)
  {}
  
  
  MPZImpl::MPZImpl(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType):
    myConeMPZ(ReturnVecVecMPZ_class(v), InputType)
  {}
  

  void LongImpl::myComputation(const libnormaliz::ConeProperties& CPs) const
  { myConeLong.compute(CPs); }


  void MPZImpl::myComputation(const libnormaliz::ConeProperties& CPs) const
  { myConeMPZ.compute(CPs); }


  void LongImpl::myComputation(libnormaliz::ConeProperty::Enum CP) const
  {
    if (!myConeLong.isComputed(CP))
      myConeLong.compute(ConeProperties(CP));
    if (!myConeLong.isComputed(CP))
      CoCoA_ERROR("not computed", "myComputation long");
  }


  void MPZImpl::myComputation(libnormaliz::ConeProperty::Enum CP) const
  {
    if (!myConeMPZ.isComputed(CP))
      myConeMPZ.compute(ConeProperties(CP));
    if (!myConeMPZ.isComputed(CP))
      CoCoA_ERROR("not computed", "myComputation MPZ");
  }


  

  void LongImpl::myHilbertBasis(std::vector<std::vector<BigInt> >& v) const
  {
    myComputation(libnormaliz::ConeProperty::HilbertBasis);
    ConvertFromNormaliz(v, myConeLong.getHilbertBasis());
  }


  void LongImpl::myHt1Elements(std::vector<std::vector<BigInt> >& v) const
  {
    myComputation(libnormaliz::ConeProperty::Ht1Elements);
    ConvertFromNormaliz(v, myConeLong.getHt1Elements());
  }


  void LongImpl::mySupportHyperplanes(std::vector<std::vector<BigInt> >& v) const
  {
    myComputation(libnormaliz::ConeProperty::SupportHyperplanes);
    ConvertFromNormaliz(v, myConeLong.getSupportHyperplanes());
  }


  void LongImpl::myHVector(std::vector<BigInt>& v) const
  {
    myComputation(libnormaliz::ConeProperty::HVector);
    v = LongVToBigIntV(myConeLong.getHVector());
  }


//   void LongImpl::myTriangulation(std::vector<std::vector<BigInt> >& v) const
//   {
//     myComputation(libnormaliz::ConeProperty::Triangulation);
//     ConvertFromNormaliz(v, myConeLong.getTriangulation());
//   }


  void MPZImpl::myHilbertBasis(std::vector<std::vector<BigInt> >& v) const
  {
    myComputation(libnormaliz::ConeProperty::HilbertBasis);
    ConvertFromNormaliz(v, myConeMPZ.getHilbertBasis());
  }


  void MPZImpl::myHt1Elements(std::vector<std::vector<BigInt> >& v) const
  {
    myComputation(libnormaliz::ConeProperty::Ht1Elements);
    ConvertFromNormaliz(v, myConeMPZ.getHt1Elements());
  }


  void MPZImpl::mySupportHyperplanes(std::vector<std::vector<BigInt> >& v) const
  {
    myComputation(libnormaliz::ConeProperty::SupportHyperplanes);
    ConvertFromNormaliz(v, myConeMPZ.getSupportHyperplanes());
  }


  void MPZImpl::myHVector(std::vector<BigInt>& v) const
  {
    myComputation(libnormaliz::ConeProperty::HVector);
    v = MPZ_classVToBigIntV(myConeMPZ.getHVector());
  }


//   void MPZImpl::myTriangulation(std::vector<std::vector<BigInt> >& v) const
//   {
//     myComputation(libnormaliz::ConeProperty::Triangulation);
//     ConvertFromNormaliz(v, myConeMPZ.getTriangulation());
//   }


} // namespace ConeImpl

  cone NewConeLong(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType)
  { return cone(new ConeImpl::LongImpl(v, InputType)); }
  
  cone NewConeBigInt(const std::vector<std::vector<BigInt> >& v, libnormaliz::Type::InputType InputType)
  { return cone(new ConeImpl::MPZImpl(v, InputType)); }

  cone NewConeLong(ConstMatrixView M, libnormaliz::Type::InputType InputType)
  { return cone(new ConeImpl::LongImpl(MatrixToVecVecBigInt(M), InputType)); }
  
  cone NewConeBigInt(ConstMatrixView M, libnormaliz::Type::InputType InputType)
  { return cone(new ConeImpl::MPZImpl(MatrixToVecVecBigInt(M), InputType)); }

  
  
  PPVector NormalToricRing(const PPVector& ppv)
  { 
    cone c = NewConeLong(PPVectorToVecVecBigInt(ppv), libnormaliz::Type::normalization);
    std::vector<std::vector<BigInt> > res = HilbertBasis(c);
    //create an emptyPPVector with the same PPMonoid
    PPVector ppv_res(PPM(ppv), DMR(ppv));
    VecVecBigIntToPPVector(ppv_res,res);
    return ppv_res;
  }

  //PPVector IntClToricRing(const PPVector& ppv) use libnormaliz::Type::integral_closure


} // namespace Normaliz
} // namespace CoCoA

// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/ExternalLibs-Normaliz.C,v 1.9 2011/11/07 11:09:32 bigatti Exp $
// $Log: ExternalLibs-Normaliz.C,v $
// Revision 1.9  2011/11/07 11:09:32  bigatti
// -- new ctors taking ConstRefRingElem
// -- added PPVector NormalToricRing(const PPVector& ppv)
// -- removed void HilbertBasis(std::vector<std::vector<BigInt> >& v, const cone& c)
//
// Revision 1.8  2011/10/04 13:03:02  bigatti
// -- new logo for gui
//
// Revision 1.7  2011/09/30 12:55:44  bigatti
// -- introduced namespace "Normaliz" and removed Normaliz from function names
// -- input of Normaliz functions in CoCoA-5 is now a matrix instead of
//    vector<vector<BigInt>>
//
// Revision 1.6  2011/08/14 15:52:17  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.5  2011/07/20 15:31:21  bigatti
// -- fixed InputType in (undefined) functions
//
// Revision 1.4  2011/07/20 13:49:37  bigatti
// -- added "Normaliz" postfix to Normaliz function calls
//
// Revision 1.3  2011/07/20 12:45:12  bigatti
// -- new normaliz interface (not yet public)
//
// Revision 1.2  2011/02/17 16:50:04  bigatti
// -- getting ready for new official veson on Normaliz: added HVector, removed Triangulation
//
// Revision 1.1  2010/11/05 14:27:18  bigatti
// -- was TmpNormaliz**
//
// Revision 1.6  2010/10/12 11:22:54  bigatti
// -- TmpNormaliz.H simplified:
//    now cone is a smart pointer to ConeBase
//    and the concrete classes are entirely in the .C file
// -- added Ht1Elements, SupportHyperplanes, Triangulation
// -- added some text in ex-Normaliz1 and 2
//
// Revision 1.5  2010/10/08 13:40:37  bigatti
// -- moved inclusion of libnormaliz.cpp and instantiation of types
//    into TmpNormalizTypes.C
//
// Revision 1.4  2010/10/08 10:39:32  bigatti
// -- extended interface for normaliz
//

#endif
