//   Copyright (c)  2007-2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


// Implementation of class PPMonoidBigEvImpl

#include "CoCoA/PPMonoidEvZZ.H"
#include "CoCoA/DivMask.H"
#include "CoCoA/MemPool.H"
#include "CoCoA/PPMonoid.H"
#include "CoCoA/PPOrdering.H"
#include "CoCoA/BigInt.H"
#include "CoCoA/assert.H"
#include "CoCoA/convert.H"
#include "CoCoA/degree.H"
#include "CoCoA/error.H"
#include "CoCoA/matrix.H"
#include "CoCoA/symbol.H"
#include "CoCoA/utils.H"

#include <algorithm>
using std::min;
using std::max;
//using std::swap;
#include <iostream>
using std::ostream;
#include<limits>
using std::numeric_limits;
#include <memory>
using std::auto_ptr;
#include <vector>
using std::vector;


namespace CoCoA
{

  /*-- class PPMonoidBigEvImpl ---------------------------------------*/
  /**

  \brief Implementation power product monoid with exponent vector

  PPMonoidBigEvImpl implements a power product monoid for safe testing.
  It stores the exponents as a std::vector<BigInt>.
  It is not designed to be efficient.

  */
  /*-----------------------------------------------------------------*/

  class PPMonoidBigEvImpl: public PPMonoidBase
  {
    typedef PPMonoidElemRawPtr RawPtr;           // just to save typing
    typedef PPMonoidElemConstRawPtr ConstRawPtr; // just to save typing

    class CmpBase
    {
    public:
      CmpBase(long NumIndets, long GradingDim);
      virtual ~CmpBase() {};
    public:
      virtual int myCmpExpvs(const BigInt* v1, const BigInt* v2) const =0;
      virtual void myWDeg(degree& d, const BigInt* v) const =0;
      virtual int myCmpWDegPartial(const BigInt* v1, const BigInt* v2, long GrDim) const =0;
      //      virtual int myCmpWDeg(const BigInt* v1, const BigInt* v2) const =0;
      int myCmpWDeg(const BigInt* v1, const BigInt* v2) const { return myCmpWDegPartial(v1, v2, myGradingDimValue); }
    protected:
      long myNumIndetsValue;
      long myGradingDimValue;
    };

    class LexImpl: public CmpBase
    {
    public:
      LexImpl(long NumIndets);
      virtual int myCmpExpvs(const BigInt* v1, const BigInt* v2) const;
      virtual void myWDeg(degree& d, const BigInt* v) const;
      virtual int myCmpWDegPartial(const BigInt* v1, const BigInt* v2, long GrDim) const;
    };


    class StdDegLexImpl: public CmpBase
    {
    public:
      StdDegLexImpl(long NumIndets);
      virtual int myCmpExpvs(const BigInt* v1, const BigInt* v2) const;
      virtual void myWDeg(degree& d, const BigInt* v) const;
      virtual int myCmpWDegPartial(const BigInt* v1, const BigInt* v2, long GrDim) const;
    };


    class StdDegRevLexImpl: public CmpBase
    {
    public:
      StdDegRevLexImpl(long NumIndets);
      virtual int myCmpExpvs(const BigInt* v1, const BigInt* v2) const;
      virtual void myWDeg(degree& d, const BigInt* v) const;
      virtual int myCmpWDegPartial(const BigInt* v1, const BigInt* v2, long GrDim) const;
    };


    class MatrixOrderingImpl: public CmpBase
    {
    public:
      MatrixOrderingImpl(long NumIndets, long GradingDim, const matrix& OrderMatrix);
      virtual int myCmpExpvs(const BigInt* v1, const BigInt* v2) const;
      virtual void myWDeg(degree& d, const BigInt* v) const;
      virtual int myCmpWDegPartial(const BigInt* v1, const BigInt* v2, long GrDim) const;
    private:
      std::vector< std::vector<BigInt> > myOrderMatrix;
    };


  public:
    PPMonoidBigEvImpl(const std::vector<symbol>& IndetNames, const PPOrdering& ord);
    ~PPMonoidBigEvImpl();
  private: // disable copy ctor and assignment
    explicit PPMonoidBigEvImpl(const PPMonoidBigEvImpl& copy);  // NEVER DEFINED -- copy ctor disabled
    PPMonoidBigEvImpl& operator=(const PPMonoidBigEvImpl& rhs); // NEVER DEFINED -- assignment disabled

  public:
    void contents() const; // FOR DEBUGGING ONLY

    long myNumIndets() const;
    const std::vector<PPMonoidElem>& myIndets() const;               ///< std::vector whose n-th entry is n-th indet as PPMonoidElem
    void mySymbols(std::vector<symbol>& SymList) const;              ///< appends PPM's symbols to SymList
    const symbol& myIndetSymbol(long indet) const;

    // The functions below are operations on power products owned by PPMonoidBigEvImpl
    const PPMonoidElem& myOne() const;
    PPMonoidElemRawPtr myNew() const;                                ///< ctor from nothing
    PPMonoidElemRawPtr myNew(PPMonoidElemConstRawPtr rawpp) const;   ///< ctor by assuming ownership
    PPMonoidElemRawPtr myNew(const std::vector<long>& expv) const;   ///< ctor from exp vector
    PPMonoidElemRawPtr myNew(const std::vector<BigInt>& EXPV) const; ///< ctor from exp vector
    void myDelete(RawPtr rawpp) const;                               ///< dtor, frees pp
    void mySwap(RawPtr rawpp1, RawPtr rawpp2) const;                 ///< swap(pp1, pp2);
    void myAssignOne(RawPtr rawpp) const;                            ///< pp = 1
    void myAssign(RawPtr rawpp, ConstRawPtr rawpp1) const;           ///< p = pp1
    void myAssign(RawPtr rawpp, const std::vector<long>& expv) const;///< pp = expv (assign from exp vector)

    void myMul(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = pp1*pp2
    void myMulIndetPower(RawPtr rawpp, long indet, long exp) const;           ///< pp *= indet^exp, assumes exp >= 0
    void myMulIndetPower(RawPtr rawpp, long indet, const BigInt& EXP) const;  ///< pp *= indet^EXP
    void myDiv(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = pp1/pp2
    void myColon(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const; ///< pp = pp1/gcd(pp1,pp2)
    void myGcd(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = gcd(pp1,pp2)
    void myLcm(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< pp = lcm(pp1,pp2)
    void myRadical(RawPtr rawpp, ConstRawPtr rawpp1) const;                   ///< pp = radical(pp1)
    void myPowerSmallExp(RawPtr rawpp, ConstRawPtr rawpp1, long exp) const;   ///< pp = pp1^exp, assumes exp >= 0

    bool myIsOne(ConstRawPtr rawpp) const;                              ///< is pp = 1?
    bool myIsIndet(long& index, ConstRawPtr rawpp) const;        ///< true iff pp is an indet
    bool myIsIndetPosPower(long& indet, BigInt& EXP, ConstRawPtr rawpp) const;
    bool myIsIndetPosPower(long& indet, long& pow, ConstRawPtr rawpp) const;
    bool myIsIndetPosPower(ConstRawPtr rawpp) const;    
    bool myIsCoprime(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;     ///< are pp1 & pp2 coprime?
    bool myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;       ///< is pp1 equal to pp2?
    bool myIsDivisible(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;   ///< does pp2 divide pp1?
    bool myIsRadical(ConstRawPtr rawpp) const;                          ///< is pp equal to its radical?

    int myCmp(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;            ///< -1,0,1 as pp1 < = > pp2
    long myStdDeg(ConstRawPtr rawpp) const;                      ///< standard degree of pp
    void myWDeg(degree& d, ConstRawPtr rawpp) const;                    ///< d = grading(pp)
    int myCmpWDeg(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const;        ///< <0, =0, >0 as wdeg(pp1) < = > wdeg(pp2)
    int myCmpWDegPartial(ConstRawPtr rawpp1, ConstRawPtr rawpp2, long GrDim ) const; ///< as myCmpWDeg wrt the first GrDim weights
    long myExponent(ConstRawPtr rawpp, long indet) const;             ///< exponent of indet in pp
    void myBigExponent(BigInt& EXP, ConstRawPtr rawpp, long indet) const; ///< EXP = exponent of indet in pp
    void myExponents(std::vector<long>& expv, ConstRawPtr rawpp) const;      ///< expv[i] = exponent(pp,i)
    void myBigExponents(std::vector<BigInt>& v, ConstRawPtr rawpp) const;  ///< get exponents, SHOULD BE myExponents ???
    void myOutputSelf(std::ostream& out) const;                      ///< out << PPM
    // INHERITED DEFINITION of virtual  void myOutput(std::ostream& out, ConstRawPtr rawpp) const;
    void myDebugPrint(std::ostream& out, ConstRawPtr rawpp) const;   ///< print pp in debugging format???


  private: // auxiliary functions
    BigInt* myExpv(RawPtr) const;
    const BigInt* myExpv(ConstRawPtr) const;

    void myComputeDivMask(DivMask& dm, const DivMaskRule& DivMaskImpl, ConstRawPtr rawpp) const; ///< used by PPWithMask
    bool myCheckExponents(const std::vector<long>& expv) const;
    bool myCheckExponents(const std::vector<BigInt>& EXPV) const;

  private: // data members
    ///@name Class members
    //@{
    const long myNumIndetsValue;
    const long myEntrySize;     ///< size in ????
    mutable MemPool myMemMgr;     // IMPORTANT: this must come *before* myIndetVector and myOnePtr.
    std::vector<PPMonoidElem> myIndetVector; ///< the indets as PPMonoidElems
    std::vector<symbol> myIndetSymbols;      ///< the names of the indeterminates
    std::auto_ptr<CmpBase> myOrdPtr;         ///< actual implementation of the ordering [should be const???]
    std::auto_ptr<PPMonoidElem> myOnePtr;
    //@}
  };


  // File local inline functions

  inline BigInt* PPMonoidBigEvImpl::myExpv(RawPtr rawpp) const
  {
    return static_cast<BigInt*>(rawpp.myRawPtr());
  }


  inline const BigInt* PPMonoidBigEvImpl::myExpv(ConstRawPtr rawpp) const
  {
    return static_cast<const BigInt*>(rawpp.myRawPtr());
  }


  bool PPMonoidBigEvImpl::myCheckExponents(const std::vector<long>& expv) const
  {
    // Check len(expv) == myNumIndetsValue.
    // Check exps are non-neg and not too big.
    if (len(expv) != myNumIndetsValue) return false;
    for (long i=0; i < myNumIndetsValue; ++i)
      if (expv[i] < 0) return false;
    return true;
  }

  bool PPMonoidBigEvImpl::myCheckExponents(const std::vector<BigInt>& expv) const
  {
    // Check len(expv) == myNumIndetsValue.
    // Check exps are non-neg and not too big.
    if (len(expv) != myNumIndetsValue) return false;
    for (long i=0; i < myNumIndetsValue; ++i)
      if (expv[i] < 0) return false;
    return true;
  }


  //----   Constructors & destructor   ----//

  PPMonoidBigEvImpl::PPMonoidBigEvImpl(const std::vector<symbol>& IndetNames, const PPOrdering& ord):
      PPMonoidBase(ord),
      myNumIndetsValue(NumIndets(ord)),
      myEntrySize(sizeof(BigInt)*NumIndets(ord)),
      myMemMgr(myEntrySize, "PPMonoidBigEvImpl.myMemMgr"),
      myIndetVector(),
      myIndetSymbols(IndetNames)
  {
    // Sanity check on the indet names given.
    CoCoA_ASSERT(myNumIndetsValue == len(myIndetSymbols));
    CoCoA_ASSERT(AreDistinct(myIndetSymbols));
    CoCoA_ASSERT(AreArityConsistent(myIndetSymbols));

    // Put the correct implementation in myOrdPtr...  [VERY UGLY CODE!!!]
    if (IsLex(ord))
      myOrdPtr.reset(new LexImpl(myNumIndetsValue));
    else if (IsStdDegLex(ord))
      myOrdPtr.reset(new StdDegLexImpl(myNumIndetsValue));
    else if (IsStdDegRevLex(ord))
      myOrdPtr.reset(new StdDegRevLexImpl(myNumIndetsValue));
    else
      myOrdPtr.reset(new MatrixOrderingImpl(myNumIndetsValue, GradingDim(ord), GetMatrix(ord)));

    myRefCountInc();  // this is needed for exception cleanliness, in case one of the lines below throws
    myOnePtr.reset(new PPMonoidElem(PPMonoid(this)));
    {
      // IMPORTANT: this block destroys pp *before* the call to myRefCountZero.
      PPMonoidElem pp(PPMonoid(this));
      vector<long> expv(myNumIndetsValue);
      for (long i=0; i < myNumIndetsValue; ++i)
      {
        expv[i] = 1;
        myAssign(raw(pp), expv);
        myIndetVector.push_back(pp);
        expv[i] = 0;
      }
    }
    myRefCountZero();
  }


  PPMonoidBigEvImpl::~PPMonoidBigEvImpl()
  {}

/////////////////////////////////////////////////////////////////////////////



  long PPMonoidBigEvImpl::myNumIndets() const
  {
    return myNumIndetsValue;
  }


  const std::vector<PPMonoidElem>& PPMonoidBigEvImpl::myIndets() const
  {
    return myIndetVector;
  }


  void PPMonoidBigEvImpl::mySymbols(std::vector<symbol>& SymList) const
  {
    SymList.insert(SymList.end(), myIndetSymbols.begin(), myIndetSymbols.end());
  }


  const symbol& PPMonoidBigEvImpl::myIndetSymbol(long indet) const
  {
    CoCoA_ASSERT(0 <= indet && indet < myNumIndetsValue);
    return myIndetSymbols[indet];
  }


  const PPMonoidElem& PPMonoidBigEvImpl::myOne() const
  {
    return *myOnePtr;
  }


  PPMonoidElemRawPtr PPMonoidBigEvImpl::myNew() const
  {
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    BigInt* expv = myExpv(rawpp);
    for (long i=0; i < myNumIndetsValue; ++i) // BUG BUG: if this throws, rawpp is leaked
      new(&expv[i]) BigInt;
    myAssignOne(rawpp); // cannot throw
    return rawpp;
  }

  PPMonoidElemRawPtr PPMonoidBigEvImpl::myNew(PPMonoidElemConstRawPtr rawcopypp) const
  {
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    BigInt* expv = myExpv(rawpp);
    for (long i=0; i < myNumIndetsValue; ++i) // BUG BUG: if this throws, rawpp is leaked
      new(&expv[i]) BigInt;
    myAssign(rawpp, rawcopypp); // cannot throw
    return rawpp;
  }


  PPMonoidElemRawPtr PPMonoidBigEvImpl::myNew(const std::vector<long>& expv) const
  {
    CoCoA_ASSERT(myCheckExponents(expv));
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    BigInt* expv1 = myExpv(rawpp);
    for (long i=0; i < myNumIndetsValue; ++i) // BUG BUG: if this throws, rawpp is leaked
      new(&expv1[i]) BigInt;
    myAssign(rawpp, expv); // cannot throw
    return rawpp;
  }


  PPMonoidElemRawPtr PPMonoidBigEvImpl::myNew(const std::vector<BigInt>& EXPV) const
  {
    CoCoA_ASSERT(myCheckExponents(EXPV));
    PPMonoidElemRawPtr rawpp(myMemMgr.alloc());
    BigInt* expv = myExpv(rawpp);
    for (long i=0; i < myNumIndetsValue; ++i) // BUG BUG: if this throws, rawpp is leaked
      new(&expv[i]) BigInt;
    for (long i = 0; i < myNumIndetsValue; ++i)
      expv[i] = EXPV[i];
    return rawpp;
  }


  void PPMonoidBigEvImpl::myAssignOne(RawPtr rawpp) const
  {
    BigInt* const expv = myExpv(rawpp);
    for (long i = 0; i < myNumIndetsValue; ++i)
      expv[i] = 0;
  }


  void PPMonoidBigEvImpl::myAssign(RawPtr rawpp, ConstRawPtr rawpp1) const
  {
    if (rawpp == rawpp1) return;

    BigInt* expv = myExpv(rawpp);
    const BigInt* const expv1 = myExpv(rawpp1);
    std::copy(&expv1[0], &expv1[myNumIndetsValue], &expv[0]);
  }

  void PPMonoidBigEvImpl::myAssign(RawPtr rawpp, const vector<long>& expv1) const
  {
    CoCoA_ASSERT(myCheckExponents(expv1));

    BigInt* const expv = myExpv(rawpp);
    for (long i = 0; i < myNumIndetsValue; ++i)
      expv[i] = expv1[i];
  }


  void PPMonoidBigEvImpl::myDelete(RawPtr rawpp) const
  {
    BigInt* expv = myExpv(rawpp);;
    for (long i=0; i < myNumIndetsValue; ++i)
      expv[i].~BigInt();
    myMemMgr.free(rawpp.myRawPtr());
  }


  void PPMonoidBigEvImpl::mySwap(RawPtr rawpp1, RawPtr rawpp2) const
  {
    if (rawpp1 == rawpp2) return;
    BigInt* expv1 = myExpv(rawpp1);
    BigInt* expv2 = myExpv(rawpp2);
    for (long i = 0; i < myNumIndetsValue; ++i)
      std::swap(expv1[i], expv2[i]);
  }


  void PPMonoidBigEvImpl::myMul(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    BigInt* expv = myExpv(rawpp);
    const BigInt* const expv1 = myExpv(rawpp1);
    const BigInt* const expv2 = myExpv(rawpp2);
    for (long i = 0; i < myNumIndetsValue; ++i)
      expv[i] = expv1[i] + expv2[i];
  }


  void PPMonoidBigEvImpl::myMulIndetPower(RawPtr rawpp, long indet, long exp) const  // assumes exp >= 0
  {
    CoCoA_ASSERT(exp >= 0);
    CoCoA_ASSERT(0 <= indet && indet < myNumIndetsValue);
    BigInt* expv = myExpv(rawpp);
    expv[indet] += exp;
  }


  // This fn overrides the default defn in PPMonoid.C (which throws for big exps)
  void PPMonoidBigEvImpl::myMulIndetPower(RawPtr rawpp, long indet, const BigInt& EXP) const
  {
    CoCoA_ASSERT(0 <= indet && indet < myNumIndetsValue);
    CoCoA_ASSERT(EXP >= 0);
    BigInt* expv = myExpv(rawpp);
    expv[indet] += EXP;
  }


  void PPMonoidBigEvImpl::myDiv(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    BigInt* expv = myExpv(rawpp);
    const BigInt* const expv1 = myExpv(rawpp1);
    const BigInt* const expv2 = myExpv(rawpp2);
    for (long i = 0; i < myNumIndetsValue; ++i)
      expv[i] = expv1[i] - expv2[i]; // assert that result is positive?
  }


  void PPMonoidBigEvImpl::myColon(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    BigInt* const expv = myExpv(rawpp);
    const BigInt* const expv1 = myExpv(rawpp1);
    const BigInt* const expv2 = myExpv(rawpp2);

    for (long i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] > expv2[i])
        expv[i] = expv1[i] - expv2[i];
      else
        expv[i] = 0;
  }


  void PPMonoidBigEvImpl::myGcd(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    BigInt* const expv = myExpv(rawpp);
    const BigInt* const expv1 = myExpv(rawpp1);
    const BigInt* const expv2 = myExpv(rawpp2);

    for (long i = 0; i < myNumIndetsValue; ++i)
      expv[i] = min(expv1[i], expv2[i]);
  }


  void PPMonoidBigEvImpl::myLcm(RawPtr rawpp, ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    // No worries about aliasing.
    BigInt* const expv = myExpv(rawpp);
    const BigInt* const expv1 = myExpv(rawpp1);
    const BigInt* const expv2 = myExpv(rawpp2);

    for (long i = 0; i < myNumIndetsValue; ++i)
      expv[i] = max(expv1[i], expv2[i]);
  }


  void PPMonoidBigEvImpl::myRadical(RawPtr rawpp, ConstRawPtr rawpp1) const
  {
    BigInt* const expv = myExpv(rawpp);
    const BigInt* const expv1 = myExpv(rawpp1);

    for (long i = 0; i < myNumIndetsValue; ++i)
      expv[i] = (expv1[i] > 0);
  }


  void PPMonoidBigEvImpl::myPowerSmallExp(RawPtr rawpp, ConstRawPtr rawpp1, long exp) const  // assumes exp >= 0
  {
    CoCoA_ASSERT(exp >= 0);
    BigInt* const expv = myExpv(rawpp);
    const BigInt* const expv1 = myExpv(rawpp1);

    for (long i = 0; i < myNumIndetsValue; ++i)
    {
      expv[i] = expv1[i]*exp;
    }
  }


  bool PPMonoidBigEvImpl::myIsOne(ConstRawPtr rawpp) const
  {
    const BigInt* const expv = myExpv(rawpp);

    for (long i = 0; i < myNumIndetsValue; ++i)
      if (!IsZero(expv[i])) return false;

    return true;
  }


  bool PPMonoidBigEvImpl::myIsIndet(long& index, ConstRawPtr rawpp) const
  {
    const BigInt* const expv = myExpv(rawpp);
    long j = myNumIndetsValue;
    for (long i = 0; i < myNumIndetsValue; ++i)
    {
      if (IsZero(expv[i])) continue;
      if (j != myNumIndetsValue || !IsOne(expv[i])) return false;
      j = i;
    }
    if (j == myNumIndetsValue) return false;
    index = j;
    return true;
  }


  bool PPMonoidBigEvImpl::myIsIndetPosPower(long& index, BigInt& EXP, ConstRawPtr rawpp) const
  {
    const BigInt* const expv = myExpv(rawpp);
    long j = myNumIndetsValue;
    for (long i = 0; i < myNumIndetsValue; ++i)
    {
      if (IsZero(expv[i])) continue;
      if (j != myNumIndetsValue || IsOne(expv[i])) return false;
      j = i;
    }
    if (j == myNumIndetsValue) return false;
    index = j;
    EXP = expv[index];
    return true;
  }


  bool PPMonoidBigEvImpl::myIsIndetPosPower(long& index, long& pow, ConstRawPtr rawpp) const
  {
    BigInt POW;
    long TmpIndex, TmpPow;
    if (!myIsIndetPosPower(TmpIndex, POW, rawpp)) return false;
    if (!IsConvertible(TmpPow, POW))
      CoCoA_ERROR(ERR::ArgTooBig, "PPMonoidBigEvImpl::myIsIndetPosPower");
    index = TmpIndex;
    pow = TmpPow;
    return true;
  }

  bool PPMonoidBigEvImpl::myIsIndetPosPower(ConstRawPtr rawpp) const
  {
    const BigInt* const expv = myExpv(rawpp);
    long j = myNumIndetsValue;
    for (long i = 0; i < myNumIndetsValue; ++i)
    {
      if (IsZero(expv[i])) continue;
      if (j != myNumIndetsValue) return false;
      j = i;
    }
    return j != myNumIndetsValue;
  }


  bool PPMonoidBigEvImpl::myIsCoprime(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const BigInt* const expv1 = myExpv(rawpp1);
    const BigInt* const expv2 = myExpv(rawpp2);

    for (long i = 0; i < myNumIndetsValue; ++i)
      if (!IsZero(expv1[i]) && !IsZero(expv2[i])) return false;

    return true;
  }


  bool PPMonoidBigEvImpl::myIsEqual(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const BigInt* const expv1 = myExpv(rawpp1);
    const BigInt* const expv2 = myExpv(rawpp2);

    for (long i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] != expv2[i]) return false;
    return true;
  }


  bool PPMonoidBigEvImpl::myIsDivisible(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    const BigInt* const expv1 = myExpv(rawpp1);
    const BigInt* const expv2 = myExpv(rawpp2);

    for (long i = 0; i < myNumIndetsValue; ++i)
      if (expv1[i] < expv2[i]) return false;

    return true;
  }


  bool PPMonoidBigEvImpl::myIsRadical(ConstRawPtr rawpp) const
  {
    const BigInt* const expv = myExpv(rawpp);

    for (long i = 0; i < myNumIndetsValue; ++i)
      if (expv[i] > 1) return false;

    return true;
  }


  int PPMonoidBigEvImpl::myCmp(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    return myOrdPtr->myCmpExpvs(myExpv(rawpp1), myExpv(rawpp2));
  }


  long PPMonoidBigEvImpl::myStdDeg(ConstRawPtr rawpp) const
  {
    const BigInt* const expv = myExpv(rawpp);
    BigInt d;
    for (long i=0; i < myNumIndetsValue; ++i)
      d += expv[i];
    long ans;
    if (!IsConvertible(ans, d))
      CoCoA_ERROR(ERR::ArgTooBig, "PPMonoidBigEvImpl::myStdDeg");
    return ans;
  }


  void PPMonoidBigEvImpl::myWDeg(degree& d, ConstRawPtr rawpp) const
  {
    myOrdPtr->myWDeg(d, myExpv(rawpp));
  }


  int PPMonoidBigEvImpl::myCmpWDeg(ConstRawPtr rawpp1, ConstRawPtr rawpp2) const
  {
    return myOrdPtr->myCmpWDeg(myExpv(rawpp1), myExpv(rawpp2));
  }


  int PPMonoidBigEvImpl::myCmpWDegPartial(ConstRawPtr rawpp1, ConstRawPtr rawpp2, long GrDim) const
  {
    return myOrdPtr->myCmpWDegPartial(myExpv(rawpp1), myExpv(rawpp2), GrDim);
  }


  long PPMonoidBigEvImpl::myExponent(ConstRawPtr rawpp, long indet) const
  {
    CoCoA_ASSERT(0 <= indet && indet < myNumIndetsValue);
    long ans;
    if (!IsConvertible(ans, myExpv(rawpp)[indet]))
      CoCoA_ERROR(ERR::ArgTooBig, "PPMonoidBigEvImpl::myExponent");
    return ans;
  }

  void PPMonoidBigEvImpl::myBigExponent(BigInt& EXP, ConstRawPtr rawpp, long indet) const
  {
    CoCoA_ASSERT(0 <= indet && indet < myNumIndetsValue);
    EXP = myExpv(rawpp)[indet];
  }


  void PPMonoidBigEvImpl::myExponents(vector<long>& v, ConstRawPtr rawpp) const
  {
    CoCoA_ASSERT(len(v) == myNumIndetsValue);
    const BigInt* const expv = myExpv(rawpp);
    bool OK = true;
    for (long i=0; i < myNumIndetsValue; ++i)
    {
      if (!IsConvertible(v[i], expv[i]))
      {
        OK = false;
        if (expv[i] > 0)
          v[i] = numeric_limits<long>::max();
        else
          v[i] = numeric_limits<long>::min();
      }
      if (!OK)
        CoCoA_ERROR(ERR::ArgTooBig, "PPMonoidBigEvImpl::myExponents");
    }
  }


  void PPMonoidBigEvImpl::myBigExponents(vector<BigInt>& expv, ConstRawPtr rawpp) const
  {
    CoCoA_ASSERT(len(expv) == myNumIndetsValue);
    const BigInt* const v = myExpv(rawpp);
    for (long i=0; i < myNumIndetsValue; ++i)  expv[i] = v[i];
  }
  

  void PPMonoidBigEvImpl::myComputeDivMask(DivMask& /*dm*/, const DivMaskRule& /*DivMaskImpl*/, ConstRawPtr /*rawpp*/) const
  {
    CoCoA_ERROR(ERR::NYI, "PPMonoidBigEvImpl::myComputeDivMask");
// BUG BUG can't do this (yet)    DivMaskImpl->myAssignFromExpv(dm, myExpv(rawpp), myNumIndetsValue);
  }


  void PPMonoidBigEvImpl::myOutputSelf(std::ostream& out) const
  {
    out << "PPMonoidBigEv(" << myNumIndetsValue << ", " << myOrd <<")";
  }


  void PPMonoidBigEvImpl::myDebugPrint(std::ostream& out, ConstRawPtr rawpp) const
  {
    out << "DEBUG PP: myNumIndetsValue=" << myNumIndetsValue << ", exps=[";
    for (long i=0; i < myNumIndetsValue; ++i)
      out << myExponent(rawpp, i) << " ";
    out << "]" << std::endl;
  }


  PPMonoid NewPPMonoidBigEv(const std::vector<symbol>& IndetNames, const PPOrdering& ord)
  {
    // Sanity check on the indet names given.
    const long nvars = NumIndets(ord);

    if (len(IndetNames) != nvars)
      CoCoA_ERROR(ERR::BadNumIndets, "NewPPMonoidBigEv(IndetNames,ord)");
    if (!AreDistinct(IndetNames))
      CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidBigEv(IndetNames,ord)");
    if (!AreArityConsistent(IndetNames))
      CoCoA_ERROR(ERR::BadIndetNames, "NewPPMonoidBigEv(IndetNames,ord)");

    return PPMonoid(new PPMonoidBigEvImpl(IndetNames, ord));
  }

  PPMonoid NewPPMonoidBigEv(const std::vector<symbol>& IndetNames, const PPOrderingCtor& ord)
  {
    return NewPPMonoidBigEv(IndetNames, ord.myCtor(len(IndetNames)));
  }


//--- orderings ----------------------------------------

  PPMonoidBigEvImpl::CmpBase::CmpBase(long NumIndets, long GradingDim):
    myNumIndetsValue(NumIndets),
    myGradingDimValue(GradingDim)
  {
    CoCoA_ASSERT(NumIndets > 0);
    CoCoA_ASSERT(NumIndets < 1000000); // complain about ridiculously large number of indets
    CoCoA_ASSERT(0 <= GradingDim && GradingDim <= NumIndets);
  }


//--- LexImpl

  PPMonoidBigEvImpl::LexImpl::LexImpl(long NumIndets):
    CmpBase(NumIndets, 0)
  {
  }


  int PPMonoidBigEvImpl::LexImpl::myCmpExpvs(const BigInt* v1, const BigInt* v2) const
  {
    for (long i=0; i<myNumIndetsValue; ++i)
      if (v1[i] != v2[i]) return (v1[i]>v2[i] ? 1 : -1 );
    return 0;
  }


  void PPMonoidBigEvImpl::LexImpl::myWDeg(degree& /*d*/, const BigInt* /*v*/) const
  {
    // deliberately does nothing because GradingDim=0
    //??? should it assign:  d = []
  }


  int PPMonoidBigEvImpl::LexImpl::myCmpWDegPartial(const BigInt* /*v1*/, const BigInt* /*v2*/, long /*GrDim*/) const
  {
    return 0; // GradingDim=0, and all degrees in Z^0 are equal
  }

//--- StdDegLexImpl

  PPMonoidBigEvImpl::StdDegLexImpl::StdDegLexImpl(long NumIndets):
    CmpBase(NumIndets, 1)
  {
  }


  void PPMonoidBigEvImpl::StdDegLexImpl::myWDeg(degree& d, const BigInt* v) const
  {
    BigInt deg;
    for (long i=0; i < myNumIndetsValue; ++i)
      deg += v[i];
    SetComponent(d, 0, deg);
  }


  int PPMonoidBigEvImpl::StdDegLexImpl::myCmpExpvs(const BigInt* v1, const BigInt* v2) const
  {
    BigInt deg1;
    BigInt deg2;
    for (long i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2) return (deg1>deg2 ? 1 : -1 );

    for (long i=0; i < myNumIndetsValue; ++i)
      if (v1[i] != v2[i]) return (v1[i]>v2[i] ? 1 : -1 );
    return 0;
  }


  int PPMonoidBigEvImpl::StdDegLexImpl::myCmpWDegPartial(const BigInt* v1, const BigInt* v2, long GrDim) const
  {
    CoCoA_ASSERT(0 <= GrDim && GrDim <= 1);
    if (GrDim == 0) return 0; // ie: wrt to 0-grading
    BigInt deg1;
    BigInt deg2;
    for (long i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2) return (deg1>deg2 ? 1 : -1 );
    return 0;
  }


//--- StdDegRevLexImpl

  PPMonoidBigEvImpl::StdDegRevLexImpl::StdDegRevLexImpl(long NumIndets):
    CmpBase(NumIndets, 1)
  {
  }


  void PPMonoidBigEvImpl::StdDegRevLexImpl::myWDeg(degree& d, const BigInt* v) const
  {
    BigInt deg;
    for (long i=0; i < myNumIndetsValue; ++i)
      deg += v[i];
    SetComponent(d, 0, deg);
  }


  int PPMonoidBigEvImpl::StdDegRevLexImpl::myCmpExpvs(const BigInt* v1, const BigInt* v2) const
  {
    BigInt deg1;
    BigInt deg2;
    for (long i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2) return (deg1>deg2 ? 1 : -1 );

    for (long i = myNumIndetsValue-1; i>0; --i)
      if (v1[i] != v2[i]) return (v1[i]>v2[i] ? -1 : 1 );
    return 0;
  }


  int PPMonoidBigEvImpl::StdDegRevLexImpl::myCmpWDegPartial(const BigInt* v1, const BigInt* v2, long GrDim) const
  {
    CoCoA_ASSERT(0 <= GrDim && GrDim <= 1);
    if (GrDim == 0) return 0; // ie: wrt to 0-grading
    BigInt deg1;
    BigInt deg2;
    for (long i=0; i < myNumIndetsValue; ++i)
    {
      deg1 += v1[i];
      deg2 += v2[i];
    }
    if (deg1 != deg2) return (deg1>deg2 ? 1 : -1 );
    return 0;
  }


//--- MatrixOrderingImpl

  PPMonoidBigEvImpl::MatrixOrderingImpl::MatrixOrderingImpl(long NumIndets, long GradingDim, const matrix& OrderMatrix):
    CmpBase(NumIndets, GradingDim)
  {
    CoCoA_ASSERT(0 <= NumIndets);
    CoCoA_ASSERT(0 <= GradingDim && GradingDim <= NumIndets);
    CoCoA_ASSERT(NumRows(OrderMatrix) == NumIndets);
    CoCoA_ASSERT(NumCols(OrderMatrix) == NumIndets);

    myOrderMatrix.resize(NumIndets, vector<BigInt>(NumIndets));
    for (long i=0; i < NumIndets; ++i)
      for (long j=0; j < NumIndets; ++j)
        IsInteger(myOrderMatrix[i][j], OrderMatrix(i,j));
  }


  void PPMonoidBigEvImpl::MatrixOrderingImpl::myWDeg(degree& d, const BigInt* v) const
  {
    BigInt deg;
    for (long g=0; g < myGradingDimValue; ++g)
    {
      deg = 0;
      for (long i=0; i < myNumIndetsValue; ++i)
        deg += v[i] * myOrderMatrix[g][i];
      SetComponent(d, g, deg);
    }
  }


  static int MatrixOrderingCmpArrays(const vector<vector<BigInt> >& M, const BigInt* v1, const BigInt* v2, long UpTo)
  {
    CoCoA_ASSERT(0 <= UpTo && UpTo <= len(M));
    BigInt s1, s2; // automatically set to 0
    const long ncols = len(M[0]);
    for (long r=0; r < UpTo; ++r)
    {
      for (long i=0; i < ncols; ++i)
      {
        s1 += v1[i]*M[r][i];
        s2 += v2[i]*M[r][i];
      }
      if (s1 != s2) return (s1>s2 ? 1 : -1 );
      s1 = 0;
      s2 = 0;
    }
    return 0;
  }


  int PPMonoidBigEvImpl::MatrixOrderingImpl::myCmpExpvs(const BigInt* v1, const BigInt* v2) const
  {
    return MatrixOrderingCmpArrays(myOrderMatrix, v1, v2, myNumIndetsValue);
  }


  int PPMonoidBigEvImpl::MatrixOrderingImpl::myCmpWDegPartial(const BigInt* v1, const BigInt* v2, long GrDim) const
  {
    CoCoA_ASSERT(0 <= GrDim && GrDim <= myGradingDimValue);
    return MatrixOrderingCmpArrays(myOrderMatrix, v1, v2, GrDim);
  }


} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/PPMonoidEvZZ.C,v 1.29 2012/02/10 17:05:06 abbott Exp $
// $Log: PPMonoidEvZZ.C,v $
// Revision 1.29  2012/02/10 17:05:06  abbott
// Added myNew pseudo-ctor from vector<BigInt>
//
// Revision 1.28  2012/01/26 16:50:55  bigatti
// -- changed back_inserter into insert
//
// Revision 1.27  2011/08/14 15:52:17  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.26  2011/03/22 15:27:53  bigatti
// -- added myIsIndetPosPower
// -- fixed myIsIndet
//
// Revision 1.25  2011/03/16 13:24:35  abbott
// Ensured that 3rd arg in myCmpWDegPartial is called GrDim; added some comments,
// and corrected arg sanity check for GrDim.
//
// Revision 1.24  2011/03/11 21:49:49  abbott
// Removed two pointless consts (to keep g++-4.3.2 quiet)
//
// Revision 1.23  2011/03/10 17:56:23  bigatti
// -- fixed another CoCoA_ASSERT
//
// Revision 1.22  2011/03/10 17:48:01  bigatti
// -- fixed some CoCoA_ASSERT
//
// Revision 1.21  2011/03/10 17:27:55  bigatti
// -- fixed checks in myCmpWDegPartial
//
// Revision 1.20  2011/03/10 16:39:34  abbott
// Replaced (very many) size_t by long in function interfaces (for rings,
// PPMonoids and modules).  Also replaced most size_t inside fn defns.
//
// Revision 1.19  2010/12/17 16:09:51  abbott
// Corrected used of myIndetSymbols in some assertions.
//
// Revision 1.18  2010/11/30 11:18:11  bigatti
// -- renamed IndetName --> IndetSymbol
//
// Revision 1.17  2010/11/05 17:37:40  bigatti
// -- added second const" to myExpv
//
// Revision 1.16  2010/11/05 17:24:29  bigatti
// -- fixed variable in assert
//
// Revision 1.15  2010/11/05 16:21:08  bigatti
// -- added ZZExponents
//
// Revision 1.14  2010/10/06 14:10:24  abbott
// Added increments to the ref count in ring and PPMonoid ctors to make
// them exception safe.
//
// Revision 1.13  2010/02/03 16:13:52  abbott
// Added new single word tags for specifying the ordering in PPMonoid
// pseudo-ctors.
//
// Revision 1.12  2010/02/02 16:44:31  abbott
// Added radical & IsRadical (via mem fns myRadical & myIsRadical)
// for PPMonoidElems.
//
// Revision 1.11  2009/12/11 11:46:32  abbott
// Changed fn  convert  into  IsConvertible.
// Added template procedure  convert.
// New version because change is not backward compatible.
//
// Revision 1.10  2009/09/22 14:01:33  bigatti
// -- added myCmpWDegPartial (ugly name, I know....)
// -- cleaned up and realigned code in PPMonoid*.C files
//
// Revision 1.9  2008/04/21 12:58:17  abbott
// Commented out some formal parameters (so compiler doesn't give annoying warnings).
//
// Revision 1.8  2007/12/05 11:06:24  bigatti
// -- changed "size_t StdDeg/myStdDeg(f)" into "long"  (and related functions)
// -- changed "log/myLog(f, i)" into "MaxExponent/myMaxExponent(f, i)"
// -- fixed bug in "IsOne(ideal)" in SparsePolyRing.C
//
// Revision 1.7  2007/12/04 14:27:06  bigatti
// -- changed "log(pp, i)" into "exponent(pp, i)"
//
// Revision 1.6  2007/10/30 17:14:07  abbott
// Changed licence from GPL-2 only to GPL-3 or later.
// New version for such an important change.
//
// Revision 1.5  2007/09/24 14:22:12  abbott
// Added big exponent myMulIndetPower.
//
// Revision 1.4  2007/05/31 14:54:31  bigatti
// -- now using AreDistinct and AreArityConsistent for sanity check on
//    indet names
//
// Revision 1.2  2007/05/21 12:42:48  abbott
// Added missing NYI error message.
//
// Revision 1.1  2007/05/03 10:35:23  abbott
// Added new PPMonoidEvZZ with (virtually) unlimited exponents.
// Modified test-PPMonoid1.C accordingly.
// Added warning in doc about silent/unchecked exponent overflow in other
// PPMonoids.
//
