//   Copyright (c)  2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RandomLongStream.H"
#include "CoCoA/error.H"
#include "CoCoA/utils.H"

#include <iostream>
using std::ostream;
#include <limits>
using std::numeric_limits; // only for myGenValue


namespace CoCoA
{

  RandomLongStream::RandomLongStream(const MachineInt& lwb, const MachineInt& upb, const MachineInt& seed):
      myLwb(AsSignedLong(lwb)),
      myUpb(AsSignedLong(upb)),
      myRange((1ul+myUpb) - myLwb), // all arith is between ulongs
      myValue(0), // initialized "properly" in the body
      myCounter(0),
      mySeed(abs(seed)),
      myState() // init in body
  {
    if (myLwb > myUpb)
      CoCoA_ERROR(ERR::BadArg, "RandomLongStream: must have lwb <= upb");
    gmp_randinit_mt(myState);
    gmp_randseed_ui(myState, mySeed);
    myGenValue();
  }

  RandomLongStream::RandomLongStream(const RandomLongStream& copy):
      myLwb(copy.myLwb),
      myUpb(copy.myUpb),
      myRange(copy.myRange),
      myValue(copy.myValue),
      myCounter(copy.myCounter),
      mySeed(copy.mySeed),
      myState() //init in body
  {
    gmp_randinit_set(myState, copy.myState);
  }

  RandomLongStream& RandomLongStream::operator=(const RandomLongStream& rhs)
  {
    if (this == &rhs) return *this; // this is necessary (because of call to gmp_randclear)
    myLwb = rhs.myLwb;
    myUpb = rhs.myUpb;
    myRange = rhs.myRange;
    myValue = rhs.myValue;
    myCounter = rhs.myCounter;
    mySeed = rhs.mySeed;
    gmp_randclear(myState);                 // Together these effect an assignment
    gmp_randinit_set(myState, rhs.myState); //

    return *this;
  }

  RandomLongStream::~RandomLongStream()
  {
    gmp_randclear(myState);
  }


  // Better define this in case some idiot wants to use it...
  // Not inline, so it is as slow as possible >-}
  RandomLongStream RandomLongStream::operator++(int)
  {
    RandomLongStream ans(*this); // Probably not cheap!
    operator++();
    return ans; // Makes another copy upon returning!
  }


  long RandomLongStream::myIndex() const
  {
    return myCounter;
  }


  void RandomLongStream::myGenValue()
  {
    if (myRange > 0)
      myValue = ULong2Long(myLwb + gmp_urandomm_ui(myState, myRange)); // myLwb is silently converted to ulong, so sum is modulo MaxUlong+1
    else
      // Get here only if myLwb = min_long & myUpb = max_long
      myValue = ULong2Long(gmp_urandomb_ui(myState, numeric_limits<unsigned long>::digits));
  }


  ostream& operator<<(ostream& out, const RandomLongStream& RLS)
  {
    out << "RandomLongStream(lwb=" << RLS.myLwb
        << ", upb=" << RLS.myUpb
        << ", seed=" << RLS.mySeed
        << ", counter=" << RLS.myCounter << ")";
    return out;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RandomLongStream.C,v 1.5 2011/11/09 14:09:53 bigatti Exp $
// $Log: RandomLongStream.C,v $
// Revision 1.5  2011/11/09 14:09:53  bigatti
// -- renamed MachineInteger --> MachineInt
//
// Revision 1.4  2011/05/03 09:46:42  abbott
// Changed default seed to 0 (used to be 1).
// Seeding with 0 no longer means "seed from system time".
//
// Revision 1.3  2011/03/14 10:26:21  abbott
// Changed size_t into long.
//
// Revision 1.2  2010/06/29 15:13:16  abbott
// RandomLongStream now allows lwb and upb to be equal.
//
// Revision 1.1  2010/02/16 10:19:29  abbott
// Added new class RandomLongStream with example and test.
//
//
