//   Copyright (c)  2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RandomSource.H"
#include "CoCoA/BigInt.H"
#include "CoCoA/GlobalManager.H" // for GlobalRandomSource()
#include "CoCoA/error.H"
#include "CoCoA/MachineInt.H"
#include "CoCoA/utils.H"

#include <iostream>
//using std::ostream;
#include <limits>
using std::numeric_limits;


namespace CoCoA
{

  RandomSource::RandomSource(const MachineInt& seed):
      mySeed(seed)
  {
    gmp_randinit_mt(myState);
    gmp_randseed(myState, mpzref(mySeed));
  }


  RandomSource::RandomSource(const BigInt& seed):
      mySeed(seed)
  {
    gmp_randinit_mt(myState);
    gmp_randseed(myState, mpzref(mySeed));
  }


  RandomSource::RandomSource(const RandomSource& copy):
      mySeed(copy.mySeed)
  {
    gmp_randinit_set(myState, copy.myState);
  }


  RandomSource& RandomSource::operator=(const RandomSource& rhs)
  {
    gmp_randclear(myState);                 // Together these effect an assignment
    gmp_randinit_set(myState, rhs.myState); //
    mySeed = rhs.mySeed;
    return *this;
  }


  RandomSource::~RandomSource()
  {
    gmp_randclear(myState);
  }


  // ANNA: this is commented out to check whether it is really needed
  // it would be called only by RandomLong(RndSrc, lwb, upb)
//   long RandomLong(RandomSource& RndSrc)
//   {
//     const unsigned long rnd = gmp_urandomb_ui(RndSrc.myState, numeric_limits<unsigned long>::digits);
//     return ULong2Long(rnd);
//   }


  long RandomLong(RandomSource& RndSrc, const MachineInt& lwb, const MachineInt& upb)
  {
    if (!IsSignedLong(lwb) || !IsSignedLong(upb))
      CoCoA_ERROR(ERR::BadArg, "RandomLong: lwb & upb must each fit into a long");
    const long lo = AsSignedLong(lwb);
    const long hi = AsSignedLong(upb);
    if (lo > hi) CoCoA_ERROR(ERR::BadArg, "RandomLong: must have lwb <= upb");
    const unsigned long range = (1ul + hi) - lo; // all arith is between ulongs.
    if (range == 0) //return RandomLong(RndSrc);
      return ULong2Long(gmp_urandomb_ui(RndSrc.myState, numeric_limits<unsigned long>::digits));
    return ULong2Long(lo + gmp_urandomm_ui(RndSrc.myState, range)); // lo is silently converted to ulong
  }


  BigInt RandomBigInt(RandomSource& RndSrc, const MachineInt& lwb, const MachineInt& upb)
  {
    return RandomBigInt(RndSrc, BigInt(lwb), BigInt(upb));
  }

  BigInt RandomBigInt(RandomSource& RndSrc, const MachineInt& lwb, const BigInt& upb)
  {
    return RandomBigInt(RndSrc, BigInt(lwb), upb);
  }

  BigInt RandomBigInt(RandomSource& RndSrc, const BigInt& lwb, const MachineInt& upb)
  {
    return RandomBigInt(RndSrc, lwb, BigInt(upb));
  }

  BigInt RandomBigInt(RandomSource& RndSrc, const BigInt& lwb, const BigInt& upb)
  {
    if (lwb > upb) CoCoA_ERROR(ERR::BadArg, "RandomBigInt: must have lwb <= upb");
    BigInt sample;
    mpz_urandomm(mpzref(sample), RndSrc.myState, mpzref(upb-lwb+1));
    return lwb + sample;
  }

// NOTE (2011-04-26): this is significantly faster than the fn above.
//   BigInt RandomBigIntmod(RandomSource& RndSrc, const BigInt& M)
//   {
//     if (M <= 0) CoCoA_ERROR(ERR::BadArg, "RandomBigIntmod: must have M > 0");
//     BigInt sample;
//     mpz_urandomm(mpzref(sample), RndSrc.myState, mpzref(M));
//     return sample;
//   }


  void reseed(RandomSource& RndSrc, const MachineInt& seed)
  {
    reseed(RndSrc, BigInt(seed));
  }

  void reseed(RandomSource& RndSrc, const BigInt& seed)
  {
    RndSrc.mySeed = seed;
    gmp_randseed(RndSrc.myState, mpzref(RndSrc.mySeed));
    // zero any counters
  }


  std::ostream& operator<<(std::ostream& out, const RandomSource& RndSrc)
  {
    out << "RandomSource(seed=" << RndSrc.mySeed << ", state=unprintable)";
    return out;
  }


  bool RandomBool()
  { return RandomBool(GlobalRandomSource()); }

  long RandomLong(const MachineInt& lwb, const MachineInt& upb)
  { return RandomLong(GlobalRandomSource(), lwb, upb); }

  BigInt RandomBigInt(const MachineInt& lwb, const MachineInt& upb)
  { return RandomBigInt(GlobalRandomSource(), lwb, upb); }

  BigInt RandomBigInt(const MachineInt& lwb, const BigInt& upb)
  { return RandomBigInt(GlobalRandomSource(), lwb, upb); }

  BigInt RandomBigInt(const BigInt& lwb, const MachineInt& upb)
  { return RandomBigInt(GlobalRandomSource(), lwb, upb); }

  BigInt RandomBigInt(const BigInt& lwb, const BigInt& upb)
  { return RandomBigInt(GlobalRandomSource(), lwb, upb); }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RandomSource.C,v 1.7 2012/02/03 10:18:46 bigatti Exp $
// $Log: RandomSource.C,v $
// Revision 1.7  2012/02/03 10:18:46  bigatti
// -- added random functions with implicit use of GlobalRandomSource()
//
// Revision 1.6  2012/02/02 15:58:22  abbott
// Added includes of BigInt & MachineInt.
// Changed formal parameter names.
//
// Revision 1.5  2011/11/09 14:09:53  bigatti
// -- renamed MachineInteger --> MachineInt
//
// Revision 1.4  2011/08/14 15:52:17  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.3  2011/05/03 09:52:50  abbott
// Changed default initial seed to 0 (was previsouly 1).
// Added reseeding functions.
// Seeding from a ZZ is now possible.
//
// Revision 1.2  2011/03/14 10:26:21  abbott
// Changed size_t into long.
//
// Revision 1.1  2010/12/17 16:03:15  abbott
// Added new class RandomSource.
//
//
