//   Copyright (c)  2011  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "CoCoA/decimal.H"

#include "CoCoA/BigRat.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"


#include <algorithm>
using std::fill;
using std::copy;
#include <iostream>
using std::ostream;
#include <limits>
using std::numeric_limits;
//#include <string>
using std::string;
#include <sstream>
using std::ostringstream;

namespace CoCoA
{

  const int MantExp::ourDefaultSigFig = 5; // initialize in .C

  namespace // anonymous for file local fns
  {

    std::string FloatStr(const MantExp& ME, long ndigits)
    {
      string mantissa; mantissa.reserve(ndigits);
      if (!IsZero(ME.myMantissa))
        IsConvertible(mantissa, ME.myMantissa); // cannot fail, except std::bad_alloc
      else
        // fill string with ndigits of zeroes...
        for (long i=0;i<ndigits;++i)mantissa+='0';///???fill(...);

      const int ExpDigits = numeric_limits<long>::digits10;
      string ans; ans.reserve(9+ndigits+ExpDigits); // 9 = size of "-.*10^(-)"

      if (ME.mySign < 0) ans += '-';
      ans += mantissa[0];
      ans += '.';
      //      copy(&mantissa[1], &mantissa[ndigits], back_inserter(ans));
      ans.insert(ans.end(), &mantissa[1], &mantissa[ndigits]);
      ans += "*10^";
      if (ME.myExponent < 0) ans += '(';
      ostringstream exp;
      exp << ME.myExponent;
      ans += exp.str();
      if (ME.myExponent < 0) ans += ')';
      return ans;
    }

    std::string DecimalStr(const MantExp& ME, long ndigits)
    {
      // See doc for info about magic number 8 in line below
      if (ME.myExponent >= ndigits || ME.myExponent < -8) return FloatStr(ME, ndigits);

      string mantissa; mantissa.reserve(ndigits);
      if (!IsZero(ME.myMantissa))
        IsConvertible(mantissa, ME.myMantissa); // cannot fail, except std::bad_alloc
      else
        // fill string with ndigits of zeroes...
        for (long i=0;i<ndigits;++i)mantissa+='0';///???fill(...);

      string ans; ans.reserve(3+ndigits); // 3 = size of "-0."
      if (ME.mySign < 0) ans += '-';
      if (ME.myExponent < 0)
      {
        ans += "0.";
        for (long i=-1; i > ME.myExponent; --i)
          ans += '0';
        //        copy(&mantissa[0], &mantissa[ndigits], back_inserter(ans));
        ans.insert(ans.end(), &mantissa[0], &mantissa[ndigits]);
        return ans;
      }

      //   copy(&mantissa[0], &mantissa[ME.myExponent+1], back_inserter(ans));
      ans.insert(ans.end(), &mantissa[0], &mantissa[ME.myExponent+1]);
      ans += '.';
      // copy(&mantissa[ME.myExponent+1], &mantissa[ndigits], back_inserter(ans));
      ans.insert(ans.end(), &mantissa[ME.myExponent+1], &mantissa[ndigits]);
      return ans;
    }

  }


  std::ostream& operator<<(std::ostream& out, const MantExp& ME)
  {
    out << "MantExp(s=" << ME.mySign << ", e=" << ME.myExponent << ", m=" << ME.myMantissa << ")";
    return out;
  }


  MantExp MantissaAndExponent(const BigInt& N, const MachineInt& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "MantissaAndExponent");
    if (IsZero(N)) return MantExp(0,0,BigInt(0));
    const long ndigits = AsSignedLong(SigFig);
    const int SignN = sign(N);
    const long e = ILogBase(N,10); // overflow???
    if (e < ndigits)
      return MantExp(SignN, e, abs(N)*power(10, ndigits-e-1));
    const BigInt HalfULP = 5*power(10, e-ndigits);
    const BigInt digits = (1+abs(N)/HalfULP)/2;
    // Must check whether digits has overflowed...
    if (abs(digits) == power(10, ndigits))
      return MantExp(SignN, e+1, digits/10);
    return MantExp(SignN, e, digits);
  }


  MantExp MantissaAndExponent(const BigRat& q, const MachineInt& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "MantissaAndExponent");
    if (IsZero(q)) return MantExp(0,0,BigInt(0));
    if (IsOneDen(q)) return MantissaAndExponent(num(q), SigFig);
    const long ndigits = AsSignedLong(SigFig);
    const int signq = sign(q);
    const long e = ILogBase(q,10); // overflow???
    BigInt digits;
    if (e < ndigits)
      digits = round(abs(q)*power(10,ndigits-e-1));
    else
      digits = round(abs(q)/power(10,1+e-ndigits)); 
    // Must check whether digits has overflowed...
    if (abs(digits) == power(10, ndigits))
      return MantExp(signq, e+1, digits/10);
    return MantExp(signq, e, digits);
  }


  std::string FloatStr(const BigInt& N, const MachineInt& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "FloatStr");
    return FloatStr(MantissaAndExponent(N, SigFig), AsSignedLong(SigFig));
  }

  std::string FloatStr(const BigRat& q, const MachineInt& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "FloatStr");
    return FloatStr(MantissaAndExponent(q, SigFig), AsSignedLong(SigFig));
  }

  std::string DecimalStr(const BigInt& N, const MachineInt& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "DecimalStr");
    if (abs(N) < power(10,SigFig)) return ConvertTo<string>(N);

    return DecimalStr(MantissaAndExponent(N, SigFig), AsSignedLong(SigFig));
  }

  std::string DecimalStr(const BigRat& q, const MachineInt& SigFig)
  {
    if (IsNegative(SigFig) || !IsSignedLong(SigFig) || IsZero(SigFig))
      CoCoA_ERROR(ERR::BadArg, "DecimalStr");
    if (IsOneDen(q)) return DecimalStr(num(q), SigFig);
    return DecimalStr(MantissaAndExponent(q, SigFig), AsSignedLong(SigFig));
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/decimal.C,v 1.9 2012/02/08 17:07:33 bigatti Exp $
// $Log: decimal.C,v $
// Revision 1.9  2012/02/08 17:07:33  bigatti
// -- definition of global in .C file
//
// Revision 1.8  2012/01/26 16:57:42  abbott
// Moved value of ourDefaultSigFig from H file to C file
// (this allows clean compilation on MSVC, also avoids
// excessive recompilation if we change the value).
//
// Revision 1.7  2012/01/26 16:47:00  bigatti
// -- changed back_inserter into insert
//
// Revision 1.6  2011/11/09 14:29:37  bigatti
// -- renamed MachineInteger --> MachineInt
//
// Revision 1.5  2011/08/24 10:32:04  bigatti
// -- renamed QQ --> BigRat
// -- sorted #include
//
// Revision 1.4  2011/08/14 15:52:16  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.3  2011/08/03 09:05:05  abbott
// Added new fn DecimalStr.
//
// Revision 1.2  2011/08/02 13:28:46  abbott
// Cleaned up impl.  Now MantissaAndExponent has default SigFig too.
// Used naming convention for members of MantExp.
//
// Revision 1.1  2011/03/01 15:21:22  abbott
// Added new functions: MantissaAndExponent, FloatStr (in CoCoALib).
//
//
