//   Copyright (c) 2010 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include <iostream>
#include <sstream>
#include <exception>
#include <cassert>
#include <vector>
#include <cstdlib>
#include <memory>
#include <boost/scope_exit.hpp>
#include <boost/foreach.hpp>
#include <boost/algorithm/string/case_conv.hpp>
#include <boost/iostreams/device/file.hpp>
#include <boost/iostreams/filtering_stream.hpp>
#include <boost/iostreams/filter/newline.hpp>
#if 0
#include <zip.h>
#endif

#include "CoCoA/library.H"
#include "CoCoALibSupplement.H"
#include "Interpreter.H"
#include "OnlineHelp.H"

using namespace std;
using namespace boost;
using namespace boost::iostreams;
using namespace CoCoA::AST;
using namespace CoCoA::LexerNS;
using namespace CoCoA::ParserNS;

namespace CoCoA {
namespace InterpreterNS {

template <> BigInt theValue<BigIntValue>(boost::intrusive_ptr<BigIntValue> x) { return x->theInteger; }
template <> const BigInt theValue<const BigIntValue>(boost::intrusive_ptr<const BigIntValue> x) { return x->theInteger; }
template <> BigRat theValue<BigRatValue>(boost::intrusive_ptr<BigRatValue> x) { return x->theRational; }
template <> const BigRat theValue<const BigRatValue>(boost::intrusive_ptr<const BigRatValue> x) { return x->theRational; }
template <> RingElem theValue<RingElemValue>(boost::intrusive_ptr<RingElemValue> x) { return x->theRingElem; }
template <> const RingElem theValue<const RingElemValue>(boost::intrusive_ptr<const RingElemValue> x) { return x->theRingElem; }
template <> ring theValue<RingValue>(boost::intrusive_ptr<RingValue> x) { return x->theRing; }
template <> const ring theValue<const RingValue>(boost::intrusive_ptr<const RingValue> x) { return x->theRing; }
template <> RingHom theValue<RingHomValue>(boost::intrusive_ptr<RingHomValue> x) { return x->theRingHom; }
template <> const RingHom theValue<const RingHomValue>(boost::intrusive_ptr<const RingHomValue> x) { return x->theRingHom; }
template <> matrix theValue<MatrixValue>(boost::intrusive_ptr<MatrixValue> x) { return x->theMatrix; }
template <> const matrix theValue<const MatrixValue>(boost::intrusive_ptr<const MatrixValue> x) { return x->theMatrix; }
template <> string theValue<StringValue>(boost::intrusive_ptr<StringValue> x) { return x->theString; }
template <> const string theValue<const StringValue>(boost::intrusive_ptr<const StringValue> x) { return x->theString; }
template <> string theValue<TypeValue>(boost::intrusive_ptr<TypeValue> x) { return x->name; }
template <> const string theValue<const TypeValue>(boost::intrusive_ptr<const TypeValue> x) { return x->name; }
template <> bool theValue<BoolValue>(boost::intrusive_ptr<BoolValue> x) { return x->theBool; }
template <> bool theValue<const BoolValue>(boost::intrusive_ptr<const BoolValue> x) { return x->theBool; }
template <> ideal theValue<IdealValue>(boost::intrusive_ptr<IdealValue> x) { return x->theIdeal; }
template <> const ideal theValue<const IdealValue>(boost::intrusive_ptr<const IdealValue> x) { return x->theIdeal; }

namespace {

typedef pair<string, intrusive_ptr<BuiltInFunction> > NameFunPair;
vector<NameFunPair> builtIns;

struct AddBuiltIn {
	AddBuiltIn(const string &fname, intrusive_ptr<BuiltInFunction> fun) {
		builtIns.push_back(make_pair(fname, fun));
	}
};

#define ARG(N) invocationExpression->args[N]

#define DECLARE_ARITYCHECK_FUNCTION(FNAME) bool builtin_check_arity_##FNAME(int nArg)
#define DECLARE_BUILTIN_FUNCTION(FNAME) \
		intrusive_ptr<RightValue> builtin_##FNAME(const intrusive_ptr<const InvocationExpression> invocationExpression, RuntimeEnvironment * const runtimeEnv); \
		DECLARE_ARITYCHECK_FUNCTION(FNAME) ;\
		AddBuiltIn add_##FNAME(#FNAME, new BuiltInFunction(builtin_##FNAME, builtin_check_arity_##FNAME)); \
		intrusive_ptr<RightValue> builtin_##FNAME(const intrusive_ptr<const InvocationExpression> invocationExpression, RuntimeEnvironment * const runtimeEnv)
#define DECLARE_STD_BUILTIN_FUNCTION(FNAME, NARG) \
	DECLARE_ARITYCHECK_FUNCTION(FNAME) { return nArg==(NARG); } \
	DECLARE_BUILTIN_FUNCTION(FNAME) { \
	invocationExpression->checkNumberOfArgs(NARG);
#define END_STD_BUILTIN_FUNCTION }

#define DECLARE_COCOALIB_FUNCTION0(FUN) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 0) \
	(void)runtimeEnv; return Value::from(FUN()); }

#define DECLARE_COCOALIB_FUNCTION1(FUN, ARG1T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 1) \
	return Value::from(FUN(theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))))); }

#define DECLARE_COCOALIB_FUNCTION2(FUN, ARG1T, ARG2T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 2) \
	return Value::from(FUN( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))), \
			theValue(runtimeEnv->evalArgAs<ARG2T>(ARG(1)))  \
			)); }

#define DECLARE_COCOALIB_FUNCTION3(FUN, ARG1T, ARG2T, ARG3T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 3) \
	return Value::from(FUN( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))), \
			theValue(runtimeEnv->evalArgAs<ARG2T>(ARG(1))), \
			theValue(runtimeEnv->evalArgAs<ARG3T>(ARG(2)))  \
			)); }

#define DECLARE_COCOALIB_FUNCTION4(FUN, ARG1T, ARG2T, ARG3T, ARG4T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 4) \
	return Value::from(FUN( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))), \
			theValue(runtimeEnv->evalArgAs<ARG2T>(ARG(1))), \
			theValue(runtimeEnv->evalArgAs<ARG3T>(ARG(2))), \
			theValue(runtimeEnv->evalArgAs<ARG4T>(ARG(3)))  \
			)); }

#define DECLARE_COCOALIB_FUNCTION5(FUN, ARG1T, ARG2T, ARG3T, ARG4T, ARG5T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 5) \
	return Value::from(FUN( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))), \
			theValue(runtimeEnv->evalArgAs<ARG2T>(ARG(1))), \
			theValue(runtimeEnv->evalArgAs<ARG3T>(ARG(2))), \
			theValue(runtimeEnv->evalArgAs<ARG4T>(ARG(3))), \
			theValue(runtimeEnv->evalArgAs<ARG5T>(ARG(4)))  \
			)); }

#define DECLARE_COCOALIBFORC5_FUNCTION0(FUN) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 0) \
	(void)runtimeEnv; return Value::from(FUN##_forC5()); }

#define DECLARE_COCOALIBFORC5_FUNCTION1(FUN, ARG1T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 1) \
	return Value::from(FUN##_forC5( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))) \
			)); }

#define DECLARE_COCOALIBFORC5_FUNCTION2(FUN, ARG1T, ARG2T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 2) \
	return Value::from(FUN##_forC5( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))), \
			theValue(runtimeEnv->evalArgAs<ARG2T>(ARG(1))) \
			)); }

#define DECLARE_COCOALIBFORC5_FUNCTION3(FUN, ARG1T, ARG2T, ARG3T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 3) \
	return Value::from(FUN##_forC5( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))), \
			theValue(runtimeEnv->evalArgAs<ARG2T>(ARG(1))), \
			theValue(runtimeEnv->evalArgAs<ARG3T>(ARG(2)))  \
			)); }

#define DECLARE_COCOALIBFORC5_FUNCTION4(FUN, ARG1T, ARG2T, ARG3T, ARG4T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 4) \
	return Value::from(FUN##_forC5( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))), \
			theValue(runtimeEnv->evalArgAs<ARG2T>(ARG(1))), \
			theValue(runtimeEnv->evalArgAs<ARG3T>(ARG(2))), \
			theValue(runtimeEnv->evalArgAs<ARG4T>(ARG(3)))  \
			)); }

#define DECLARE_COCOALIBFORC5_FUNCTION5(FUN, ARG1T, ARG2T, ARG3T, ARG4T, ARG5T) \
	DECLARE_STD_BUILTIN_FUNCTION(FUN, 5) \
	return Value::from(FUN##_forC5( \
			theValue(runtimeEnv->evalArgAs<ARG1T>(ARG(0))), \
			theValue(runtimeEnv->evalArgAs<ARG2T>(ARG(1))), \
			theValue(runtimeEnv->evalArgAs<ARG3T>(ARG(2))), \
			theValue(runtimeEnv->evalArgAs<ARG4T>(ARG(3))), \
			theValue(runtimeEnv->evalArgAs<ARG5T>(ARG(4)))  \
			)); }

// functions on BigInt
DECLARE_COCOALIB_FUNCTION1(factorial, BigIntValue)
DECLARE_COCOALIB_FUNCTION1(fibonacci, BigIntValue)
DECLARE_COCOALIB_FUNCTION1(isqrt, BigIntValue) // AMB
DECLARE_COCOALIB_FUNCTION1(NextProbPrime, BigIntValue) // AMB
DECLARE_COCOALIB_FUNCTION2(binomial, BigIntValue, BigIntValue)
DECLARE_COCOALIB_FUNCTION2(iroot, BigIntValue, BigIntValue) // AMB
DECLARE_COCOALIB_FUNCTION3(PowerMod, BigIntValue, BigIntValue, BigIntValue)
DECLARE_COCOALIB_FUNCTION1(IsEven, BigIntValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsOdd, BigIntValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsPrime, BigIntValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsProbPrime, BigIntValue) // AMB
DECLARE_COCOALIBFORC5_FUNCTION1(NextPrime, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION2(random, BigIntValue, BigIntValue)
// functions on BigRat
DECLARE_COCOALIB_FUNCTION2(CFApprox, BigRatValue, BigRatValue) // AMB
// functions on RingElem
DECLARE_COCOALIB_FUNCTION1(ClearDenom, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION1(CommonDenom, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsConstant, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsIndet, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsIndetPosPower, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsInvertible, RingElemValue)
DECLARE_COCOALIB_FUNCTION1(IsMonomial, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION1(LC, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION1(characteristic, RingValue)
DECLARE_COCOALIB_FUNCTION1(content, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION1(monic, RingElemValue) // AMB //if LC(f)==1???
DECLARE_COCOALIB_FUNCTION2(IsDivisible, RingElemValue, RingElemValue)
DECLARE_COCOALIB_FUNCTION2(deriv, RingElemValue, RingElemValue) // AMB
DECLARE_COCOALIB_FUNCTION2(power, RingElemValue, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION1(IsTerm, RingElemValue)
DECLARE_COCOALIBFORC5_FUNCTION1(LM, RingElemValue)
DECLARE_COCOALIBFORC5_FUNCTION1(LPP, RingElemValue)
DECLARE_COCOALIBFORC5_FUNCTION1(monomials, RingElemValue)
DECLARE_COCOALIBFORC5_FUNCTION1(support, RingElemValue)
DECLARE_COCOALIBFORC5_FUNCTION1(wdeg, RingElemValue)
DECLARE_COCOALIBFORC5_FUNCTION2(CoeffOfTerm, RingElemValue, RingElemValue)
DECLARE_COCOALIBFORC5_FUNCTION2(DensePoly, RingValue, BigIntValue) // AMB
// functions on ring/PolyRing
DECLARE_COCOALIB_FUNCTION0(RingQQ)
DECLARE_COCOALIB_FUNCTION0(RingZZ)
DECLARE_COCOALIB_FUNCTION1(IsCommutative, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsField, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsFractionField, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsGCDDomain, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsIntegralDomain, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsOrderedDomain, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsPolyRing, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsQQ, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsQuotientRing, RingValue)
DECLARE_COCOALIB_FUNCTION1(IsZZ, RingValue)
DECLARE_COCOALIB_FUNCTION1(NewFractionField, RingValue)  // AMB
DECLARE_COCOALIB_FUNCTION1(NewRingFp, BigIntValue)  // AMB
DECLARE_COCOALIB_FUNCTION1(one, RingValue)
DECLARE_COCOALIB_FUNCTION1(owner, RingElemValue)
DECLARE_COCOALIB_FUNCTION1(zero, RingValue)
//DECLARE_COCOALIB_FUNCTION1(NewWeylAlgebra, ..)  // AMB
DECLARE_COCOALIB_FUNCTION2(NewQuotientRing, RingValue, IdealValue)  // AMB
DECLARE_COCOALIBFORC5_FUNCTION1(GradingDim, RingValue) // AMB
DECLARE_COCOALIBFORC5_FUNCTION1(WeightsMatrix, RingValue) // AMB
DECLARE_COCOALIBFORC5_FUNCTION1(ord, RingValue) // AMB

// functions for homomorphisms
DECLARE_COCOALIB_FUNCTION1(IdentityHom, RingValue) // AMB
DECLARE_COCOALIB_FUNCTION1(codomain, RingHomValue) // AMB
DECLARE_COCOALIB_FUNCTION1(domain, RingHomValue) // AMB
DECLARE_COCOALIB_FUNCTION2(CanonicalHom, RingValue, RingValue) // GL
DECLARE_COCOALIB_FUNCTION2(TmpChainCanonicalHom, RingValue, RingValue) // AMB

// functions on matrices
//DECLARE_COCOALIB_FUNCTION1(FrobeniusNorm2, MatrixValue) // AMB
//DECLARE_COCOALIB_FUNCTION1(OperatorNorm1, MatrixValue) // AMB
//DECLARE_COCOALIB_FUNCTION1(OperatorNormInfinity, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsAntiSymmetric, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsDiagonal, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsSymmetric, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsTermOrdering, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(NumCols, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(NumRows, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(PseudoInverse, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(adjoint, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(det, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(inverse, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION1(rank, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION2(ConcatAntiDiag, MatrixValue, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION2(ConcatDiag, MatrixValue, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION2(ConcatHor, MatrixValue, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION2(ConcatVer, MatrixValue, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION2(TensorMat, MatrixValue, MatrixValue) // AMB
DECLARE_COCOALIB_FUNCTION4(BlockMat, MatrixValue, MatrixValue, MatrixValue, MatrixValue) // AMB
DECLARE_COCOALIBFORC5_FUNCTION1(ExtractOrdMat, MatrixValue)
DECLARE_COCOALIBFORC5_FUNCTION1(LexMat, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION1(RevLexMat, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION1(StdDegLexMat, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION1(StdDegRevLexMat, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION1(XelMat, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION1(transposed, MatrixValue)
DECLARE_COCOALIBFORC5_FUNCTION2(GetCol, MatrixValue, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION2(GetRow, MatrixValue, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION2(IdentityMat, RingValue, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION3(NewDenseMat, RingValue, BigIntValue, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION3(ZeroMat, RingValue, BigIntValue, BigIntValue)
DECLARE_COCOALIBFORC5_FUNCTION4(FilledMat, RingValue, BigIntValue, BigIntValue, RingElemValue)

// functions on ideals
DECLARE_COCOALIB_FUNCTION1(AlexanderDual, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION1(AmbientRing, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION1(AreGensMonomial, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION1(GBasis, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION1(IsZeroDim, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION1(PrimaryDecomposition, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION1(gens, IdealValue)
DECLARE_COCOALIB_FUNCTION2(IsContained, IdealValue, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION2(IsElem, RingElemValue, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION2(NF, RingElemValue, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION2(colon, IdealValue, IdealValue) // AMB
DECLARE_COCOALIB_FUNCTION2(intersect, IdealValue, IdealValue) // AMB
DECLARE_COCOALIBFORC5_FUNCTION1(HilbertNumQuot, IdealValue) // AMB

DECLARE_COCOALIBFORC5_FUNCTION0(CpuTime)
DECLARE_COCOALIBFORC5_FUNCTION0(date)

DECLARE_STD_BUILTIN_FUNCTION(IsInteger, 2) {
	intrusive_ptr<LeftValue> resultRef = intrusive_ptr_cast<LeftValue>(runtimeEnv->evalArg(ARG(0), RuntimeEnvironment::EVAL_BY_REF));
	intrusive_ptr<RingElemValue> poly = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
	BigInt N;
	bool isInteger(IsInteger(N, poly->theRingElem));
	if (isInteger) {
		if (resultRef->assignmentNeedsOwnership())
			resultRef->obtainOwnership();
		const intrusive_ptr<const Expression> resultRefExp = ARG(0).exp;
		resultRef->assign(new BigIntValue(N), resultRefExp->getBegin(), resultRefExp->getEnd(), runtimeEnv);
	}
	return BoolValue::fromBool(isInteger);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsRational, 2) {
	intrusive_ptr<LeftValue> resultRef = intrusive_ptr_cast<LeftValue>(runtimeEnv->evalArg(ARG(0), RuntimeEnvironment::EVAL_BY_REF));
	intrusive_ptr<RingElemValue> poly = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
	BigRat qq;
	bool isRational(IsRational(qq, poly->theRingElem));
	if (isRational) {
		if (resultRef->assignmentNeedsOwnership())
			resultRef->obtainOwnership();
		const intrusive_ptr<const Expression> resultRefExp = ARG(0).exp;
		resultRef->assign(new BigRatValue(qq), resultRefExp->getBegin(), resultRefExp->getEnd(), runtimeEnv);
	}
	return BoolValue::fromBool(isRational);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(aliases, 0) {
	runtimeEnv->topLevelAliases->dump(runtimeEnv->getOutputStream());
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(pause, 0) {
#ifdef C5IDE
	runtimeEnv->interpreter->singleStepExecution = true;
#else // #ifdef C5IDE
	(void)runtimeEnv; // keeps the compiler happy (or, at least, silent ;-) )
#endif // #ifdef C5IDE
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

#ifdef C5IDE
DECLARE_STD_BUILTIN_FUNCTION(sleep, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(arg);
	long l;
	if (!IsConvertible(l, N->theInteger) || l<0 || l>1000)
		throw RuntimeException("The argument must be in the range [0..1000]", arg.exp);
	this_thread::sleep(boost::posix_time::seconds(static_cast<int>(l)));
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(debug, 1) {
	const Argument &arg = ARG(0);
	string filename = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	try {
		runtimeEnv->interpreter->singleStepExecution = true;
		runtimeEnv->interpreter->readAndExecute(filename, false, false);
	} catch (const RuntimeException &) {
		throw;
	} catch (const BaseException &be) {
		throw RuntimeException(be.reason, invocationExpression->getBegin(), invocationExpression->getEnd());
	}
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION
#endif // #ifdef C5IDE

#if 0  // 2 functions for accessing zip files, currently disabled
DECLARE_STD_BUILTIN_FUNCTION(ZipFileList, 1) {
	const Argument &arg = ARG(0);
	const string filename = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	int err;
	struct zip *zf = ::zip_open(filename.c_str(), ZIP_CHECKCONS, &err);
	if (!zf)
		throw RuntimeException("Cannot read Zip file \""+filename+"\"", arg.exp);
	intrusive_ptr<ListValue> result(new ListValue);
	const int numEntries = zip_get_num_files(zf);
	for(int a=0; a<numEntries; ++a)
		result->addValue(new StringValue(zip_get_name(zf, a, ZIP_FL_UNCHANGED)));
	::zip_close(zf);
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ZipRead, 2) {
	const Argument &argFilename = ARG(0);
	const Argument &argEntryName = ARG(1);
	const string filename = runtimeEnv->evalArgAs<StringValue>(argFilename)->theString;
	const string entryName = runtimeEnv->evalArgAs<StringValue>(argEntryName)->theString;
	int err;
	struct zip *zf = ::zip_open(filename.c_str(), ZIP_CHECKCONS, &err);
	if (!zf)
		throw RuntimeException("Cannot read Zip file \""+filename+"\"", argFilename.exp);
	const int index = zip_name_locate(zf, entryName.c_str(), /* flags */ 0);
	if (index<0) {
		::zip_close(zf);
		throw RuntimeException("Cannot find an entry named \""+entryName+" inside the Zip file \""+filename+"\"", argEntryName.exp);
	}
	struct zip_file *entry = ::zip_fopen_index(zf, index, /* flags */ 0);
	if (!entry) {
		::zip_close(zf);
		throw RuntimeException("Error reading the entry named \""+entryName+" inside the Zip file \""+filename+"\"", argEntryName.exp);
	}
	string result;
	for(;;) {
		char buf[1024];
		int nRead = ::zip_fread(entry, buf, sizeof(buf));
		if (nRead<=0) {
			::zip_fclose(entry);
			::zip_close(zf);
			return new StringValue(result);
		}
		result.append(buf, buf+nRead);
	}
}
END_STD_BUILTIN_FUNCTION
#endif

DECLARE_STD_BUILTIN_FUNCTION(len, 1) {
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (intrusive_ptr<StringValue> s = dynamic_pointer_cast<StringValue>(v))
		return Value::from(BigInt(s->theString.length()));
	if (intrusive_ptr<ListValue> list = dynamic_pointer_cast<ListValue>(v))
		return Value::from(BigInt(list->size()));
	if (intrusive_ptr<RingElemValue> f = dynamic_pointer_cast<RingElemValue>(v))
		return Value::from(BigInt(NumTerms(f->theRingElem)));
	if (dynamic_pointer_cast<MatrixValue>(v))
    throw RuntimeException("len(Matrix) not allowed, use NumRows(matrix) instead", ARG(0).exp);
	if (dynamic_pointer_cast<MatrixRowValue>(v))
    throw RuntimeException("len(MatrixRow) not allowed, use NumCols(matrix) instead", ARG(0).exp);
	throw WrongTypeException(ListValue::type->name + " or " +
                           RingElemValue::type->name + " or " +
                           StringValue::type->name,
                           v->getType()->name, ARG(0).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(sum) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(sum) {
	invocationExpression->checkNumberOfArgs(1,2);
	const Argument &a0 = ARG(0);
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(a0);
	ListValue::ContainerType::size_type size = list->size();
  intrusive_ptr<RightValue> result = BigIntValue::zero;
  if (invocationExpression->args.size()==1)
  {
    if (size==0) return result;
    result = list->getValue(--size);
  }
  else
    result = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  for(ListValue::ContainerType::size_type a=size; a!=0; /**/)
    try	{
			result = runtimeEnv->binaryOperatorDispatch(list->getValue(--a), result, RuntimeEnvironment::opPlusMap, a0.exp->getBegin(), a0.exp->getEnd());
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the sum", a0.exp);
	}
	return result;
}

DECLARE_ARITYCHECK_FUNCTION(product) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(product) {
	invocationExpression->checkNumberOfArgs(1,2);
	const Argument &a0 = ARG(0);
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(a0);
	ListValue::ContainerType::size_type size = list->size();
	intrusive_ptr<RightValue> result = BigIntValue::one;
  if (invocationExpression->args.size()==1)
  {
    if (size==0) return result;
    result = list->getValue(--size);
  }
  else
    result = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  for(ListValue::ContainerType::size_type a=size; a!=0; /**/)
		try	{
			result = runtimeEnv->binaryOperatorDispatch(list->getValue(--a), result, RuntimeEnvironment::opStarMap, a0.exp->getBegin(), a0.exp->getEnd());
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the product", a0.exp);
	}
	return result;
}

DECLARE_ARITYCHECK_FUNCTION(NewList) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(NewList) {
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
  long n;
  if (!IsConvertible(n, N->theInteger) || n<0)
     throw RuntimeException("invalid length", ARG(0).exp);
	intrusive_ptr<ListValue> returnValue = new ListValue();
  intrusive_ptr<RightValue> val = BigIntValue::zero;
  if (invocationExpression->args.size()==2)
    val = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  for(long i=0; i<n; ++i)  returnValue->addValue(val);
	return returnValue;
}

DECLARE_ARITYCHECK_FUNCTION(first) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(first) {
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(ARG(0));
  if (invocationExpression->args.size()==1)  return list->getValue(0);
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger) || n<0)
    throw RuntimeException("invalid length", ARG(1).exp);
	intrusive_ptr<ListValue> returnValue = new ListValue();
	for(long i=0; i<n; ++i)  returnValue->addValue(list->getValue(i));
	return returnValue;
}

DECLARE_ARITYCHECK_FUNCTION(last) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(last) {
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(ARG(0));
  ///////// here
  ListValue::ContainerType::size_type s = list->size();
  if (invocationExpression->args.size()==1) return list->getValue(s-1);
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger) || n<0)
    throw RuntimeException("invalid length", ARG(1).exp);
  if ((unsigned long)n>s)
    throw RuntimeException("value greater than list length", ARG(1).exp);
	intrusive_ptr<ListValue> returnValue = new ListValue();
	for(unsigned long i=s-n; i<s; ++i)  returnValue->addValue(list->getValue(i));
	return returnValue;
}

DECLARE_STD_BUILTIN_FUNCTION(ScalarProduct, 2) {
	const Argument &arg1 = ARG(0);
	intrusive_ptr<ListValue> l1 = runtimeEnv->evalArgAs<ListValue>(arg1);
	const Argument &arg2 = ARG(1);
	intrusive_ptr<ListValue> l2 = runtimeEnv->evalArgAs<ListValue>(arg2);
	const ListValue::ContainerType::size_type size = l1->size();
	const CharPointer &begin = arg1.exp->getBegin();
	const CharPointer &end = arg2.exp->getEnd();
	if (l2->size()!=size)
		throw RuntimeException("The arguments must be lists of the same size", begin, end);
	intrusive_ptr<RightValue> result = BigIntValue::zero;
	try	{
		for(ListValue::ContainerType::size_type a=0; a<size; ++a)
			result = runtimeEnv->binaryOperatorDispatch(
						result,
						runtimeEnv->binaryOperatorDispatch(l1->getValue(a), l2->getValue(a), RuntimeEnvironment::opStarMap, begin, end),
						RuntimeEnvironment::opPlusMap,
						begin,
						end);
	} catch (const InterruptException &) {
		throw;
	} catch (const RuntimeException &) {
		throw RuntimeException("Some elements have incompatible types for the scalar product", begin, end);
	}
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(concat) { (void)nArg; return true; }

DECLARE_BUILTIN_FUNCTION(concat) {
	intrusive_ptr<ListValue> returnValue = new ListValue();
	BOOST_FOREACH(const Argument &arg, invocationExpression->args) {
		intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(arg);
		int size = list->size();
		for(int a=0; a<size; ++a)
			returnValue->addValue(list->getValue(a));
	}
	return returnValue;
}

DECLARE_STD_BUILTIN_FUNCTION(reversed, 1) {
	intrusive_ptr<ListValue> list = runtimeEnv->evalArgAs<ListValue>(ARG(0));
	intrusive_ptr<ListValue> returnValue = new ListValue();
	for(ListValue::ContainerType::size_type a=list->size(); a!=0; /**/)
		returnValue->addValue(list->getValue(--a));
	return returnValue;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(reverse, 1) {
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnshared<ListValue>(ARG(0));
	const ListValue::ContainerType::size_type size = list->size();
	if (size) {
		ListValue::ContainerType::size_type left = 0, right = size-1;
		while(left < right) {
			const intrusive_ptr<RightValue> tmp(list->getValue(left));
			list->setValue(left++, list->getValue(right));
			list->setValue(right--, tmp);
		}
	}
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(deg) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(deg) {  // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<RingElemValue> a = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  if (invocationExpression->args.size()==1)
    return Value::from(deg(a->theRingElem));
	intrusive_ptr<RingElemValue> b = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
  long i;
  if (!IsIndet(i,b->theRingElem))
    throw RuntimeException("not an indet", ARG(1).exp);
  return Value::from(MaxExponent(a->theRingElem, i));
}

DECLARE_ARITYCHECK_FUNCTION(coefficients) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(coefficients) {  // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<RingElemValue> a = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  if (invocationExpression->args.size()==1)
    return Value::from(coefficients_forC5(a->theRingElem));
	vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(1));
  vector<RingElem> res;
  for (long i=0; i<len(v); ++i)
    res.push_back(CoeffOfTerm_forC5(a->theRingElem, v[i]));
  return Value::from(res);
}

DECLARE_ARITYCHECK_FUNCTION(NewMat) { return (2<=nArg) && (nArg<=3); }
DECLARE_BUILTIN_FUNCTION(NewMat) {
	invocationExpression->checkNumberOfArgs(2,3);
  if (invocationExpression->args.size()==2)
    throw RuntimeException("NewMat(R,C) not allowed, use NewMat(Ring,R,C) or NewMat(R,C,Entry) instead", ARG(0).exp);
  long nrows, ncols;
	intrusive_ptr<BigIntValue> nrows_ptr, ncols_ptr;
	intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	intrusive_ptr<RightValue> v1 = runtimeEnv->evalArgAs<RightValue>(ARG(1));
	intrusive_ptr<RightValue> v2 = runtimeEnv->evalArgAs<RightValue>(ARG(2));
	if (const intrusive_ptr<const RingValue> R = dynamic_pointer_cast<const RingValue>(v0))
  {
    if ( !(nrows_ptr = dynamic_pointer_cast<BigIntValue>(v1)) 
         || !IsConvertible(nrows, nrows_ptr->theInteger)
         || nrows<0)
      throw RuntimeException("invalid number of rows", ARG(1).exp);
    if ( !(ncols_ptr = dynamic_pointer_cast<BigIntValue>(v2)) 
         || !IsConvertible(ncols, ncols_ptr->theInteger)
         || ncols<0)
      throw RuntimeException("invalid number of columns", ARG(2).exp);
    return Value::from(ZeroMat(R->theRing, nrows, ncols));
  }
    if ( !(nrows_ptr = dynamic_pointer_cast<BigIntValue>(v0)) 
         || !IsConvertible(nrows, nrows_ptr->theInteger)
         || nrows<0)
      throw RuntimeException("invalid number of rows", ARG(0).exp);
    if ( !(ncols_ptr = dynamic_pointer_cast<BigIntValue>(v1)) 
         || !IsConvertible(ncols, ncols_ptr->theInteger)
         || ncols<0)
      throw RuntimeException("invalid number of columns", ARG(1).exp);
	if (intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(v2))
    return Value::from(FilledMat(owner(x->theRingElem), nrows, ncols, x->theRingElem));
	if (intrusive_ptr<BigIntValue> x = dynamic_pointer_cast<BigIntValue>(v2))
    return Value::from(FilledMat(RingQQ(), nrows, ncols, RingElem(RingQQ(), x->theInteger)));
	if (intrusive_ptr<BigRatValue> x = dynamic_pointer_cast<BigRatValue>(v2))
    return Value::from(FilledMat(RingQQ(), nrows, ncols, RingElem(RingQQ(), x->theRational)));
	throw WrongTypeException(RingElemValue::type->name + " or " + 
                           BigIntValue::type->name + " or " + 
                           BigRatValue::type->name, v2->getType()->name, ARG(2).exp);
}

DECLARE_ARITYCHECK_FUNCTION(jacobian) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(jacobian) {  // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	vector<RingElem> v0 = runtimeEnv->evalArgAsRingElemList(ARG(0));
  if (invocationExpression->args.size()==1)
  {
    if (v0.size()==0) throw RuntimeException("Empty list", ARG(0).exp);
    return new MatrixValue(jacobian(v0, indets(AsPolyRing(owner(v0[0])))));
  }
	vector<RingElem> v1 = runtimeEnv->evalArgAsRingElemList(ARG(1));
  return new MatrixValue(jacobian(v0, v1));
}


DECLARE_ARITYCHECK_FUNCTION(CompleteToOrd) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(CompleteToOrd) {  // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<const MatrixValue> M0 = runtimeEnv->evalArgAs<const MatrixValue>(ARG(0));
  if (invocationExpression->args.size()==1)
    return Value::from(NewMatCompleteOrd(M0->theMatrix));
	intrusive_ptr<const MatrixValue> M1 = runtimeEnv->evalArgAs<const MatrixValue>(ARG(1));
  return Value::from(NewMatCompleteOrd(M0->theMatrix, M1->theMatrix));
}


DECLARE_STD_BUILTIN_FUNCTION(HomogElimMat, 2) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	vector<BigInt> v1 = runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1));
  return Value::from(HomogElimMat_forC5(M->theMatrix, v1));
}
END_STD_BUILTIN_FUNCTION


// DECLARE_ARITYCHECK_FUNCTION(GrammSchmidtRows) { return (1<=nArg) && (nArg<=2); }
// DECLARE_BUILTIN_FUNCTION(GrammSchmidtRows) {  // AMB
// 	invocationExpression->checkNumberOfArgs(1,2);
// 	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
//   if (invocationExpression->args.size()==1)
//   {
//     GrammSchmidtRows(M->theMatrix);
//     return VoidValue::theInstance;
//   }
// 	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
//   long n;
//   if (!IsConvertible(n, N->theInteger))
//     throw RuntimeException("invalid row index", ARG(1).exp);
//   GrammSchmidtRows(M->theMatrix, n);
//   return VoidValue::theInstance;
// }


DECLARE_STD_BUILTIN_FUNCTION(append, 2) {
	intrusive_ptr<RightValue> elem = runtimeEnv->evalArgAs<RightValue>(ARG(1));
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnshared<ListValue>(ARG(0));
	list->addValue(elem);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(remove, 2) {
	const Argument indexArg = ARG(1);
	intrusive_ptr<RightValue> shouldBeAnIndex = runtimeEnv->evalArgAs<RightValue>(indexArg);
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnshared<ListValue>(ARG(0));
	ListValue::ContainerType::size_type index = list->checkIndex(shouldBeAnIndex, indexArg.exp);
	list->removeValue(index);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(insert, 3) {
	const Argument &indexArg = ARG(1);
	intrusive_ptr<RightValue> shouldBeAnIndex = runtimeEnv->evalArgAs<RightValue>(indexArg);
	intrusive_ptr<RightValue> elem = runtimeEnv->evalArgAs<RightValue>(ARG(2));
	intrusive_ptr<ListValue> list = runtimeEnv->obtainUnshared<ListValue>(ARG(0));
	ListValue::ContainerType::size_type index = list->checkIndex(shouldBeAnIndex, indexArg.exp);
	list->insertValue(index, elem);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(div, 2) {
	intrusive_ptr<BigIntValue> a = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> b = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	BigInt result;
	div(result, a->theInteger, b->theInteger);
	return new BigIntValue(result);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(mod, 2) {
	intrusive_ptr<BigIntValue> a = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> b = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	BigInt result;
	mod(result, a->theInteger, b->theInteger);
	return new BigIntValue(result);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NumDigits, 2) {
	intrusive_ptr<BigIntValue> a = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> b = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	long base;
	if (!IsConvertible(base, b->theInteger) || base<2 || base>36)
		throw RuntimeException("Base must be in the range 2..36", ARG(1).exp);
	return new BigIntValue(NumDigits(a->theInteger, base));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(sign, 1) {
	intrusive_ptr<const RightValue> v(runtimeEnv->evalArgAs<RightValue>(ARG(0)));
	if (const intrusive_ptr<const RingElemValue> poly = dynamic_pointer_cast<const RingElemValue>(v))
		return BigIntValue::fromInt(sign(poly->theRingElem));
	if (const intrusive_ptr<const BigIntValue> N = dynamic_pointer_cast<const BigIntValue>(v))
		return BigIntValue::fromInt(sign(N->theInteger));
	if (const intrusive_ptr<const BigRatValue> qq = dynamic_pointer_cast<const BigRatValue>(v))
		return BigIntValue::fromInt(sign(qq->theRational));
	throw WrongTypeException(BigIntValue::type->name+", "+BigRatValue::type->name+" or "+RingElemValue::type->name, v->getType()->name, ARG(0).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsZero, 1) {
	intrusive_ptr<const RightValue> v(runtimeEnv->evalArgAs<RightValue>(ARG(0)));
	if (const intrusive_ptr<const RingElemValue> poly = dynamic_pointer_cast<const RingElemValue>(v))
		return BoolValue::fromBool(IsZero(poly->theRingElem));
	if (const intrusive_ptr<const BigIntValue> N = dynamic_pointer_cast<const BigIntValue>(v))
		return BoolValue::fromBool(IsZero(N->theInteger));
	if (const intrusive_ptr<const BigRatValue> qq = dynamic_pointer_cast<const BigRatValue>(v))
		return BoolValue::fromBool(IsZero(qq->theRational));
	if (const intrusive_ptr<const IdealValue> I = dynamic_pointer_cast<const IdealValue>(v))
		return BoolValue::fromBool(IsZero(I->theIdeal));
	throw WrongTypeException(BigIntValue::type->name+", "+
                           BigRatValue::type->name+" or "+
                           RingElemValue::type->name+" or "+
                           IdealValue::type->name,
                           v->getType()->name, ARG(0).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsOne, 1) {
	intrusive_ptr<const RightValue> v(runtimeEnv->evalArgAs<RightValue>(ARG(0)));
	if (const intrusive_ptr<const RingElemValue> poly = dynamic_pointer_cast<const RingElemValue>(v))
		return BoolValue::fromBool(IsOne(poly->theRingElem));
	if (const intrusive_ptr<const BigIntValue> N = dynamic_pointer_cast<const BigIntValue>(v))
		return BoolValue::fromBool(IsOne(N->theInteger));
	if (const intrusive_ptr<const BigRatValue> qq = dynamic_pointer_cast<const BigRatValue>(v))
		return BoolValue::fromBool(IsOne(qq->theRational));
	throw WrongTypeException(BigIntValue::type->name+", "+BigRatValue::type->name+" or "+RingElemValue::type->name, v->getType()->name, ARG(0).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsMinusOne, 1) {
	intrusive_ptr<const RightValue> v(runtimeEnv->evalArgAs<RightValue>(ARG(0)));
	if (const intrusive_ptr<const RingElemValue> poly = dynamic_pointer_cast<const RingElemValue>(v))
		return BoolValue::fromBool(IsMinusOne(poly->theRingElem));
	if (const intrusive_ptr<const BigIntValue> N = dynamic_pointer_cast<const BigIntValue>(v))
		return BoolValue::fromBool(IsMinusOne(N->theInteger));
	if (const intrusive_ptr<const BigRatValue> qq = dynamic_pointer_cast<const BigRatValue>(v))
		return BoolValue::fromBool(IsMinusOne(qq->theRational));
	throw WrongTypeException(BigIntValue::type->name+", "+BigRatValue::type->name+" or "+RingElemValue::type->name, v->getType()->name, ARG(0).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NumTerms, 1) { // AMB
	return Value::from(BigInt(NumTerms(runtimeEnv->evalArgAs<RingElemValue>(ARG(0))->theRingElem)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(RingElem, 2) {
	intrusive_ptr<const RingValue> R(runtimeEnv->evalArgAs<RingValue>(ARG(0)));
	intrusive_ptr<const RightValue> v(runtimeEnv->evalArgAs<RightValue>(ARG(1)));
	if (const intrusive_ptr<const BigIntValue> N = dynamic_pointer_cast<const BigIntValue>(v))
    return new RingElemValue(RingElem(R->theRing, N->theInteger));
	if (const intrusive_ptr<const BigRatValue> qq = dynamic_pointer_cast<const BigRatValue>(v))
    return new RingElemValue(RingElem(R->theRing, qq->theRational));
	if (const intrusive_ptr<const StringValue> s = dynamic_pointer_cast<const StringValue>(v))
    return new RingElemValue(RingElem(R->theRing, symbol(s->theString)));
	if (const intrusive_ptr<const ListValue> l = dynamic_pointer_cast<const ListValue>(v))
  {
    ListValue::ContainerType::size_type size = l->size();
    string str;
		if (const boost::intrusive_ptr<StringValue> s = boost::dynamic_pointer_cast<StringValue>(l->getValue(0)))
			str = s->theString;
		else
			throw RuntimeException("String expected", ARG(1).exp);
    std::vector<BigInt> indices;
    for(ListValue::ContainerType::size_type a=1; a<size; ++a) {
      if (const boost::intrusive_ptr<BigIntValue> elem = boost::dynamic_pointer_cast<BigIntValue>(l->getValue(a)))
        indices.push_back(theValue(elem));
      else
        throw RuntimeException("Integer expected", ARG(1).exp);
    }
    return new RingElemValue(RingElem(R->theRing, symbol(str, VectorLong(indices, "RingElem"))));
  }
// 	if (const intrusive_ptr<const RingElemValue> x = dynamic_pointer_cast<const RingElemValue>(v))
//     return new RingElemValue(RingElem(R->theRing, x->theRingElem));
	throw WrongTypeException(BigIntValue::type->name+", "+BigRatValue::type->name+", "+StringValue::type->name+" or "+ListValue::type->name, v->getType()->name, ARG(1).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(homog, 2) {
	intrusive_ptr<const RightValue> v(runtimeEnv->evalArgAs<RightValue>(ARG(0)));
	intrusive_ptr<const RingElemValue> h(runtimeEnv->evalArgAs<RingElemValue>(ARG(1)));
	if (const intrusive_ptr<const RingElemValue> f = dynamic_pointer_cast<const RingElemValue>(v))
    return Value::from(homog(f->theRingElem, h->theRingElem));
	if (const intrusive_ptr<const IdealValue> i = dynamic_pointer_cast<const IdealValue>(v))
    throw RuntimeException("not yet implemented for ideal", ARG(0).exp);
	if (dynamic_pointer_cast<const ListValue>(v))
  {
    vector<RingElem> v1 = runtimeEnv->evalArgAsRingElemList(ARG(0));
    intrusive_ptr<ListValue> returnValue = new ListValue();
    for(long a=0; a<len(v1); ++a)
      returnValue->addValue(Value::from(homog(v1[a], h->theRingElem)));
    return returnValue;
  }
	throw WrongTypeException(RingElemValue::type->name+", "+IdealValue::type->name+" or "+ListValue::type->name, v->getType()->name, ARG(0).exp);
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(abs, 1) {
	bool isBigInt;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(ARG(0), isBigInt);
	if (isBigInt)
		return new BigIntValue(abs(intrusive_ptr_cast<BigIntValue>(v)->theInteger));
	return new BigRatValue(abs(intrusive_ptr_cast<BigRatValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(floor, 1) {
	bool isBigInt;
  intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(ARG(0), isBigInt);
	if (isBigInt)
		return v;
	return new BigIntValue(floor(intrusive_ptr_cast<BigRatValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ceil, 1) {
	bool isBigInt;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(ARG(0), isBigInt);
	if (isBigInt)
		return v;
	return new BigIntValue(ceil(intrusive_ptr_cast<BigRatValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(round, 1) {
	bool isBigInt;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(ARG(0), isBigInt);
	if (isBigInt)
		return v;
	return new BigIntValue(round(intrusive_ptr_cast<BigRatValue>(v)->theRational));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(fields, 1) {
	intrusive_ptr<RecordValue> r = runtimeEnv->evalArgAs<RecordValue>(ARG(0));
	return r->fieldNames();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ConcatList, 1) {
	intrusive_ptr<ListValue> listOfList = runtimeEnv->evalArgAs<ListValue>(ARG(0));
	intrusive_ptr<ListValue> result(new ListValue);
	const ListValue::ContainerType::size_type nLists = listOfList->size();
	for(ListValue::ContainerType::size_type a=0; a<nLists; ++a) {
		intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(listOfList->getValue(a));
		if (!l)
			throw RuntimeException("The argument is not a list of lists", ARG(0).exp);
		const ListValue::ContainerType::size_type len = l->size();
		for(ListValue::ContainerType::size_type b=0; b<len; ++b)
			result->addValue(l->getValue(b));
	}
	return result;
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(gcd) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(gcd) {
	invocationExpression->checkNumberOfArgs(1,2);
  if (invocationExpression->args.size()==1)
  {
    intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
    if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(v)) {
      if (l->size() == 0) return BigIntValue::zero;
      // if (RingFor(ARG(0))==RingZZ()) ....// try evalArgAsListOf<BigIntValue>
      return new RingElemValue(gcd_forC5(runtimeEnv->evalArgAsRingElemList(ARG(0))));
    }
    if (dynamic_pointer_cast<BigIntValue>(v)) return v;
    if (dynamic_pointer_cast<RingElemValue>(v)) return v;
    throw WrongTypeException(ListValue::type->name + " or " +
                             RingElemValue::type->name + " or " +
                             BigIntValue::type->name,
                             v->getType()->name, ARG(0).exp);
  }
	bool isBigInt;
	intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAsT1orT2<BigIntValue, RingElemValue>(ARG(0), isBigInt);
	if (isBigInt)
  {
    intrusive_ptr<BigIntValue> v1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
		return new BigIntValue(gcd(intrusive_ptr_cast<BigIntValue>(v0)->theInteger, v1->theInteger));
  }
  intrusive_ptr<RingElemValue> v1 = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
  return new RingElemValue(gcd(intrusive_ptr_cast<RingElemValue>(v0)->theRingElem, v1->theRingElem));
}

DECLARE_ARITYCHECK_FUNCTION(lcm) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(lcm) {
	invocationExpression->checkNumberOfArgs(1,2);
  if (invocationExpression->args.size()==1)
  {
    intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
    if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(v)) {
      if (l->size() == 0) return BigIntValue::one;
      // if (RingFor(ARG(0))==RingZZ()) ....// try evalArgAsListOf<BigIntValue>
      return new RingElemValue(lcm_forC5(runtimeEnv->evalArgAsRingElemList(ARG(0))));
    }
    if (dynamic_pointer_cast<BigIntValue>(v)) return v;
    if (dynamic_pointer_cast<RingElemValue>(v)) return v;
    throw WrongTypeException(ListValue::type->name + " or " +
                             RingElemValue::type->name + " or " +
                             BigIntValue::type->name,
                             v->getType()->name, ARG(0).exp);
  }
	bool isBigInt;
	intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAsT1orT2<BigIntValue, RingElemValue>(ARG(0), isBigInt);
	if (isBigInt)
  {
    intrusive_ptr<BigIntValue> v1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
		return new BigIntValue(lcm(intrusive_ptr_cast<BigIntValue>(v0)->theInteger, v1->theInteger));
  }
  intrusive_ptr<RingElemValue> v1 = runtimeEnv->evalArgAs<RingElemValue>(ARG(1));
  return new RingElemValue(lcm(intrusive_ptr_cast<RingElemValue>(v0)->theRingElem, v1->theRingElem));
}

DECLARE_STD_BUILTIN_FUNCTION(num, 1) { // AMB (changed)
	const Argument &arg = ARG(0);
  int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(arg, which);
	if (which==1) return v;
	if (which==2)
    return new BigIntValue(num(intrusive_ptr_cast<BigRatValue>(v)->theRational));
  //	if (which==3)
  return new RingElemValue(num(intrusive_ptr_cast<RingElemValue>(v)->theRingElem));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(den, 1) {
	const Argument &arg = ARG(0);
  int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(arg, which);
	if (which==1) return new BigIntValue(BigInt(1));
	if (which==2)
    return new BigIntValue(den(intrusive_ptr_cast<BigRatValue>(v)->theRational));
  //	if (which==3)
  return new RingElemValue(den(intrusive_ptr_cast<RingElemValue>(v)->theRingElem));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(sprint, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<OutputStringStreamValue> outs(new OutputStringStreamValue);
	outs->print(runtimeEnv, runtimeEnv->evalArgAs<RightValue>(arg), arg.exp);
	return outs->close(runtimeEnv, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GetErrMesg, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<ErrorValue> err = runtimeEnv->evalArgAs<ErrorValue>(arg);
	return new StringValue(err->message);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(close, 1) {
	const Argument &arg = ARG(0);
	return runtimeEnv->evalArgAs<OutputStreamValue>(arg)->close(runtimeEnv, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(OpenOString, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	return new OutputStringStreamValue();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(GetEnv, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<StringValue> varName = runtimeEnv->evalArgAs<StringValue>(arg);
	char *value = ::getenv(varName->theString.c_str());
	if (!value)
		return StringValue::empty;
	return new StringValue(string(value));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(OpenOFile, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<StringValue> filename = runtimeEnv->evalArgAs<StringValue>(arg);
	return new OutputFileStreamValue(filename, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ContentsOfFile, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<StringValue> filename = runtimeEnv->evalArgAs<StringValue>(arg);
	filtering_istream input;
	input.push(newline_filter(newline::posix));
	const file_source fs(filename->theString.c_str());
	if (!fs.is_open())
		throw RuntimeException("Cannot open file \""+filename->theString+"\" for reading.", invocationExpression);
	input.push(fs);
	string wholeFile;
	for(;;) {
		string line;
		getline(input, line);
		if (input.bad())
			throw RuntimeException("Cannot read from \""+filename->theString+"\".", invocationExpression);
		wholeFile += line;
		if (input.eof())
			break;
		wholeFile += '\n';
	}
	return new StringValue(wholeFile);
}
END_STD_BUILTIN_FUNCTION

const string invalidTagError("The tag must be a valid identifier");

DECLARE_STD_BUILTIN_FUNCTION(tagged, 2) {
	const Argument &argValue = ARG(0);
	const Argument &argTag = ARG(1);
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(argValue);
	const string tag = runtimeEnv->evalArgAs<StringValue>(argTag)->theString;
	// TODO: must be fixed
	//if (!LexerNS::Lexer::isValidIdentifier(tag))
	//	throw RuntimeException(invalidTagError, argTag.exp);
	return new TaggedValue(v->untagged(), invocationExpression->packageName+"."+tag);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(untagged, 1) {
	return runtimeEnv->evalArgAs<RightValue>(ARG(0))->untagged();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(tag, 1) {
	const Argument &argValue = ARG(0);
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(argValue);
	if (intrusive_ptr<TaggedValue> taggedValue = dynamic_pointer_cast<TaggedValue>(v)) {
		const string::size_type l = invocationExpression->packageName.length()+1;
		if (taggedValue->tag.substr(0, l)==invocationExpression->packageName+".")
			return new StringValue(taggedValue->tag.substr(l, string::npos));
		return new StringValue(taggedValue->tag);
	}
	return StringValue::empty;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(type, 1) {
	return runtimeEnv->evalArgAs<RightValue>(ARG(0))->getType();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PkgName, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	return new StringValue(invocationExpression->packageName);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(/* uppercase because it's a backward-compatibility hack */ Functions, 1) {
	(void)runtimeEnv; // to avoid a compilation warning
	throw RuntimeException("Use the statement Describe to obtain package information", invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(TAGGED, 1) {
	const Argument &arg = ARG(0);
	const string tag = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	if (!LexerNS::Lexer::isValidIdentifier(tag))
		throw RuntimeException(invalidTagError, arg.exp);
	return TypeValue::tagType(tag);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(types, 0) {
	(void)runtimeEnv; // to avoid a compilation warning
	throw RuntimeException("Cannot return a list of all types because they're infinite (CoCoA 4 lies about this, I don't). You might want to use CurrentTypes() to get a list of currently instantiated types", invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CurrentTypes, 0) {
	return runtimeEnv->currentTypes();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(error, 1) {
	const Argument &arg = ARG(0);
	const string message = runtimeEnv->evalArgAs<StringValue>(arg)->theString;
	throw RuntimeException(message, invocationExpression);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ascii, 1) {
	const Argument &arg = ARG(0);
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(arg);
	if (intrusive_ptr<StringValue> sv = dynamic_pointer_cast<StringValue>(v)) {
		intrusive_ptr<ListValue> l(new ListValue());
		const string s(sv->theString);
		BOOST_FOREACH(const char c, s)
			l->addValue(new BigIntValue(static_cast<int>(c)));
		return l;
	}
	if (intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(v)) {
		long l;
		if (!IsConvertible(l, N->theInteger) || l<0 || l>255)
			throw RuntimeException("When the argument is an integer, it must be in the range [0..255]", arg.exp);
		return new StringValue(string(1, static_cast<char>(l)));
	}
	if (intrusive_ptr<ListValue> list = dynamic_pointer_cast<ListValue>(v)) {
		string retValue;
		ListValue::ContainerType::size_type size = list->size();
		for(ListValue::ContainerType::size_type a=0; a<size; ++a) {
			intrusive_ptr<BigIntValue> N = dynamic_pointer_cast<BigIntValue>(list->getValue(a));
			long l;
			if (!N || !IsConvertible(l, N->theInteger) || l<0 || l>255)
				throw RuntimeException("When the argument is a list, it must be a list of integers only in the range [0..255]", arg.exp);
			retValue += static_cast<char>(l);
		}
		return new StringValue(retValue);
	}
	throw WrongTypeException(ListValue::type->name+", "+StringValue::type->name + " or "+BigIntValue::type->name, v->getType()->name, arg.exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ILogBase, 2) { // AMB
	const Argument &a0 = ARG(0);
  intrusive_ptr<BigIntValue> base = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	bool isBigInt;
	intrusive_ptr<RightValue> v0 = runtimeEnv->evalArgAsT1orT2<BigIntValue, BigRatValue>(a0, isBigInt);
	if (isBigInt)
    return new BigIntValue(ILogBase(runtimeEnv->evalArgAs<BigIntValue>(a0)->theInteger, base->theInteger));
    return new BigIntValue(ILogBase(runtimeEnv->evalArgAs<BigRatValue>(a0)->theRational, base->theInteger));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(log, 1) { // AMB
  intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  vector<BigInt> expv;
  if (NumIndets(AsPolyRing(owner(f->theRingElem)))==1)
    expv.push_back(BigInt(deg(f->theRingElem)));
  else 
    BigExponents(expv, LPP(f->theRingElem));
	return new ListValue(expv);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CoeffRing, 1) { // AMB
	const Argument &arg = ARG(0);
	return new RingValue(CoeffRing(AsPolyRing((runtimeEnv->evalArgAs<RingValue>(arg))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(DefiningIdeal, 1) { // AMB
	const Argument &arg = ARG(0);
	return Value::from(DefiningIdeal(AsQuotientRing((runtimeEnv->evalArgAs<RingValue>(arg))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsHomog, 1) { // AMB
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (const intrusive_ptr<const RingElemValue> f = dynamic_pointer_cast<const RingElemValue>(v))
		return BoolValue::fromBool(IsHomog(f->theRingElem));
	if (const intrusive_ptr<const IdealValue> I = dynamic_pointer_cast<const IdealValue>(v))
    return BoolValue::fromBool(IsHomog(I->theIdeal));
  if (const intrusive_ptr<const ListValue> l = dynamic_pointer_cast<const ListValue>(v))
  {
    ListValue::ContainerType::size_type size = l->size();
    for(ListValue::ContainerType::size_type a=1; a<size; ++a)
    {
      if (const boost::intrusive_ptr<const RingElemValue> f = boost::dynamic_pointer_cast<const RingElemValue>(l->getValue(a)))
        if (!IsHomog(f->theRingElem)) return BoolValue::fromBool(false);
    }
    return BoolValue::fromBool(true);
  }
	throw WrongTypeException(RingValue::type->name + " or " +
                           ListValue::type->name + " or " +
                           IdealValue::type->name,
                           v->getType()->name, ARG(0).exp);
  return BoolValue::fromBool(true); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(LT, 1) { // AMB
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (const intrusive_ptr<const RingElemValue> f = dynamic_pointer_cast<const RingElemValue>(v))
		return Value::from(LT_forC5(f->theRingElem));
	if (const intrusive_ptr<const IdealValue> I = dynamic_pointer_cast<const IdealValue>(v))
    return Value::from(LT(I->theIdeal));
//   if (const intrusive_ptr<const ListValue> l = dynamic_pointer_cast<const ListValue>(v))
//   {
//     ListValue::ContainerType::size_type size = l->size();
//     for(ListValue::ContainerType::size_type a=1; a<size; ++a)
//     {
//       if (const boost::intrusive_ptr<const RingElemValue> f = boost::dynamic_pointer_cast<const RingElemValue>(l->getValue(a)))
//         if (!IsHomog(f->theRingElem)) return BoolValue::fromBool(false);
//     }
//     return BoolValue::fromBool(true);
//   }
	throw WrongTypeException(RingValue::type->name + " or " +
                           ListValue::type->name + " or " +
                           IdealValue::type->name,
                           v->getType()->name, ARG(0).exp);
  return BoolValue::fromBool(true); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(RingOf, 1) { // AMB
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (const intrusive_ptr<RingElemValue> f = dynamic_pointer_cast<RingElemValue>(v))
    return Value::from(owner(f->theRingElem));
	if (const intrusive_ptr<IdealValue> I = dynamic_pointer_cast<IdealValue>(v))
    return Value::from(AmbientRing(I->theIdeal));
	if (const intrusive_ptr<MatrixValue> M = dynamic_pointer_cast<MatrixValue>(v))
    return Value::from(BaseRing(M->theMatrix));
//   if (const intrusive_ptr<RingValue> R = dynamic_pointer_cast<RingValue>(v))
//   {
//     if (IsFractionField(R->theRing))
//       return Value::from(BaseRing(AsFractionField(R->theRing)));
//     if (IsQuotientRing(R->theRing))
//       return Value::from(BaseRing(AsQuotientRing(R->theRing)));
//   }
	throw WrongTypeException(RingElemValue::type->name + " or " +
                           IdealValue::type->name + " or " +
                           MatrixValue::type->name,
                           v->getType()->name, ARG(0).exp);
  return new RingValue(RingZZ()); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(BaseRing, 1) { // AMB
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(0));
	if (const intrusive_ptr<MatrixValue> M = dynamic_pointer_cast<MatrixValue>(v))
    return Value::from(BaseRing(M->theMatrix));
  if (const intrusive_ptr<RingValue> R = dynamic_pointer_cast<RingValue>(v))
  {
    if (IsFractionField(R->theRing))
      return Value::from(BaseRing(AsFractionField(R->theRing)));
    if (IsQuotientRing(R->theRing))
      return Value::from(BaseRing(AsQuotientRing(R->theRing)));
  }
	throw WrongTypeException(MatrixValue::type->name + " or " +
                           RingValue::type->name,
                           v->getType()->name, ARG(0).exp);
  return new RingValue(RingZZ()); // just to keep the compiler quiet
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CoeffEmbeddingHom, 1) { // AMB
  PolyRing P = AsPolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing);
	return new RingHomValue(CoeffEmbeddingHom(P));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(EmbeddingHom, 1) { // AMB
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
	return new RingHomValue(EmbeddingHom(AsFractionField(R->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(InducedHom, 2) { // AMB
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<RingHomValue> phi = runtimeEnv->evalArgAs<RingHomValue>(ARG(1));
  if (IsFractionField(R->theRing))
    return Value::from(InducedHom(AsFractionField(R->theRing),phi->theRingHom));
  if (IsQuotientRing(R->theRing))
    return Value::from(InducedHom(AsQuotientRing(R->theRing), phi->theRingHom));
  throw RuntimeException("FractionField or QuotientRing expected", ARG(0).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PolyAlgebraHom, 3) { // AMB
  intrusive_ptr<RingValue> P = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(1));
  vector<RingElem> IndetImages = runtimeEnv->evalArgAsRingElemList(ARG(2), R->theRing);
	return Value::from(PolyAlgebraHom(AsPolyRing(P->theRing), R->theRing, IndetImages));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(PolyRingHom, 4) { // AMB
  intrusive_ptr<RingValue> P = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(1));
  intrusive_ptr<RingHomValue> CoeffHom = runtimeEnv->evalArgAs<RingHomValue>(ARG(2));
  vector<RingElem> IndetImages = runtimeEnv->evalArgAsRingElemList(ARG(3), R->theRing);
	return new RingHomValue(PolyRingHom(AsPolyRing(P->theRing), R->theRing, CoeffHom->theRingHom, IndetImages));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(apply, 2) {
	intrusive_ptr<RingHomValue> phi = runtimeEnv->evalArgAs<RingHomValue>(ARG(0));
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAs<RightValue>(ARG(1));
	if (intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(v))
		return Value::from(apply(phi->theRingHom, x->theRingElem));
	if (intrusive_ptr<MatrixValue> M = dynamic_pointer_cast<MatrixValue>(v))
		return Value::from(apply(phi->theRingHom, M->theMatrix));
	if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(v))
  {
    vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(1), domain(phi->theRingHom));
		return Value::from(apply(phi->theRingHom, v));
  }
	throw WrongTypeException(RingElemValue::type->name + " or " +
                           ListValue::type->name + " or " +
                           MatrixValue::type->name,
                           v->getType()->name, ARG(1).exp);
}
END_STD_BUILTIN_FUNCTION

// DECLARE_STD_BUILTIN_FUNCTION(ord, 1) { // AMB
// 	return Value::from(ordering(PPM(AsSparsePolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing))));
// }
// END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(indets, 1) { // AMB
	return Value::from(indets(AsPolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NumIndets, 1) { // AMB
	return Value::from(NumIndets(AsPolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(0)))->theRing)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(indet, 2) { // AMB
	intrusive_ptr<RingValue> a = runtimeEnv->evalArgAs<RingValue>(ARG(0));
	intrusive_ptr<BigIntValue> b = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, b->theInteger))
    throw RuntimeException("invalid indet index", ARG(1).exp);
	return Value::from(indet(AsPolyRing(a->theRing), n-1));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IndetIndex, 1) { // AMB
  long i;
  if (!IsIndet(i, runtimeEnv->evalArgAs<RingElemValue>(ARG(0))->theRingElem))
    throw RuntimeException("not an indet", ARG(0).exp);
  return Value::from(i+1);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IndetName, 1) { // AMB
	intrusive_ptr<RingElemValue> a = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  long i;
  if (!IsIndet(i,a->theRingElem))
    throw RuntimeException("not an indet", ARG(0).exp);
  symbol s=IndetSymbol(AsPolyRing(owner(a->theRingElem)), i);
	return Value::from(head(s));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IndetInd, 1) { // AMB
	intrusive_ptr<RingElemValue> a = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  long i;
  if (!IsIndet(i,a->theRingElem))
    throw RuntimeException("not an indet", ARG(0).exp);
  symbol s=IndetSymbol(AsPolyRing(owner(a->theRingElem)), i);
  vector<long> ind;
  for (long n=0; n<NumIndices(s); ++n) ind.push_back(index(s,n));
	return Value::from(ind);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IndetsCalled, 2) { // AMB
	return Value::from(IndetsCalled(
                                  AsPolyRing((runtimeEnv->evalArgAs<RingValue>(ARG(1)))->theRing),
                                  (runtimeEnv->evalArgAs<StringValue>(ARG(0)))->theString));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(ContentWRT, 2) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
	intrusive_ptr<RightValue> idt = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  if (intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(idt))
    return Value::from(ContentWRT(f->theRingElem, x->theRingElem));
  if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(idt))
  {
//     vector<long> v1 = VectorLongDecr1(runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1)), ERR::BadIndetIndex, "ContentWRT");
//     return Value::from(ContentWRT(f->theRingElem, v1));
    vector<RingElem> v1 = runtimeEnv->evalArgAsRingElemList(ARG(1), owner(f->theRingElem));
    return Value::from(ContentWRT_forC5(f->theRingElem, v1));
  }
  throw WrongTypeException(ListValue::type->name + " or " +
                           RingElemValue::type->name,
                           idt->getType()->name, ARG(1).exp);
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CoefficientsWRT, 2) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
	intrusive_ptr<RightValue> idt = runtimeEnv->evalArgAs<RightValue>(ARG(1));
  std::vector<CoeffPP> CoeffWRT;
  if (intrusive_ptr<RingElemValue> x = dynamic_pointer_cast<RingElemValue>(idt))
    CoeffWRT = CoefficientsWRT(f->theRingElem, x->theRingElem);
  else
  if (intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(idt))
  {
//     vector<long> v1 = VectorLongDecr1(runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1)), ERR::BadIndetIndex, "ContentWRT");
//     CoeffWRT = CoefficientsWRT(f->theRingElem, v1);
    vector<RingElem> v1 = runtimeEnv->evalArgAsRingElemList(ARG(1), owner(f->theRingElem));
    CoeffWRT = CoefficientsWRT_forC5(f->theRingElem, v1);
  }
  else
    throw WrongTypeException(ListValue::type->name + " or " +
                             RingElemValue::type->name,
                             idt->getType()->name, ARG(1).exp);
  // create return value: list of ...
  SparsePolyRing P = AsSparsePolyRing(owner(f->theRingElem));
	intrusive_ptr<ListValue> returnValue(new ListValue);
  BOOST_FOREACH(const CoeffPP cpp, CoeffWRT)
  {
    // create return value: ... record
    intrusive_ptr<RecordValue> cpp5(new RecordValue);
    cpp5->setField("Coeff", new RingElemValue(cpp.myCoeff));
    cpp5->setField("PP",    new RingElemValue(monomial(P,1,cpp.myPP)));
    returnValue->addValue(cpp5);
	}
  return returnValue;
}
END_STD_BUILTIN_FUNCTION



//-- matrix -----

DECLARE_STD_BUILTIN_FUNCTION(SetEntry, 4) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	intrusive_ptr<BigIntValue> J = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
  int which;
	intrusive_ptr<RightValue> v = runtimeEnv->evalArgAsT1orT2orT3<BigIntValue, BigRatValue, RingElemValue>(ARG(3), which);
  RingElem x(BaseRing(M->theMatrix));
	if (which==1) x = intrusive_ptr_cast<BigIntValue>(v)->theInteger;
	if (which==2) x = intrusive_ptr_cast<BigRatValue>(v)->theRational;
	if (which==3) x = intrusive_ptr_cast<RingElemValue>(v)->theRingElem;
	SetEntry_forC5(M->theMatrix, I->theInteger, J->theInteger, x);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SetRow, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(2), BaseRing(M->theMatrix));
	SetRow_forC5(M->theMatrix, I->theInteger, v);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SetCol, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(2), BaseRing(M->theMatrix));
	SetCol_forC5(M->theMatrix, I->theInteger, v);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SwapRows, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	intrusive_ptr<BigIntValue> I2 = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
	SwapRows_forC5(M->theMatrix, I1->theInteger, I2->theInteger);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SwapCols, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> I1 = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
	intrusive_ptr<BigIntValue> I2 = runtimeEnv->evalArgAs<BigIntValue>(ARG(2));
	SwapCols_forC5(M->theMatrix, I1->theInteger, I2->theInteger);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(AssignZero, 1) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->obtainUnshared<MatrixValue>(ARG(0));
	AssignZero(M->theMatrix);
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsZeroRow, 2) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger))
    throw RuntimeException("invalid row index", ARG(1).exp);
  return BoolValue::fromBool(IsZeroRow(M->theMatrix, n-1));
	//return new BoolValue(IsZeroRow(M->theMatrix, n));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(IsZeroCol, 2) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long n;
  if (!IsConvertible(n, N->theInteger))
    throw RuntimeException("invalid col index", ARG(1).exp);
  return BoolValue::fromBool(IsZeroCol(M->theMatrix, n-1));
	//return new BoolValue(IsZeroCol(M->theMatrix, n));
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(IsPositiveGrading) { return (1<=nArg) && (nArg<=2); }
DECLARE_BUILTIN_FUNCTION(IsPositiveGrading) { // AMB
	invocationExpression->checkNumberOfArgs(1,2);
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
  if (invocationExpression->args.size()==1)
    return BoolValue::fromBool(IsPositiveGrading(M->theMatrix)); // 1 arg: done
  intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  long GrDim = NumRows(M->theMatrix);
  if (!IsConvertible(GrDim, N->theInteger))
    throw RuntimeException("invalid row index", ARG(1).exp);
  return BoolValue::fromBool(IsPositiveGrading(M->theMatrix, GrDim));
}
//END_STD_BUILTIN_FUNCTION // no: variable number of args

DECLARE_STD_BUILTIN_FUNCTION(submat, 3) { // AMB
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
  vector<long> rows = VectorLongDecr1(runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1)), ERR::BadRowIndex, "submat");
  vector<long> cols = VectorLongDecr1(runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(2)), ERR::BadColIndex, "submat");
	return new MatrixValue(NewDenseMat(submat(M->theMatrix, rows, cols)));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(factor, 1) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
  factorization<RingElem> facs = factor(f->theRingElem);
  intrusive_ptr<RecordValue> facsC5(new RecordValue);
  facsC5->setField("Factors", new ListValue(facs.myFactors));
  facsC5->setField("Exponents", new ListValue(facs.myExponents));
  facsC5->setField("RemainingFactor", new RingElemValue(facs.myRemainingFactor));
  return facsC5;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(SmoothFactor, 2) { // AMB
	intrusive_ptr<BigIntValue> N = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
	intrusive_ptr<BigIntValue> TrialLimit = runtimeEnv->evalArgAs<BigIntValue>(ARG(1));
  factorization<BigInt> facs = SmoothFactor_forC5(N->theInteger, TrialLimit->theInteger);
  intrusive_ptr<RecordValue> facsC5(new RecordValue);
  facsC5->setField("Factors", new ListValue(facs.myFactors));
  facsC5->setField("Exponents", new ListValue(facs.myExponents));
  facsC5->setField("RemainingFactor", new BigIntValue(facs.myRemainingFactor));
  return facsC5;
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NR, 2) { // AMB
	intrusive_ptr<RingElemValue> f = runtimeEnv->evalArgAs<RingElemValue>(ARG(0));
	vector<RingElem> v = runtimeEnv->evalArgAsListOf<RingElemValue>(ARG(1));
	return Value::from(NR(f->theRingElem,v));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(BinSequentialToric, 2) { // AMB
  intrusive_ptr<IdealValue> I = runtimeEnv->evalArgAs<IdealValue>(ARG(0));
  vector<BigInt> indices = runtimeEnv->evalArgAsListOf<BigIntValue>(ARG(1));
	return Value::from(SequentialToric_C(I->theIdeal, VectorLong(indices, "BinSequentialToric")));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(MatSequentialToric, 2) { // AMB
  intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
	return Value::from(SequentialToric_C(AsSparsePolyRing(R->theRing), M->theMatrix));
}
END_STD_BUILTIN_FUNCTION

DECLARE_ARITYCHECK_FUNCTION(ideal) { return 1<=nArg; }
DECLARE_BUILTIN_FUNCTION(ideal) {
	const int nArgs = invocationExpression->args.size();
	if (nArgs==0)
		throw RuntimeException("Wrong number of arguments; found: 0, expecting: at least 1", invocationExpression);
	intrusive_ptr<RightValue> x = runtimeEnv->evalArgAs<RightValue>(ARG(0));
  if (invocationExpression->args.size()==2)
    if (intrusive_ptr<RingValue> R = dynamic_pointer_cast<RingValue>(x))
      return Value::from(ideal(R->theRing, runtimeEnv->evalArgAsRingElemList(ARG(1), R->theRing)));
	return Value::from(ideal(runtimeEnv->evalAllArgsAsListOf<RingElemValue>(invocationExpression)));
}

DECLARE_STD_BUILTIN_FUNCTION(NewRingTwinFloat, 1) { // AMB
	intrusive_ptr<BigIntValue> prec = runtimeEnv->evalArgAs<BigIntValue>(ARG(0));
  long d;
  if (!IsConvertible(d, prec->theInteger))
		throw RuntimeException("invalid precision", ARG(0).exp);
	return new RingValue(NewRingTwinFloat(d));
}
END_STD_BUILTIN_FUNCTION


DECLARE_ARITYCHECK_FUNCTION(NewPolyRing) { return (nArg==2) || (nArg==4); }
DECLARE_BUILTIN_FUNCTION(NewPolyRing) {
  //	invocationExpression->checkNumberOfArgs(2,4);
	const int nArgs = invocationExpression->args.size();
	if (nArgs!=2 && nArgs!=4)
		throw RuntimeException("Wrong number of arguments; found: "+boost::lexical_cast<std::string>(nArgs)+", expecting: 2 or 4", invocationExpression);
	intrusive_ptr<RingValue> R = runtimeEnv->evalArgAs<RingValue>(ARG(0));
  //	vector<string> v = runtimeEnv->evalArgAsListOf<StringValue>(ARG(1));
	intrusive_ptr<ListValue> l = runtimeEnv->evalArgAs<ListValue>(ARG(1));
	ListValue::ContainerType::size_type size = l->size();
  vector<symbol> v;
  for (unsigned long i=0; i<size; ++i)
  {
		if (const boost::intrusive_ptr<StringValue> s = boost::dynamic_pointer_cast<StringValue>(l->getValue(i)))
      v.push_back(symbol(s->theString));
    else if (const boost::intrusive_ptr<RecordValue> r = boost::dynamic_pointer_cast<RecordValue>(l->getValue(i)))
    {
      const boost::intrusive_ptr<StringValue> s = boost::dynamic_pointer_cast<StringValue>(r->getField("head"));
      const boost::intrusive_ptr<ListValue> inds = boost::dynamic_pointer_cast<ListValue>(r->getField("indices"));
      vector<long> indices;
      long tmp;
      ListValue::ContainerType::size_type NumIndices = inds->size();
      for (unsigned long j=0; j<NumIndices; ++j)
        if (IsConvertible(tmp, (boost::dynamic_pointer_cast<BigIntValue>(inds->getValue(j)))->theInteger))
          indices.push_back(tmp);
        else
          throw RuntimeException("All indices must fit into a machine-integer", ARG(1).exp);
      v.push_back(symbol(s->theString, indices));
    }
    else throw RuntimeException("List of symbols must have strings or records", ARG(1).exp);
  }
  if (nArgs==2)
    return new RingValue(NewPolyRing(R->theRing, v));
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(2));
	intrusive_ptr<BigIntValue> GradingDim = runtimeEnv->evalArgAs<BigIntValue>(ARG(3));
  long d;
  if (!IsConvertible(d, GradingDim->theInteger))
		throw RuntimeException("invalid GradingDim", ARG(3).exp);
  const PPOrdering PPO = NewMatrixOrdering(v.size(), d, M->theMatrix);
	return new RingValue(NewPolyRing(R->theRing, NewPPMonoid(v, PPO)));
}
  //END_STD_BUILTIN_FUNCTION // no: variable number of args


DECLARE_STD_BUILTIN_FUNCTION(elim, 2) { // AMB
	vector<RingElem> v1 = runtimeEnv->evalArgAsListOf<RingElemValue>(ARG(0));
	intrusive_ptr<IdealValue> I = runtimeEnv->evalArgAs<IdealValue>(ARG(1));
  ideal J = I->theIdeal;
  MakeUnique(J)->myElim(v1); // I->myElim(v1) ~~~> const
	return Value::from(J);
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(saturate, 2) { // AMB
	intrusive_ptr<IdealValue> I0 = runtimeEnv->evalArgAs<IdealValue>(ARG(0));
	intrusive_ptr<IdealValue> I1 = runtimeEnv->evalArgAs<IdealValue>(ARG(1));
  ideal J = I0->theIdeal;
  MakeUnique(J)->mySaturate(I1->theIdeal);
	return Value::from(J);
}
END_STD_BUILTIN_FUNCTION


DECLARE_STD_BUILTIN_FUNCTION(TopLevelFunctions, 0) { // GL
	return runtimeEnv->topLevelFunctions();
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(CocoaPackagePath, 0) { // GL
	(void)runtimeEnv; // keeps the compiler happy (or, at least, silent ;-) )
	return new StringValue(packageDir);
}
END_STD_BUILTIN_FUNCTION

//---- manual CoCoAHelp ----

DECLARE_STD_BUILTIN_FUNCTION(ReloadMan, 0) { // AMB
	try {
    ostringstream os;
    OnlineHelp::ReloadMan(os);
    runtimeEnv->getOutputStream()->print(os.str())->flush();
	} catch (const std::exception& err) {
		throw RuntimeException(err.what(), invocationExpression->getBegin(), invocationExpression->getEnd());
	}
	return VoidValue::theInstance;
}
END_STD_BUILTIN_FUNCTION

//---- Normaliz ----

DECLARE_STD_BUILTIN_FUNCTION(NormalizHilbertBasis, 1) {
// 	intrusive_ptr<ListValue> listOfList = runtimeEnv->evalArgAs<ListValue>(ARG(0));
//   vector<vector<BigInt> > v;
// 	const ListValue::ContainerType::size_type nLists = listOfList->size();
// 	for (ListValue::ContainerType::size_type a=0; a<nLists; ++a) {
// 		intrusive_ptr<ListValue> l = dynamic_pointer_cast<ListValue>(listOfList->getValue(a));
// 		if (!l)
// 			throw RuntimeException("The argument is not a list of lists", ARG(0).exp);
//     v.push_back(vector<BigInt>());
// 		const ListValue::ContainerType::size_type len = l->size();
// 		for (ListValue::ContainerType::size_type b=0; b<len; ++b)
//     {
//       intrusive_ptr<BigIntValue> n = dynamic_pointer_cast<BigIntValue>(l->getValue(b));
//       if (!n)
//         throw RuntimeException("The argument is not a list of lists on INT", ARG(0).exp);
// 			v[a].push_back(n->theInteger);
//     }
// 	}
	intrusive_ptr<MatrixValue> M = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
  return Value::from(NormalizHilbertBasis_forC5(M->theMatrix));
}
END_STD_BUILTIN_FUNCTION

DECLARE_STD_BUILTIN_FUNCTION(NormalizNormalToricRing, 1) {
  vector<RingElem> v = runtimeEnv->evalArgAsRingElemList(ARG(0));
  return Value::from(NormalizNormalToricRing_forC5(v));
}
END_STD_BUILTIN_FUNCTION


//---- GSL ----

#ifdef CoCoA_WITH_GSL
  DECLARE_COCOALIB_FUNCTION1(LU, MatrixValue)

  DECLARE_STD_BUILTIN_FUNCTION(SolverLU, 2) {
    intrusive_ptr<MatrixValue> A = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
    vector<RingElem> b = runtimeEnv->evalArgAsRingElemList(ARG(1));
    vector<RingElem> x = SolverLU(A->theMatrix, b);
    return new ListValue(x);
  }
  END_STD_BUILTIN_FUNCTION

  DECLARE_COCOALIB_FUNCTION1(CholeskyDecomposition, MatrixValue)
  DECLARE_COCOALIB_FUNCTION1(SVD, MatrixValue)
  DECLARE_COCOALIB_FUNCTION1(SingularValues, MatrixValue)
  
  DECLARE_STD_BUILTIN_FUNCTION(SolverSVD, 2) {
    intrusive_ptr<MatrixValue> A = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
    vector<RingElem> b = runtimeEnv->evalArgAsRingElemList(ARG(1));//, R->theRing);
    vector<RingElem> x = SolveSVD(A->theMatrix, b);
    return new ListValue(x);
  }
  END_STD_BUILTIN_FUNCTION

  DECLARE_COCOALIB_FUNCTION1(QR, MatrixValue)
  DECLARE_COCOALIB_FUNCTION1(QRPT, MatrixValue)
 
  DECLARE_STD_BUILTIN_FUNCTION(SolverQR, 2) {
    intrusive_ptr<MatrixValue> A = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
    vector<RingElem> b = runtimeEnv->evalArgAsRingElemList(ARG(1));
    vector<RingElem> x = SolverQR(A->theMatrix, b);
    return new ListValue(x);
  }
  END_STD_BUILTIN_FUNCTION

  DECLARE_STD_BUILTIN_FUNCTION(BidiagDecomposition, 4) {
    intrusive_ptr<MatrixValue> A = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
    intrusive_ptr<MatrixValue> D = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
    vector<RingElem> tau_U = runtimeEnv->evalArgAsListOf<RingElemValue>(ARG(2));
    vector<RingElem> tau_V = runtimeEnv->evalArgAsListOf<RingElemValue>(ARG(3)); 
    BidiagDecomposition(A->theMatrix, D->theMatrix, tau_U, tau_V);
    return VoidValue::theInstance;
  }
  END_STD_BUILTIN_FUNCTION

  DECLARE_STD_BUILTIN_FUNCTION(BidiagDecompUnpack, 5) {
    intrusive_ptr<MatrixValue> A = runtimeEnv->evalArgAs<MatrixValue>(ARG(0));
    intrusive_ptr<MatrixValue> U = runtimeEnv->evalArgAs<MatrixValue>(ARG(1));
    intrusive_ptr<MatrixValue> V = runtimeEnv->evalArgAs<MatrixValue>(ARG(2));
    vector<RingElem> tau_U = runtimeEnv->evalArgAsListOf<RingElemValue>(ARG(3));
    vector<RingElem> tau_V = runtimeEnv->evalArgAsListOf<RingElemValue>(ARG(4)); 
    BidiagDecompUnpack(A->theMatrix, U->theMatrix, V->theMatrix, tau_U, tau_V);
    return VoidValue::theInstance;
  }
  END_STD_BUILTIN_FUNCTION
#endif //  CoCoA_WITH_GSL



} // namespace {

void RuntimeEnvironment::initBuiltInFunctions()
{
	BOOST_FOREACH(NameFunPair &p, builtIns)
		this->setTopLevelVar(p.first, p.second, VariableSlot::VSF_SystemProtected);
}

} // namespace AST
} // namespace CoCoA
