//   Copyright (c) 2010 Giovanni Lagorio (lagorio@disi.unige.it)
//
//   This file is part of the source of CoCoALib, the CoCoA Library.
//
//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.
//
//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "CoCoALibSupplement.H"
#include "CoCoA/library.H"

#include <string>
//#include <vector> // already included by library.H->io.H
using std::vector;

namespace CoCoA {

  std::vector<long> VectorLong(const std::vector<BigInt>& BigIntVec, const std::string& FuncName)
  {
    vector<long> v;
    long tmp;
    for (long i=0; i<len(BigIntVec); ++i)
      if (IsConvertible(tmp, BigIntVec[i])) v.push_back(tmp);
      else CoCoA_ERROR(ERR::BadSymbolRange, FuncName);
    return v;
  }
  

  std::vector<long> VectorLongDecr1(const std::vector<BigInt>& BigIntVec, const ERR::ID& ErrID, const std::string& FuncName)
  {
    vector<long> v;
    long tmp;
    for (long i=0; i<len(BigIntVec); ++i)
      if (IsConvertible(tmp, BigIntVec[i])) v.push_back(tmp-1);
      else CoCoA_ERROR(ErrID, FuncName);
    return v;
  }
  

  bool IsTerm_forC5(ConstRefRingElem f)
  {
    return IsMonomial(f)&&IsOne(LC(f));
  }

  
  RingElem LT_forC5(ConstRefRingElem f)
  {
    return LPP_forC5(f);
  }  ///< NB result belongs to owner(f)


  RingElem LPP_forC5(ConstRefRingElem f)
  {
    ring R = owner(f);
    if (!IsPolyRing(R)) CoCoA_ERROR(ERR::NotElemPolyRing, "LPP_forC5");
    if (IsSparsePolyRing(R)) return monomial(AsSparsePolyRing(R), 1, LPP(f));
    PolyRing Rx = AsPolyRing(R); // univariate case
    return IndetPower(Rx, 0, deg(f));
  }


  RingElem LM_forC5(ConstRefRingElem f)
  {
    ring R = owner(f);
    if (!IsPolyRing(R)) CoCoA_ERROR(ERR::NotElemPolyRing, "LM_forC5");
    if (IsSparsePolyRing(R)) return monomial(AsSparsePolyRing(R), LC(f), LPP(f));
    PolyRing Rx = AsPolyRing(R);
    return CoeffEmbeddingHom(Rx)(LC(f))*IndetPower(Rx, 0, deg(f));    
  }

  namespace
  {
    RingElem CoeffOfTermSparse(ConstRefRingElem f, ConstRefPPMonoidElem pp)
    {
      for (SparsePolyIter itf=BeginIter(f); !IsEnded(itf); ++itf)
        if (PP(itf) == pp) return coeff(itf);
      return zero(CoeffRing(AsPolyRing(owner(f))));
    }
    
    std::vector<RingElem> CoefficientsDense(ConstRefRingElem f)
    {
      vector<RingElem> v;
      DenseUPolyRing P = AsDenseUPolyRing(owner(f));
      for (long i=P->myDegPlus1(raw(f))-1; i>=0; --i)
        if (!IsZero(coeff(f,i))) v.push_back(coeff(f,i));
      return v;
    }
    
    std::vector<RingElem> CoefficientsSparse(ConstRefRingElem f)
    {
      vector<RingElem> v;
      for (SparsePolyIter itf=BeginIter(f); !IsEnded(itf); ++itf)
        v.push_back(coeff(itf));
      return v;
    }
    
    std::vector<RingElem> MonomialsDense(ConstRefRingElem f)
    {
      vector<RingElem> v;
      DenseUPolyRing P = AsDenseUPolyRing(owner(f));
      for (long i=P->myDegPlus1(raw(f))-1; i>=0; --i)
        if (!IsZero(coeff(f,i))) v.push_back(monomial(P, coeff(f,i), i));
      return v;
    }
    
    std::vector<RingElem> MonomialsSparse(ConstRefRingElem f)
    {
      vector<RingElem> v;
      SparsePolyRing P = AsSparsePolyRing(owner(f));
      for (SparsePolyIter itf=BeginIter(f); !IsEnded(itf); ++itf)
        v.push_back(monomial(P, coeff(itf), PP(itf)));
      return v;
    }
    
    std::vector<RingElem> SupportDense(ConstRefRingElem f)
    {
      CoCoA_ERROR("SupportDense is disabled", "SupportDense");
      vector<RingElem> v;
      DenseUPolyRing P = AsDenseUPolyRing(owner(f));
      for (long i=P->myDegPlus1(raw(f))-1; i>=0; --i)
        if (!IsZero(coeff(f,i))) v.push_back(monomial(P, one(CoeffRing(P)), i));
      return v;
    }

    std::vector<RingElem> SupportSparse(ConstRefRingElem f)
    {
      vector<RingElem> v;
      SparsePolyRing P = AsSparsePolyRing(owner(f));
      for (SparsePolyIter itf=BeginIter(f); !IsEnded(itf); ++itf)
        v.push_back(monomial(P, one(CoeffRing(P)), PP(itf)));
      return v;
    }  
  }
  

  RingElem CoeffOfTerm_forC5(ConstRefRingElem f, ConstRefRingElem t)
  {
    ring R = owner(f);
    if (!IsPolyRing(R)) CoCoA_ERROR(ERR::NotElemPolyRing, "CoeffOfTerm_forC5");
    if (!IsMonomial(t)) CoCoA_ERROR(ERR::BadArg, "CoeffOfTerm_forC5");
    if (!IsOne(LC(t))) CoCoA_ERROR(ERR::BadArg, "CoeffOfTerm_forC5");
    if (IsSparsePolyRing(R))
      return CoeffOfTermSparse(f, LPP(t));
    return coeff(f, deg(t));
  }
  
  
  std::vector<RingElem> coefficients_forC5(ConstRefRingElem f)
  {
    ring R = owner(f);
    if (!IsPolyRing(R)) CoCoA_ERROR(ERR::NotElemPolyRing, "coefficients_forC5");
    if (IsSparsePolyRing(R))  return CoefficientsSparse(f);
    return CoefficientsDense(f);
  }
  
  
  std::vector<RingElem> monomials_forC5(ConstRefRingElem f)
  {
    ring R = owner(f);
    if (!IsPolyRing(R)) CoCoA_ERROR(ERR::NotElemPolyRing, "monomials_forC5");
    if (IsSparsePolyRing(R))  return MonomialsSparse(f);
    return MonomialsDense(f);
  }
  
  
  std::vector<RingElem> support_forC5(ConstRefRingElem f)
  {
    ring R = owner(f);
    if (!IsPolyRing(R)) CoCoA_ERROR(ERR::NotElemPolyRing, "support_forC5");
    if (IsSparsePolyRing(R))  return SupportSparse(f);
    return SupportDense(f);
  }


  namespace
  {
    // needs a decent implementation, this one isn't
    RingElem DensePolyRec(const PolyRing& P, long n, long d)
    {
      if (n==1) return IndetPower(P,0,d);
      if (d==1)
      {
        RingElem s=indet(P,0);
        for (long i=1; i<n; ++i)  s+=indet(P,i);
        return s;
      }
      return indet(P,n-1)*DensePolyRec(P,n,d-1) + DensePolyRec(P,n-1,d);
    }    
  }


  RingElem DensePoly_forC5(const ring& R, const BigInt& D)
  {
    const PolyRing P = AsPolyRing(R);
    long d;
    if (!IsConvertible(d,D))  CoCoA_ERROR(ERR::ArgTooBig, "DensePoly_forC5");
    if (d<0)  CoCoA_ERROR(ERR::NotNonNegative, "DensePoly_forC5");
    if (d==0) return one(P);
    return DensePolyRec(P, NumIndets(P), d);
  }


  std::vector<BigInt> wdeg_forC5(ConstRefRingElem f)
  {
    degree d = wdeg(f);
    vector<BigInt> v;
    for (long i=0; i<GradingDim(d); ++i)  v.push_back(d[i]);
    return v;
  }


  BigInt NextPrime_forC5(BigInt N)
  {
    if (N <= 0) CoCoA_ERROR(ERR::NotNonNegative, "NextPrime_forC5");
    unsigned long n;
    if (!IsConvertible(n, N)) CoCoA_ERROR(ERR::ArgTooBig, "NextPrime_forC5");
    unsigned long ans = NextPrime(n);
    if (ans == 0) CoCoA_ERROR(ERR::ArgTooBig, "NextPrime_forC5");
    return BigInt(ans);
  }

  factorization<BigInt> SmoothFactor_forC5(const BigInt& N, const BigInt& TrialLimit)
  {
    unsigned long tl;
    if (!IsConvertible(tl, TrialLimit)) CoCoA_ERROR(ERR::ArgTooBig, "SmoothFactor_forC5");
    return SmoothFactor(N, tl);
  }


  BigRat CpuTime_forC5()
  {
    return ConvertTo<BigRat>(CpuTime()); 
//     BigRat res;
//     IsConvertible(res, CpuTime());
//     return res;
  }
  

  std::string date_forC5()
  {
    const static std::string d(__DATE__);
    const static std::string t(__TIME__);
    // fix day of the week
    return "    " + d.substr(0,7) + t + d.substr(6,5);
  }
  

  const BigInt random_forC5(const BigInt& lo, const BigInt& hi)
  {
    return RandomBigInt(GlobalRandomSource(), lo, hi);
  }
  

  RingElem lcm_forC5(const std::vector<RingElem>& v)
  {  // v is not empty
    RingElem res = v[0];
    for (long i=1; i<len(v); ++i)  res = lcm(res, v[i]);
    return res;
  }
  
  RingElem gcd_forC5(const std::vector<RingElem>& v)
  {  // v is not empty
    RingElem res = v[0];
    for (long i=1; i<len(v); ++i)  res = gcd(res, v[i]);
    return res;
  }
  

  RingElem ContentWRT_forC5(ConstRefRingElem f, const std::vector<RingElem>& v)
  {
    vector<long> indices(len(v));
    for (long i=0; i<len(v); ++i)
      if (!IsIndet(indices[i], v[i]))
        CoCoA_ERROR(ERR::NotIndet, "ContentWRT_forC5");
    return ContentWRT(f, indices);
  }
  

  std::vector<CoeffPP> CoefficientsWRT_forC5(ConstRefRingElem f, const std::vector<RingElem>& v)
  {
    vector<long> indices(len(v));
    for (long i=0; i<len(v); ++i)
      if (!IsIndet(indices[i], v[i]))
        CoCoA_ERROR(ERR::NotIndet, "ContentWRT_forC5");
    return CoefficientsWRT(f, indices);
  }
  

  RingElem HilbertNumQuot_forC5(const ideal& I)
  {
    PPOrdering PPO(ordering(PPM(AsSparsePolyRing(AmbientRing(I)))));
    if (GradingDim(PPO) ==0)  // aff
      CoCoA_ERROR(ERR::NYI, "HilbertNumQuot_forC5");
    //    if (GradingDim(PPO) >1)  // multigraded
    if (GradingDim(PPO) ==1)  // standard case
    {
      for (long i=0; i<NumIndets(PPO); ++i)
        if (GetMatrix(PPO)(0,i) != 1) // non-standard
        {
          // return SingleWeights...
          SparsePolyRing HSRing = NewPolyRing(RingQQ(), GradingDim(PPO));
          return MGHilbertNumQuot(HSRing, I);
        }
      return HilbertNumQuot(I);
    }
    //    if (GradingDim(PPO) >1)  // multigraded
    SparsePolyRing HSRing = NewPolyRing(RingQQ(), GradingDim(PPO));
    return MGHilbertNumQuot(HSRing, I);
  }
  
  
  //----- matrix --------------------------------------------------

  matrix NewDenseMat_forC5(const ring& R, const BigInt& NR, const BigInt& NC)
  {
    long r,c;
    if (!IsConvertible(r, NR)) CoCoA_ERROR(ERR::BadRowIndex, "NewDenseMat_forC5");
    if (!IsConvertible(c, NC)) CoCoA_ERROR(ERR::BadColIndex, "NewDenseMat_forC5");
    return NewDenseMat(R, r, c);
  }
  

  matrix HomogElimMat_forC5(ConstMatrixView M, const std::vector<BigInt>& ElimInds)
  {
    std::vector<long> v = VectorLong(ElimInds, "HomogElimMat_forC5");
    for (long i=0; i<len(v); ++i)
      --v[i];
    return NewHomogElimMat(M, v);
  }
  

  matrix RevLexMat_forC5(const BigInt& N)
  {
    long n;
    if (!IsConvertible(n, N)) CoCoA_ERROR(ERR::BadRowIndex, "RevLexMat_forC5");
    return NewDenseMatRevLex(n);
  }
  

  matrix StdDegLexMat_forC5(const BigInt& N)
  {
    long n;
    if (!IsConvertible(n, N)) CoCoA_ERROR(ERR::BadRowIndex, "StdDegLexMat_forC5");
    return NewDenseMatStdDegLex(n);
  }
  

  matrix StdDegRevLexMat_forC5(const BigInt& N)
  {
    long n;
    if (!IsConvertible(n, N)) CoCoA_ERROR(ERR::BadRowIndex, "StdDegRevLexMat_forC5");
    return NewDenseMatStdDegRevLex(n);
  }
  

  matrix XelMat_forC5(const BigInt& N)
  {
    long n;
    if (!IsConvertible(n, N)) CoCoA_ERROR(ERR::BadRowIndex, "XelMat_forC5");
    return NewDenseMatXel(n);
  }
  

  matrix LexMat_forC5(const BigInt& N)
  {
    long n;
    if (!IsConvertible(n, N)) CoCoA_ERROR(ERR::BadRowIndex, "LexMat_forC5");
    return NewDenseMat(IdentityMat(RingQQ(), n));
  }
  

  matrix ExtractOrdMat_forC5(const matrix& M)
  {
    return NewMatMinimize(M);
  }
  

  void SetEntry_forC5(MatrixView& M, const BigInt& I, const BigInt& J, ConstRefRingElem x)
  {
    long i,j;
    if (!IsConvertible(i, I)) CoCoA_ERROR(ERR::BadRowIndex, "SetEntry_forC5");
    if (!IsConvertible(j, J)) CoCoA_ERROR(ERR::BadColIndex, "SetEntry_forC5");
    SetEntry(M, i-1, j-1, x);
  }
  

  std::vector<RingElem> GetRow_forC5(ConstMatrixView M, const BigInt& I)
  {
    long i;
    if (!IsConvertible(i, I)) CoCoA_ERROR(ERR::BadRowIndex, "GetRow_forC5");
    vector<RingElem> v;
    --i;
    for (long j=0; j<NumCols(M); ++j)  v.push_back(M(i,j));
    return v;
  }
  

  std::vector<RingElem> GetCol_forC5(ConstMatrixView M, const BigInt& I)
  {
    long i;
    if (!IsConvertible(i, I)) CoCoA_ERROR(ERR::BadColIndex, "GetCol_forC5");
    vector<RingElem> v;
    --i;
    for (long j=0; j<NumRows(M); ++j)  v.push_back(M(j,i));
    return v;
  }
  

  void SetRow_forC5(MatrixView& M, const BigInt& I, const std::vector<RingElem>& v)
  {
    long i;
    if (!IsConvertible(i, I)) CoCoA_ERROR(ERR::BadRowIndex, "SetRow_forC5");
    --i;
    for (long j=0; j<NumCols(M); ++j)  SetEntry(M, i, j, v[j]);
  }
  

  void SetCol_forC5(MatrixView& M, const BigInt& I, const std::vector<RingElem>& v)
  {
    long i;
    if (!IsConvertible(i, I)) CoCoA_ERROR(ERR::BadColIndex, "SetCol_forC5");
    --i;
    for (long j=0; j<NumRows(M); ++j)  SetEntry(M, j, i, v[j]);
  }
  

  void SwapRows_forC5(matrix& M, BigInt i1, BigInt i2)
  {
    long j1, j2;
    if (!IsConvertible(j1, i1)) CoCoA_ERROR(ERR::BadRowIndex, "SwapRows_forC5");
    if (!IsConvertible(j2, i2)) CoCoA_ERROR(ERR::BadRowIndex, "SwapRows_forC5");
    SwapRows(M, --j1, --j2);
  }
  

  void SwapCols_forC5(matrix& M, BigInt i1, BigInt i2)
  {
    long j1, j2;
    if (!IsConvertible(j1, i1)) CoCoA_ERROR(ERR::BadColIndex, "SwapCols_forC5");
    if (!IsConvertible(j2, i2)) CoCoA_ERROR(ERR::BadColIndex, "SwapCols_forC5");
    SwapCols(M, --j1, --j2);
  }


  MatrixView IdentityMat_forC5(const ring& R, const BigInt& N)
  {
    long n;
    if (!IsConvertible(n, N)) CoCoA_ERROR(ERR::BadMatrixSize, "IdentityMat_forC5");
    return IdentityMat(R, n);
  }
  

  MatrixView ZeroMat_forC5(const ring& R, const BigInt& NRows, const BigInt& NCols)
  {
    long nrows, ncols;
    if (!IsConvertible(nrows, NRows)) CoCoA_ERROR(ERR::BadMatrixSize, "ZeroMat_forC5");
    if (!IsConvertible(ncols, NCols)) CoCoA_ERROR(ERR::BadMatrixSize, "ZeroMat_forC5");
    return ZeroMat(R, nrows, ncols);
  }
  

  MatrixView FilledMat_forC5(const ring& R, const BigInt& NRows, const BigInt& NCols, ConstRefRingElem x)
  {
    long nrows, ncols;
    if (!IsConvertible(nrows, NRows)) CoCoA_ERROR(ERR::BadMatrixSize, "ZeroMat_forC5");
    if (!IsConvertible(ncols, NCols)) CoCoA_ERROR(ERR::BadMatrixSize, "ZeroMat_forC5");
    return FilledMat(R, nrows, ncols, x);
  }
  

  ConstMatrixView transposed_forC5(ConstMatrixView M)
  { return transpose(M); }


  long GradingDim_forC5(const ring& P)
  {
    return GradingDim(AsSparsePolyRing(P));
  }
  

  matrix WeightsMatrix_forC5(const ring& P)
  {
    PPOrdering PPO(ordering(PPM(AsSparsePolyRing(P))));
    vector<long> rows, cols;
    for (long i=0; i<GradingDim(PPO); ++i)  rows.push_back(i);
    for (long i=0; i<NumIndets(PPO); ++i)  cols.push_back(i);    
    return NewDenseMat(submat(GetMatrix(PPO), rows, cols));
  }
  

  matrix ord_forC5(const ring& P)
  {
    PPOrdering PPO(ordering(PPM(AsSparsePolyRing(P))));
    return GetMatrix(PPO);
  }


//   std::vector<RingElem> minors_forC5(const BigInt& N, ConstMatrixView M)
//   {
//     vector<RingElem> v;
//     size_t n;
//     if (!IsConvertible(n, N))
//       CoCoA_ERROR(ERR::BadRowIndex, "minors_forC5");
//     if (n>NumRows(M))  CoCoA_ERROR(ERR::BadRowIndex, "minors_forC5");
//     if (n>NumCols(M))  CoCoA_ERROR(ERR::BadColIndex, "minors_forC5");
//     vector<size_t> rows(n);
//     vector<size_t> cols(n);
// ---- we need a tuple-generator ----
//         v.push_back(det(submat(M,rows,cols)));
//     return v;
//   }
//   }

  //-------- NORMALIZ --------------------------------------------------
  
  matrix NormalizHilbertBasis_forC5(ConstMatrixView M)
  {
#ifdef CoCoA_WITH_NORMALIZ
    return NewDenseMat(BaseRing(M), Normaliz::HilbertBasis(Normaliz::NewConeLong(M, libnormaliz::Type::normalization)));
#else
    CoCoA_ERROR("No Normaliz in this CoCoALib", "NormalizHilbertBasis2_forC5");
    return NewDenseMat(M); // never reached, just to keep compiler quiet
#endif
  }

  std::vector<RingElem> NormalizNormalToricRing_forC5(const std::vector<RingElem>& v)
  {
#ifdef CoCoA_WITH_NORMALIZ
/*    if(!AreMonomial(v)) {
      CoCoA_ERROR("Expected list of monomials","NormalizNormalToricRing_forC5");
    }*/
    if(v.size()==0) {
      CoCoA_ERROR("List of monomials has to be non-empty","NormalizNormalToricRing_forC5");
    }
    //convert it to a PPVector
    PPVector ppv(PPM(AsSparsePolyRing(owner(v[0]))), NewDivMaskNull());
    convert(ppv,v);
    //run Normaliz and convert back to vector<RingElem>
    std::vector<RingElem> res;
    convert(res,owner(v[0]),Normaliz::NormalToricRing(ppv));
    return res;
#else
    CoCoA_ERROR("No Normaliz in this CoCoALib", "NormalizNormalToricRing_forC5");
    return v; // never reached, just to keep compiler quiet
#endif
  }


}

