//   Copyright (c)  2010  Anna Bigatti

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.

#include "OnlineHelp.H"
#include "CoCoA/error.H"

#include <algorithm>
using std::transform;
#include <cctype>
//using std::tolower;
#include <fstream>
using std::ifstream;
#include <iostream>
using std::clog;
using std::endl;
#include <string>
using std::string;
#include <vector>
using std::vector;


extern string packageDir;

namespace CoCoA
{
namespace OnlineHelp
{

  class index
  {
  public: 
    class entry
    {
    public:
      entry(const std::string& title): myTitleValue(title) {}
      // ~entry(); // default is OK
      const std::string& myTitle() const {return myTitleValue;}
      const std::vector<std::string> myKeys() const {return myKeysValue;}
      const std::vector<std::string> myTypes() const {return myTypesValue;}
      const std::vector<std::string> mySeeAlso() const {return mySeeAlsoValue;}
      void myAddKey(const std::string& key);
      void myAddType(const std::string& key);
      void myAddSee(const std::string& key);
      void myAddKeys(std::ifstream& in);
      void myAddTypes(std::ifstream& in);
      void myAddSeeAlso(std::ifstream& in);

    private: // data members
      std::string myTitleValue;
      std::vector<std::string> myKeysValue;
      std::vector<std::string> myTypesValue;
      std::vector<std::string> mySeeAlsoValue;
    };
 
  public:
    index();
    // ~index(); // default is OK
    void myLoad(std::ostream &out);
    const std::string& myTitle(std::size_t i) const {return myVec[i].myTitle();}
    const std::vector<std::string> myKeys(std::size_t i) const {return myVec[i].myKeys();}
    const std::vector<std::string> myTypes(std::size_t i) const {return myVec[i].myTypes();}
    const std::vector<std::string> mySeeAlso(std::size_t i) const {return myVec[i].mySeeAlso();}
    std::size_t mySize() const {return myVec.size();}

  private:
    void myAddEntry(std::ifstream& in, const std::string& title, const std::string& ClosingTag);
  
  private: // data members
    std::vector<entry> myVec;
  };

  // ----------------------------------------------------------------------
  // fwd decl -- defined later in this file
  std::string LowerCase(const std::string& str);
  bool IsStr1BeforeStr2(std::ifstream& in, const std::string& s1, const std::string& s2);
  std::string StringInTag(const std::string& line, const std::string& XMLTag);
  std::string SkipToStringInTag(std::ifstream& in, const std::string& XMLTag);
  void PrintCommAndFunc(std::string& line, std::ostream &out);
  void PrintSeeAlso(const std::string& s, std::ostream &out);

  // ----------------------------------------------------------------------
  inline bool IsSubstring(const std::string& s, const std::string& line) 
  { return line.find(s) != string::npos; }

  // ----------------------------------------------------------------------

  //-- index --

  //  const index& UniqueIndex();
  index& UniqueIndex();

  index::index()
  {
    myLoad(clog);
  }


  void index::myLoad(std::ostream &out)
  {
    out << "Loading index ..." << std::flush;
    ifstream in(XMLFileName());
    if (in == 0)
      CoCoA_ERROR("Cannot find input file `CoCoAHelp.xml`", "OnlineHelp::index()");
    string line; // used only inside while loop
    string LastCommand;
    string NewCommand = "a";
    myVec.clear();
    while (!in.eof())
    {
      getline(in, line);
      if (IsSubstring("<!--", line)) continue;
      if (IsSubstring("<command>", line))
      {
        LastCommand = NewCommand;
        NewCommand = SkipToStringInTag(in, "<title>");
        // check for proper sorting
        if (false && LowerCase(LastCommand).compare(LowerCase(NewCommand))>0)
          std::cout << "OnlineHelp: unsorted entries: "
                    << LastCommand << " -- " << NewCommand << std::endl;
        myAddEntry(in, NewCommand, "</command>");
      }
      if (IsSubstring("<section>", line))
        myAddEntry(in, SkipToStringInTag(in, "<title>"), "</section>");
    }
    out << "... index loaded" << endl;
  }


  void index::myAddEntry(std::ifstream& in, const std::string& title, const std::string& ClosingTag)
  {
    myVec.push_back(entry(title));
    entry& e(myVec.back());
    e.myAddKey(title);
    string line;
    getline(in, line);
    while (!in.eof() && !IsSubstring(ClosingTag, line))
    {
      if (IsSubstring("<keys>", line)) e.myAddKeys(in);
      if (IsSubstring("<types>", line)) e.myAddTypes(in);
      if (IsSubstring("<seealso>", line)) e.myAddSeeAlso(in);
      getline(in, line);
    }
  }


  //-- index::entry --
  void index::entry::myAddKey(const std::string& key)
  { myKeysValue.push_back(LowerCase(key)); }

  void index::entry::myAddType(const std::string& t)
  { myTypesValue.push_back(LowerCase(t)); }

  void index::entry::myAddSee(const std::string& t)
  { mySeeAlsoValue.push_back(t); }

  void index::entry::myAddKeys(std::ifstream& in)
  {
    string line;
    getline(in, line);
    string s; // used only inside while loop
    while ((!in.eof()) && (!IsSubstring("</keys>",line)))
    {
      if ((s=StringInTag(line, "<key>")) != "") myAddKey(s);
      getline(in, line);
    }
  }

  void index::entry::myAddTypes(std::ifstream& in)
  {
    string line;
    getline(in, line);
    string s; // used only inside while loop
    while ((!in.eof()) && (!IsSubstring("</types>",line)))
    {
      if ((s=StringInTag(line, "<type>")) != "") myAddType(s);
      getline(in, line);
    }
  }


  void index::entry::myAddSeeAlso(std::ifstream& in)
  {
    string line;
    getline(in, line);
    string s; // used only inside while loop
    while ((!in.eof()) && (!IsSubstring("</seealso>",line)))
    {
      if ((s=StringInTag(line, "<see>")) != "") myAddSee(s);
      getline(in, line);
    }
  }


  //----------------------------------------------------------------

  std::string LowerCase(const std::string& str)
  {
    string s(str); // slightly wasteful, but readable
    transform(s.begin(), s.end(), s.begin(), tolower);
    return s;
  }


  bool IsStr1BeforeStr2(std::ifstream& in, const std::string& s1, const std::string& s2)
  {
    string line;
    getline(in, line);
    while (!in.eof())
    {
      if ( IsSubstring(s2, line) ) return false;
      if ( IsSubstring(s1, line) ) return true;
      getline(in, line);
    }
    CoCoA_ERROR(CoCoA::ERR::SERIOUS, "IsStr1BeforeStr2");
    return false;
  }
  

  // only for opening and closing tag in the same line
  std::string StringInTag(const std::string& line, const std::string& XMLTag)
  {
    size_t open;
    if ( (open=line.find(XMLTag)) == string::npos) return "";
    
    string ClosedXMLTag = XMLTag;
    ClosedXMLTag.replace(0, 1, "</");
    size_t close = line.find(ClosedXMLTag);
    if (close == string::npos)
      CoCoA_ERROR(XMLTag+" closing tag not found in this line", "StringInTag");
    size_t StartPos = open + XMLTag.length();
    return line.substr(StartPos, close-StartPos);
  }


  std::string SkipToStringInTag(std::ifstream& in, const std::string& XMLTag)
  {
    string line;
    getline(in, line);
    string s;
    while (!in.eof())
    {
      if ( (s=StringInTag(line, XMLTag)) != "") return s;
      getline(in, line);
    }
    return "";
  }
  

  void ReplaceWith(std::string& line, const std::string& SFrom, const std::string& STo)
  {
    size_t open;
    while ( (open=line.find(SFrom)) != string::npos)
      line.replace(open, SFrom.length(), STo);
  }
  

  void ReplaceWithQuotes(std::string& line, const std::string& s)
  { ReplaceWith(line, s, "\""); }
  

  std::string& CleanupLine(std::string& line)
  {
    ReplaceWithQuotes(line, "<quotes>"); ReplaceWithQuotes(line, "</quotes>");
    ReplaceWithQuotes(line, "<em>");     ReplaceWithQuotes(line, "</em>");
    ReplaceWith(line, "&lt;", "<");
    ReplaceWith(line, "&gt;", ">");
    ReplaceWith(line, "&apos;", "'");
    ReplaceWith(line, "&amp;", "&");
    ReplaceWith(line, "<backslash/>", "\\");
    ReplaceWith(line, "<less_eq/>", "<=");
    return line;
  }


  std::string& ProcessLine(std::string& line)
  {
    ReplaceWithQuotes(line, "<quotes>"); ReplaceWithQuotes(line, "</quotes>");
    ReplaceWithQuotes(line, "<tt>");     ReplaceWithQuotes(line, "</tt>");
    ReplaceWithQuotes(line, "<code>");   ReplaceWithQuotes(line, "</code>");
    ReplaceWithQuotes(line, "<ttref>");  ReplaceWithQuotes(line, "</ttref>");
    ReplaceWithQuotes(line, "<ref>");    ReplaceWithQuotes(line, "</ref>");
    ReplaceWithQuotes(line, "<em>");     ReplaceWithQuotes(line, "</em>");
    ReplaceWith(line, "<verbatim>", ""); ReplaceWith(line, "</verbatim>", "");
    ReplaceWith(line, "<formula>", " "); ReplaceWith(line, "</formula>", " ");
    ReplaceWith(line, "&lt;", "<");
    ReplaceWith(line, "&gt;", ">");
    ReplaceWith(line, "&apos;", "'");
    ReplaceWith(line, "&amp;", "&");
    ReplaceWith(line, "<backslash/>", "\\");
    ReplaceWith(line, "<less_eq/>", "<=");
    ReplaceWith(line, "<par/>", "");
    ReplaceWith(line, "<par />", "");
    return line;
  }

  void PrintExample(std::ifstream& in, std::ostream &out)
  {
    out << "--====<  example  >====--" << endl;
    string line;
    getline(in, line);
    while ( !IsSubstring("</example>", line) )
    {
      out << CleanupLine(line) << endl;
      getline(in, line);
    }
    out << "--=========< end example >=========--" << endl;
  }
  

  void PrintExampleWithoutOutput(std::ifstream& in, std::ostream &out)
  {
    out << "--====<  example  >====--" << endl;
    string line;
    getline(in, line);
    while ( !IsSubstring("</example>", line) )
    {
      if (IsSubstring("/**/", line))  out << CleanupLine(line) << endl;
      getline(in, line);
    }
    out << "--=========< end example >=========--" << endl;
  }
  

  void PrintSyntax(std::ifstream& in, std::ostream &out)
  {
    out << "--====<  syntax  >====--" << endl;
    string line;
    getline(in, line);
    while ( !IsSubstring("</syntax>", line) )
    {
      out << "--  " << CleanupLine(line) << endl;
      getline(in, line);
    }
  }
  

void PrintSearchExact(std::string title, std::ostream &out)
{
  ifstream in(XMLFileName());
  if (in == 0)
  {
    std::cout << "Cannot find input file `CoCoAHelp.xml`.  Aborting." << endl;
    abort();
  }
  string s = SkipToStringInTag(in, "<title>");
  while (!in.eof() && s != title)  s = SkipToStringInTag(in, "<title>");
  if (in.eof())
    CoCoA_ERROR("Not found exact title: " + title, "PrintSearchExact");
//   out << 
// "======================<  !! WARNING !!  >=======================\n"
// " this is the CoCoA-4 manual: some functions might have changed  \n"
// " name or syntax and some are still missing.  See: \n"
// " http://cocoa.dima.unige.it/cocoalib/doc/CoCoALib-functions.html\n"
// "======================<  !! WARNING !!  >=======================\n"
//  << endl;
  out << "==============[ " << title << " ]===============" << endl;
  string line;
  getline(in, line);
  while (!IsSubstring("<description>", line))  getline(in, line);
  out << "--====<  description  >====--" << endl;
  getline(in, line);
  while (!IsSubstring("</description>", line))
  {
    if ( IsSubstring("<commands_and_functions_for", line))  PrintCommAndFunc(line, out);
    else
    {
      if ( IsSubstring("<example>", line))  PrintExample(in, out);
      else
        out << ProcessLine(line) << endl;
    }
    getline(in, line);
  }
  getline(in, line);
  while (!IsSubstring("<", line))   getline(in, line);
  if ( IsSubstring("<syntax>", line))    PrintSyntax(in, out);
//   out << 
// "======================<  !! WARNING !!  >=======================\n"
// " this is the CoCoA-4 manual: some functions might have changed  \n"
// " name or syntax and some are still missing.  See: \n"
// " http://cocoa.dima.unige.it/cocoalib/doc/CoCoALib-functions.html\n"
// "======================<  !! WARNING !!  >=======================\n"
//  << endl;
  PrintSeeAlso(title, out);
}


#if 0 
  // hopefully useless
  std::size_t SkipToLAngle(std::ifstream& in, std::string& line, std::size_t from)
  {
    size_t pos;
    while (!in.eof())
    {
      if ( (pos=line.find("<")) != string::npos) return pos;
      getline(in, line);
    }
    return string::npos;
  }
#endif


//----------------------------------------------------------------------

void SearchMatch(std::vector<std::string>& MatchingTitles, const std::string& s)
{
  string ls(LowerCase(s));
  
  const index& index(UniqueIndex());
  for (size_t i=0; i<index.mySize(); ++i)
    for (size_t j=0; j<index.myKeys(i).size(); ++j)
      if (IsSubstring(ls, index.myKeys(i)[j]))
      {
        MatchingTitles.push_back(index.myTitle(i));
        break;
      }
}


void SearchType(std::vector<std::string>& MatchingTitles, const std::string& s)
{
  string ls(LowerCase(s));
  
  const index& index(UniqueIndex());
  for (size_t i=0; i<index.mySize(); ++i)
    for (size_t j=0; j<index.myTypes(i).size(); ++j)
      if (IsSubstring(ls, index.myTypes(i)[j]))
      {
        MatchingTitles.push_back(index.myTitle(i));
        break;
      }
}


  void PrintCommAndFunc(std::string& TypeLine, std::ostream &out)
  {
    ReplaceWith(TypeLine, "<commands_and_functions_for type=\"", "");
    ReplaceWith(TypeLine, "\"></commands_and_functions_for>", "");
    std::vector<std::string> found;
    SearchType(found, TypeLine);
    for (size_t i=0; i<found.size(); ++i)
    {
      ifstream in(XMLFileName());  // if (in == 0) ...
      string s = SkipToStringInTag(in, "<title>");
      while (!in.eof() && s != found[i])  s = SkipToStringInTag(in, "<title>");
      if (in.eof())
        CoCoA_ERROR("Not found exact title: " + found[i], "PrintCommAndFunc");
      out << "? " << s << " -- ";
      string line;
      getline(in, line);
      while (!IsSubstring("<short_description>", line))  getline(in, line);
      out << StringInTag(line, "<short_description>") << endl;
    }
    out << "--==================<>===================--" << endl;
  }
  


  void PrintMan(const std::string& s, std::ostream &out)
  {
    vector<string> found;
    string PrintedFound = "";
    // remove trailing comments, spaces, ";"
    string str(s);
    if (IsSubstring("--",str))  str.resize(str.find("--"));
    if (IsSubstring("//",str))  str.resize(str.find("//"));
    for (long i=str.length()-1; i>=0; --i)
      if (str[i]==' ' || str[i]=='\t' || str[i]==';')  str.resize(i);
      else break;
    SearchMatch(found, str);
    if (found.size() == 0)
    {
      out << "\nNo matches for \"" << str << "\"" << endl;
      return;
    }
    if (found.size() == 1)
    {
      PrintSearchExact(found[0], out);
      return;
    }
    for (size_t i=0; i<found.size(); ++i)
      if (LowerCase(found[i]) == LowerCase(str))
      {
        PrintedFound = found[i];
        PrintSearchExact(PrintedFound, out);
      }
    if (PrintedFound == "")
      out << "\nAll matches for \"" << str << "\":" << endl;
    else
      out << "\nAll other matches for \"" << str << "\":" << endl;
    for (size_t i=0; i<found.size(); ++i)
      if (found[i] != PrintedFound)
        out << " ? " << found[i] << endl;
  }


  void ReloadMan(std::ostream &out)
  {
    UniqueIndex().myLoad(out);
  }
  

  void PrintSeeAlso(const std::string& s, std::ostream &out)
  {
    const index& index(UniqueIndex());
    for (size_t i=0; i<index.mySize(); ++i)
      if ( LowerCase(index.myTitle(i)) == LowerCase(s) )
      {
        if (index.mySeeAlso(i).size() != 0)
          out << "\nSee Also:" << endl;
        for (size_t j=0; j<index.mySeeAlso(i).size(); ++j)
          out << " ? " << index.mySeeAlso(i)[j] << endl;
        break;
      }
  }
  

void PrintAllExamples(std::ostream &out)
{
  ifstream in(XMLFileName());
  if (in == 0)
  {
    std::cout << "Cannot find input file `CoCoAHelp.xml`.  Aborting." << endl;
    abort();
  }
  string s = "";
  while (!in.eof())
  {
    s = SkipToStringInTag(in, "<title>");
    if (in.eof()) break;
    out << "--==============[ " << s << " ]===============--" << endl;
    string line;
    getline(in, line);
    while (!IsSubstring("<description>", line))  getline(in, line);
    getline(in, line);
    while (!IsSubstring("</description>", line))
    {
      if ( IsSubstring("<example>", line))  PrintExample(in, out);
      getline(in, line);
    }
  }
  out << "--- PrintAllExamples: done ---" << endl;
}


void PrintAllExamplesWithoutOutput(std::ostream &out)
{
  ifstream in(XMLFileName());
  if (in == 0)
  {
    std::cout << "Cannot find input file `CoCoAHelp.xml`.  Aborting." << endl;
    abort();
  }
  string s = "";
  while (!in.eof())
  {
    s = SkipToStringInTag(in, "<title>");
    if (in.eof()) break;
    out << "--==============[ " << s << " ]===============--" << endl;
    string line;
    getline(in, line);
    while (!IsSubstring("<description>", line))  getline(in, line);
    getline(in, line);
    while (!IsSubstring("</description>", line))
    {
      if ( IsSubstring("<example>", line))  PrintExample(in, out);
      getline(in, line);
    }
  }
  out << "--- PrintAllExamples: done ---" << endl;
}


const char* XMLFileName()
{
  static const string UniqueCopy(packageDir+"/../CoCoAManual/CoCoAHelp.xml");
  return UniqueCopy.c_str();
}


//const index& UniqueIndex()
index& UniqueIndex()
{
  static index UniqueCopy;
  return UniqueCopy;
}

} // namespace OnlineHelp
} // namespace CoCoA



// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/CoCoA-5/OnlineHelp.C,v 1.31 2012/04/04 13:56:35 bigatti Exp $
// $Log: OnlineHelp.C,v $
// Revision 1.31  2012/04/04 13:56:35  bigatti
// -- added PrintAllExamplesWithoutOutput
//
// Revision 1.30  2012/04/02 15:15:37  bigatti
// -- added check for trailing tab
//
// Revision 1.29  2012/03/20 14:34:09  bigatti
// -- added recognition of some xml tags
// -- minor cleaning
// -- Printing "matches for .." only when meaningful
//
// Revision 1.28  2012/03/13 15:59:12  bigatti
// -- xml source moved into CoCoAManual/
//
// Revision 1.27  2012/03/09 17:01:27  bigatti
// -- removed warning about old manual
// -- added "end example"
//
// Revision 1.26  2012/03/07 18:10:40  bigatti
// -- fixed interpretation of <backslash/>
//
// Revision 1.25  2012/02/24 13:10:12  bigatti
// -- added ReloadMan
//
// Revision 1.24  2012/01/25 14:33:15  bigatti
// -- added code for checking proper sorting (inactive)
//
// Revision 1.23  2011/11/02 14:41:16  bigatti
// -- added syntax
// -- removing trailing comments and spaces to searching string
// -- minor aesthetics change
//
// Revision 1.22  2011/07/06 15:52:13  bigatti
// -- improved IsSubstring
//
// Revision 1.21  2011/05/25 12:19:15  bigatti
// -- added case <par />
//
// Revision 1.20  2011/05/24 08:17:16  bigatti
// -- added warning also at the end of description
//
// Revision 1.19  2011/05/23 13:30:37  bigatti
// -- chencged include CoCoA/library --> CoCoA/error
//
// Revision 1.18  2011/05/13 10:26:37  bigatti
// -- added warning "thi is CoCoA-4 manual"
//
// Revision 1.17  2011/05/04 12:03:50  lagorio
// *** empty log message ***
//
// Revision 1.16  2011/03/23 17:31:22  bigatti
// -- added <code>..</code>
//
// Revision 1.15  2011/02/16 17:22:40  bigatti
// -- added "See Also"
//
// Revision 1.14  2011/02/16 16:14:09  bigatti
// -- class deefinition moved into .C file
// -- added storing of types, and function for <commands_and_functions_for>
// -- cleaning up
//
// Revision 1.13  2011/02/14 10:10:08  bigatti
// -- fixed &amp; &apos; <backslash/>  ...
// -- added function PrintAllExamples
//
// Revision 1.12  2011/01/25 15:45:23  bigatti
// -- added tags verbatim and formula
//
// Revision 1.11  2010/12/26 13:14:44  abbott
// Changed GlobalLogput to clog.
//
// Revision 1.10  2010/11/22 17:45:02  bigatti
// -- fixed bug in storing keys
//
// Revision 1.9  2010/11/22 17:37:44  bigatti
// -- improved message for entries with no matching
//
// Revision 1.8  2010/09/02 13:04:08  bigatti
// -- added conversion for <em>
//
// Revision 1.7  2010/09/01 13:24:48  bigatti
// -- moved all manual functions into CoCoA::OnlineHelp namespace
//
// Revision 1.6  2010/09/01 12:27:52  lagorio
// *** empty log message ***
//
// Revision 1.5  2010/09/01 09:23:19  bigatti
// -- added some "const"
// -- using CoCoA::GlobalLogput and CoCoA_ERROR instead of clog
//
// Revision 1.4  2010/09/01 08:28:35  lagorio
// *** empty log message ***
//
// Revision 1.3  2010/09/01 07:46:19  lagorio
// *** empty log message ***
//
// Revision 1.2  2010/08/31 14:55:02  bigatti
// -- pretty basic fixes
//

